import copy
import io
import os
import re
import subprocess
import tempfile
import traceback
from typing import List, Callable, Optional

from astra_installer.logs.logger import Logger
from astra_installer.settings.archive import Archive
from astra_installer.settings.progress_manager import ProgressManager


def readfile(pathfile: str) -> List[str]:
    with open(pathfile, 'r') as file:
        lines = file.readlines()
    return lines


def create_file(pathfile: str, text: str) -> None:
    Logger.log('creating file "{}" with content "{}"'.format(pathfile, text))
    with open(pathfile, 'w') as file:
        file.write(text)
        file.write('\n')

def remove_file(pathfile: str) -> None:
    if os.path.isfile(pathfile):
        Logger.log('removing file "{}"'.format(pathfile))
        os.remove(pathfile)


def append_file(pathfile: str, text: str) -> None:
    Logger.log('appending "{}" to file "{}"'.format(text, pathfile))
    with open(pathfile, 'a') as file:
        file.write('\n' + text)
        file.write('\n')


class CommandResult:
    def __init__(self, code: int, stdout: List[str], stderr: List[str]) -> None:
        self.code = code
        self.stdout = stdout
        self.stderr = stderr

    @property
    def is_success(self) -> bool:
        return self.code == 0

    @property
    def is_fail(self) -> bool:
        return self.code != 0


def run_command(command, *, spargs=None,
                stdin: Optional[str] = None, redirect_stderr: bool = True,
                enable_logging: bool = True, expect_success: bool = True,
                stdout_line_callback: Optional[Callable[[str], None]] = None) -> CommandResult:
    def log(message: str) -> None:
        if enable_logging:
            Logger.log(message)

    log('command: {command}'.format(command=command))
    arguments = {'shell': True, 'start_new_session': True,
                 'stdout': subprocess.PIPE, 'stderr': subprocess.PIPE}
    if spargs:
        arguments.update(spargs)
    if stdin:
        arguments.update({'stdin': subprocess.PIPE})
    if redirect_stderr:
        arguments.update({'stderr': subprocess.STDOUT})
    stdout = []  # type: List[str]
    stderr = []  # type: List[str]
    with subprocess.Popen(command, **arguments) as proc:
        if stdin:
            procstdin = io.TextIOWrapper(proc.stdin)
            procstdin.write(stdin)
            procstdin.close()
        for line in io.TextIOWrapper(proc.stdout, errors='backslashreplace'):
            line = line.strip()
            stdout.append(line)
            log(line)
            if stdout_line_callback:
                try:
                    stdout_line_callback(line)
                except Exception:
                    # Some processes like apt in media-change state can wait for terminal input forever
                    # and stdout_line_callback exception can't kill such process by itself
                    proc.terminate()
                    raise  # rethrow
        if not redirect_stderr:
            for line in io.TextIOWrapper(proc.stderr, errors='backslashreplace'):
                line = line.strip()
                stderr.append(line)
                log(line)
    result = CommandResult(proc.returncode, stdout, stderr)
    if result.is_fail and expect_success:
        raise RuntimeError
    else:
        return result


def run_debootstrap(arguments: List[str], use_eatmydata: bool = False) -> None:
    eatmydata = ['eatmydata', '--'] if use_eatmydata else []
    command_debootstrap = ' '.join([*eatmydata, 'debootstrap', *arguments])
    command_debootstrap_deps = ' '.join(
        [*eatmydata, 'debootstrap', '--print-debs', '--keep-debootstrap-dir', *arguments])
    result = run_command(command_debootstrap_deps, redirect_stderr=False)
    packages_to_install = []
    for line in result.stdout:
        packages = line.split()
        packages_to_install.extend(packages)
    # packages count * five steps of debootstrap for each package:
    # Retrieving, Validating, Extracting, Unpacking, Configuring
    expected_lines_count = len(packages_to_install) * 5
    lines_count = 0
    percentage = 0

    def report_progress(line: str) -> None:
        nonlocal lines_count
        nonlocal percentage
        lines_count += 1
        percentage_next = int(ProgressManager.percentage(lines_count, expected_lines_count))
        # Строк выводится несколько больше, чем ожидается, но этим можно пренебречь
        if percentage <= 100 and percentage != percentage_next:
            percentage = percentage_next
            ProgressManager.increment_by_percent_of_current_interval(percentage)

    run_command(command_debootstrap, stdout_line_callback=report_progress)


def run_async(command) -> None:
    subprocess.Popen(command)


def read_archive_size(archive: str) -> Optional[int]:
    sizefile = '{archive}.size'.format(archive=archive)
    Logger.log('Trying to get archive size from {sizefile}'.format(sizefile=sizefile))
    try:
        return int(readfile(sizefile)[0].strip())
    except:
        return None


def calculate_archive_size(archive: str) -> Optional[int]:
    Logger.log('Calculating size of {archive}'.format(archive=archive))
    try:
        result = run_command('zcat {archive} | wc -c'.format(archive=archive), redirect_stderr=False)
        return int(result.stdout[0].strip())
    except:
        return None


def extract_archive(archive: Archive, target: str) -> None:
    file = archive.path
    total = archive.size or read_archive_size(file) or calculate_archive_size(file)
    regex = re.compile(r'^(?!tar:).*?:\s*(?P<bytes>\d+).*$')
    percentage = 0

    def report_progress(line: str) -> None:
        nonlocal percentage
        match = regex.match(line)
        if match:
            extracted_bytes = int(match.group('bytes'))
            percentage_next = int(ProgressManager.percentage(extracted_bytes, total))
            if percentage != percentage_next:
                percentage = percentage_next
                ProgressManager.increment_by_percent_of_current_interval(percentage)

    arguments = ['tar',
                 '--acls',
                 '--xattrs',
                 '--xattrs-include=*',
                 '--totals',
                 '--checkpoint=2500',
                 '--checkpoint-action=totals',
                 '-pxvzf {file} -C {target}'.format(file=file, target=target),
                 '1>{logfile}'.format(logfile='/var/log/astra-installer-untar.log')]
    Logger.log('Extracting {archive}'.format(archive=file))
    run_command(' '.join(arguments), stdout_line_callback=report_progress)


# TODO: Can be optimized
# Just break loop if current package is uninstalled
def is_task_installed(task: str) -> bool:
    cmd = "tasksel --task-packages {task} 2>/dev/null | " \
          "grep --invert-match 'baytrail-install\\|fly-all-main' | " \
          "xargs dpkg-query --show --showformat='${{db:Status-Status}}\n' | " \
          "grep --invert-match '^installed'".format(task=task)
    result = run_command(cmd, expect_success=False)
    return not (result.stdout or result.stderr)


def is_security_feature_enabled(feature: str, chrooted: bool = True) -> bool:
    args = [feature, '' if chrooted else '/', 'status']
    return run_command(' '.join(args), expect_success=False).is_success


def check_media_change(line) -> None:
    if bool(re.search(r'media-change:', line)):
        raise RuntimeError('Disk requested')


def has_valid_repository() -> bool:
    Logger.log('Checking repository availability')
    try:
        chroot = Chroot('/')
        chroot.apt_get(['update'])
        chroot.apt_get(['check'])
        with tempfile.TemporaryDirectory() as tmpdir:
            # Apt uses APT::Status-FD=<file descriptor> to output progress
            # 2 is a standard output file descriptor
            # See: https://salsa.debian.org/apt-team/apt/blob/master/doc/progress-reporting.md
            command_template = 'cd {tmpdir} && {aptcom} -o APT::Status-FD=2 download tasksel'
            command = command_template.format(tmpdir=tmpdir, aptcom=' '.join(chroot.APT_ARGS))
            chroot.run_command(command, stdout_line_callback=check_media_change)
        return True
    except Exception:
        Logger.log('Repository unavailable:')
        Logger.log(traceback.format_exc())
        return False


def safe_exec(f: Callable, *args) -> None:
    try:
        f(*args)
    except Exception as e:
        Logger.log(str(e))


class Chroot:
    APT_ARGS = ['DEBIAN_FRONTEND=noninteractive',
                'apt-get', '-y',
                '--no-install-recommends',
                '-o', 'APT::Install-Recommends=no']
    APT_CONF_OPTS = ['-o', 'Dpkg::Options::=--force-confnew']

    def __init__(self, chroot: str) -> None:
        self.chroot = chroot

    def path(self, path: str) -> str:
        path = path.lstrip('/')
        return os.path.join(self.chroot, path)

    def isfile(self, path: str) -> bool:
        return os.path.isfile(self.path(path))

    def readfile(self, pathfile: str) -> List[str]:
        return readfile(self.path(pathfile))

    def create_file(self, pathfile: str, text: str) -> None:
        create_file(self.path(pathfile), text)

    def remove_file(self, pathfile: str) -> None:
        remove_file(self.path(pathfile))

    def append_file(self, pathfile: str, text: str) -> None:
        append_file(self.path(pathfile), text)

    def run_command(self, command, *, spargs=None, stdin: Optional[str] = None, redirect_stderr=True,
                    expect_success=True, stdout_line_callback: Optional[Callable[[str], None]] = None) -> CommandResult:
        command = 'chroot {target} /bin/bash -c \'{command}\''.format(target=self.chroot, command=command)
        return run_command(command, spargs=spargs, stdin=stdin, redirect_stderr=redirect_stderr,
                           expect_success=expect_success, stdout_line_callback=stdout_line_callback)

    def apt_get(self, arguments: List[str], confopts: Optional[List[str]] = None, spargs=None) -> None:
        args = copy.copy(self.APT_ARGS)
        if confopts:
            args.extend(confopts)
        else:
            args.extend(self.APT_CONF_OPTS)
        args.extend(arguments)
        regex = re.compile(r'^.*?:\s+\[\s*?(?P<progress>\d+)%\]$')

        def report_progress(line):
            match = regex.match(line)
            if match:
                ProgressManager.increment_by_percent_of_current_interval(int(match.group('progress')))

        self.run_command(command=' '.join(args), spargs=spargs, stdout_line_callback=report_progress)

    def apt_get_update(self) -> None:
        self.apt_get(['update'])

    def apt_get_dist_upgrade(self, report_progress=False) -> None:
        arguments = ['dist-upgrade']
        if report_progress:
            arguments = ['--show-progress', *arguments]
        self.apt_get(arguments)

    def apt_get_install(self, packages: List[str], report_progress=False) -> None:
        arguments = ['install', *packages]
        if report_progress:
            arguments = ['--show-progress', *arguments]
        self.apt_get(arguments)

    def apt_get_purge(self, packages: List[str], report_progress=False) -> None:
        arguments = ['purge', *packages]
        if report_progress:
            arguments = ['--show-progress', *arguments]
        self.apt_get(arguments)

    def is_installed(self, package: str) -> bool:
        command = 'dpkg -l | grep {package}'.format(package=package)
        return self.run_command(command, expect_success=False).is_success

    def set_selections(self, answers: List[str]) -> None:
        self.run_command('echo -e "{answers}" | debconf-set-selections'.format(answers='\n'.join(answers)))

    def get_packages_from_tasks(self, tasks: List[str]) -> List[str]:
        if len(tasks) == 0:
            return []
        arguments = ['--task-packages {task}'.format(task=task) for task in tasks]
        command = 'tasksel {arguments}'.format(arguments=' '.join(arguments))
        return self.run_command(command, redirect_stderr=False).stdout

    def aptitude(self, arguments: List[str], spargs=None) -> None:
        args = ['DEBIAN_FRONTEND=noninteractive',
                'aptitude', '--without-recommends', '-o', 'APT::Install-Recommends=no',
                '-o', 'Aptitude::CmdLine::Progress::Retain-Completed=true', '-y', '--allow-untrusted']
        args.extend(arguments)
        regex = re.compile(r'^.*?:\s+\[\s*?(?P<progress>\d+)%\]$')

        def report_progress(line):
            match = regex.match(line)
            if match:
                ProgressManager.increment_by_percent_of_current_interval(int(match.group('progress')))

        self.run_command(command=' '.join(args), spargs=spargs, stdout_line_callback=report_progress)

    def aptitude_install(self, packages: List[str], report_progress=False) -> None:
        arguments = ['install', *packages]
        self.aptitude(arguments)
