from typing import List, Dict, Any

from astra_installer.helpers import count
from astra_installer.logs.logger import Logger
from astra_installer.settings.archive import Archive
from astra_installer.settings.components import ComponentsSetting


class ArchivesSetting(ComponentsSetting):
    def __init__(self, archives: List[Archive]) -> None:
        ArchivesSetting.validate_archives(archives)
        self.__archives = archives

    @property
    def components(self) -> List[Archive]:
        return self.__archives

    @components.setter
    def components(self, archives: List[Archive]) -> None:
        ArchivesSetting.validate_archives(archives)
        self.__archives = archives

    @property
    def exclusive_components(self) -> bool:
        return True

    def load_components(self) -> None:
        pass

    @property
    def is_components_loaded(self) -> bool:
        return True

    @property
    def error(self) -> str:
        return ''

    @property
    def is_valid(self) -> bool:
        return True

    def log_selected(self) -> None:
        Logger.log('Selected archive is "{archive}"'.format(archive=str(self.selected_archive())))

    def dump_selected(self) -> Dict[str, Any]:
        return {'archive': str(self.selected_archive())}

    def selected_archive(self) -> Archive:
        return [archive for archive in self.__archives if archive.selected][0]

    @staticmethod
    def validate_archives(archives: List[Archive]) -> None:
        ArchivesSetting.check_archives_non_empty(archives)
        ArchivesSetting.check_only_one_archive_selected(archives)

    @staticmethod
    def check_archives_non_empty(archives: List[Archive]) -> None:
        if not archives:
            raise RuntimeError('Archives list can not be empty')

    @staticmethod
    def check_only_one_archive_selected(archives: List[Archive]) -> None:
        selected_archives_count = count(lambda archive: archive.selected, archives)
        if selected_archives_count != 1:
            raise RuntimeError('Exactly one archive should be selected. '
                               'Selected archives: {count}'.format(count=selected_archives_count))
