from typing import Dict, Any, TYPE_CHECKING, Optional

from astra_installer.settings.manager import SettingsManager
from astra_installer.settings.system_locale import Locale
from astra_installer.settings.system_locale import SystemLocaleSetting
from astra_installer.settings.tasks import TasksSetting
from astra_installer.storage.template.storage_template import StorageTemplate

if TYPE_CHECKING:
    from pyudev import Device
    from gettext import gettext as _


class AutoinstallSettingsManager(SettingsManager):
    @classmethod
    def create_settings(cls, config: Dict[str, Any]) -> None:
        cls.create_systeminfo_setting(config)
        cls.create_installer_locale_setting(config)
        cls.create_system_locale_setting(config)
        cls.create_layout_setting(config)
        cls.create_timezone_setting(config)
        cls.create_user_setting(config)
        cls.create_hostname_setting(config)
        cls.create_kernel_setting(config)
        cls.create_components_setting(config)
        cls.create_security_setting(config)
        cls.create_mode_setting(config)
        cls.create_disk_setting(config)
        cls.create_storage(config)
        cls.create_datetime_settings()

    @classmethod
    def create_systeminfo_setting(cls, config: Dict[str, Any]) -> None:
        super().create_systeminfo_setting(config)

    @classmethod
    def create_installer_locale_setting(cls, config: Dict[str, Any]) -> None:
        super().create_installer_locale_setting(config)

    @classmethod
    def create_hostname_setting(cls, config: Dict[str, Any]) -> None:
        super().create_hostname_setting(config)
        if 'hostname' in config['preseed']:
            hostname = config['preseed']['hostname']
            SettingsManager.hostname.hostname = hostname

    @classmethod
    def create_timezone_setting(cls, config: Dict[str, Any]) -> None:
        super().create_timezone_setting(config)
        if 'timezone' in config['preseed']:
            timezone = config['preseed']['timezone']
            SettingsManager.timezone.selected = timezone

    @classmethod
    def create_user_setting(cls, config: Dict[str, Any]) -> None:
        super().create_user_setting(config)
        if 'username' in config['preseed']['user']:
            username = config['preseed']['user']['username']
            SettingsManager.user.username = username
        if 'password' in config['preseed']['user']:
            password = config['preseed']['user']['password']
            SettingsManager.user.password = password
        if 'password-encrypted' in config['preseed']['user']:
            password = config['preseed']['user']['password-encrypted']
            SettingsManager.user.password = password
            SettingsManager.user.encrypted = True
        if 'grub-password' in config['preseed']['user']:
            grub_password = config['preseed']['user']['grub-password']
            SettingsManager.user.grub_password = grub_password
        if 'grub-password-encrypted' in config['preseed']['user']:
            grub_password = config['preseed']['user']['grub-password-encrypted']
            SettingsManager.user.grub_password = grub_password
            SettingsManager.user.grub_password_encrypted = True

    @classmethod
    def create_kernel_setting(cls, config: Dict[str, Any]) -> None:
        super().create_kernel_setting(config)
        if SettingsManager.kernel.package == 'linux-image-*-generic':
            return
        if 'kernel-package' in config['preseed']:
            kernel_package = config['preseed']['kernel-package']
            if not kernel_package in SettingsManager.kernel.available_packages:
                return
            SettingsManager.kernel.package = kernel_package

    @classmethod
    def create_system_locale_setting(cls, config: Dict[str, Any]) -> None:
        super().create_system_locale_setting(config)
        if 'locales' in config['preseed']:
            locales = config['preseed']['locales']
            selected_locales = []
            for locale in locales:
                locale = Locale.from_string(locale)
                lnt = locale.language_and_territory()
                selected_locales.append(lnt)
                if lnt not in SettingsManager.system_locale.locales:
                    SettingsManager.system_locale.locales.append(lnt)
            selected_locales.append(SystemLocaleSetting.additional_locale.language_and_territory())
            SettingsManager.system_locale.selected = selected_locales

    @classmethod
    def create_layout_setting(cls, config: Dict[str, Any]) -> None:
        super().create_layout_setting(config)
        if 'layouts' in config['preseed']:
            layouts = config['preseed']['layouts']
            SettingsManager.layout.selected_keys = layouts
        if 'layout_switch_keys' in config['preseed']:
            layout_switch_keys = config['preseed']['layout_switch_keys']
            SettingsManager.layout.selected_keys = layout_switch_keys

    @classmethod
    def create_tasks_setting(cls, config: Dict[str, Any]) -> TasksSetting:
        task_setting = super().create_tasks_setting(config)
        if 'tasks' in config['preseed']:
            tasks = config['preseed']['tasks']
            for task in task_setting.tasks:
                task.selected = task.ident in tasks
        return task_setting

    @classmethod
    def create_security_setting(cls, config: Dict[str, Any]) -> None:
        super().create_security_setting(config)
        if 'features' in config['preseed']:
            features = config['preseed']['features']
            for feature in SettingsManager.security.features:
                feature.selected = feature.ident in features

    @classmethod
    def create_mode_setting(cls, config: Dict[str, Any]) -> None:
        super().create_mode_setting(config)
        if 'mode' in config['preseed']:
            selected_mode_id = config['preseed']['mode']
            if selected_mode_id not in {0, 1, 2}:
                raise Exception(_('Wrong mode value: ') + selected_mode_id)
            for mode in SettingsManager.mode.modes:
                if mode.ident == selected_mode_id:
                    mode.selected = True
                else:
                    mode.selected = False
            mode_features = {feature.ident: feature for feature in SettingsManager.mode.selected_mode().features}
            for feature in SettingsManager.security.features:
                if feature.ident in mode_features and not mode_features[feature.ident].enabled:
                    feature.enabled = mode_features[feature.ident].enabled
                    feature.selected = mode_features[feature.ident].selected
        else:
            raise Exception(_('The security level is not specified'))

    @classmethod
    def create_disk_setting(cls, config: Dict[str, Any]) -> None:
        if 'repos' in config['preseed']:
            config['repos'] = config['preseed']['repos']
            config['repo_source'] = 'config'
        if 'late_command' in config['preseed']:
            config['late_command'] = config['preseed']['late_command']
        if 'late_command_intarget' in config['preseed']:
            config['late_command_intarget'] = config['preseed']['late_command_intarget']
        super().create_disk_setting(config)
        SettingsManager.disk.reboot_after_install = True

    @classmethod
    def create_storage(cls, config: Dict[str, Any]) -> None:
        super().create_storage(config)

        if 'storage' not in config['preseed']:
            raise Exception(_('Storage is required section'))

        if 'config' in config['preseed']['storage']:
            cls.setup_storage_by_config(config)
        else:
            cls.setup_storage_by_template(config)

    @classmethod
    def create_datetime_settings(cls) -> None:
        super().create_datetime_settings()

    @classmethod
    def setup_storage_by_config(cls, config: Dict[str, Any]) -> None:
        config = config['preseed']['storage']['config']

        report, state = SettingsManager.storage.config_text_state(config)
        if not report.is_valid():
            error_text = '\n'.join([str(message) for message in report.get_messages()])
            raise Exception(_('Configuration error: ') + error_text)

        template = StorageTemplate(name=_('Custom disk layout'))
        template.config_text = config
        template.storage_state = state

        SettingsManager.storage.templates = SettingsManager.storage.templates + [template]
        SettingsManager.storage.current_template = template

        error = SettingsManager.storage.validate()
        if error:
            raise Exception(error)

    @classmethod
    def setup_storage_by_template(cls, config: Dict[str, Any]) -> None:
        devices = SettingsManager.storage.all_physical_devices()
        requested_device = config['preseed']['storage']['device']
        selected_device = None  # type: Optional[Device]
        for device in devices:
            if requested_device == device.device_node:
                selected_device = device
                break
        else:
            raise Exception(_('Storage device not found'))

        templates = SettingsManager.storage.templates
        requested_template = config['preseed']['storage'].get('template', 'default_template.txt')
        selected_template = None  # type: Optional[StorageTemplate]
        for template in templates:
            if requested_template == template.config_file.filename:
                selected_template = template
                break
        else:
            raise Exception(_('Template not found'))

        template.usable_device = selected_device
        if 'gpt' in config['preseed']['storage']:
            template.gpt_is_used = config['preseed']['storage']['gpt']
        if 'bios' in config['preseed']['storage']:
            template.install_into_bios = config['preseed']['storage']['bios']

        report, state = SettingsManager.storage.init_template(selected_template)
        if not report.is_valid():
            error_text = '\n'.join([str(message) for message in report.get_messages()])
            raise Exception(_('Configuration error: ') + error_text)

        template.storage_state = state
        SettingsManager.storage.current_template = template

        error = SettingsManager.storage.validate()
        if error:
            raise Exception(error)
