import os
from typing import Dict, Any, List, TYPE_CHECKING

import yaml

from astra_installer.logs.logger import Logger
from astra_installer.utils import Utils

if TYPE_CHECKING:
    from gettext import gettext as _


class ConfigLoader:
    def __init__(self, oem: bool = False) -> None:
        self.oem = oem

    def load(self) -> Dict[str, Any]:
        Logger.log('Loading installer configs')
        installer_config_file = Utils.file_path('installer.conf.yaml')
        installer_config = ConfigLoader.load_file(installer_config_file)
        common_config_file = Utils.file_path('configs', 'common.conf.yaml')
        common_config = ConfigLoader.load_file(common_config_file)
        security_scripts_path = Utils.file_path('configs', 'security')
        hooks_path = Utils.file_path('configs', 'hooks')
        arch = Utils.value_or_auto(installer_config.get('arch', 'auto'), Utils.get_arch_from_environment).lower()
        test = installer_config.get('test', False)
        section = 'oem' if self.oem else 'installer'
        specific_configs_path = Utils.file_path('configs', section, arch)
        specific_config = ConfigLoader.load_file(Utils.file_path(specific_configs_path, 'conf.yaml'))
        config = {'scripts_path': specific_configs_path,
                  'security_scripts_path': security_scripts_path,
                  'hooks_path': hooks_path,
                  'oem': self.oem,
                  'test': test}
        config.update(common_config)
        config.update(specific_config)

        disk_layout_templates = ConfigLoader.get_disk_layout_templates(config.get('disk_layout_templates'),
                                                                       section, arch)
        config.update({'disk_layout_templates': disk_layout_templates})

        reserved_usernames_file = Utils.file_path('configs', 'reserved-usernames.yaml')
        reserved_usernames = ConfigLoader.load_file(reserved_usernames_file)
        config.update(reserved_usernames)

        return config

    @staticmethod
    def load_file(path) -> Dict[str, Any]:
        Logger.log('Loading file {file}'.format(file=path))
        with open(path) as file:
            return yaml.load(file)

    @staticmethod
    def get_disk_layout_templates(requested_templates_data: List[Dict[str, str]],
                                  section: str, architecture: str) -> List[Dict[str, str]]:
        disk_layout_templates = []  # type: List[Dict[str, str]]
        for template_data in requested_templates_data:
            filename = template_data.get('filename')
            common_path = Utils.file_path('configs', 'installer', 'common', 'disk_layout_templates', filename)
            specific_path = Utils.file_path('configs', section, architecture, 'disk_layout_templates', filename)
            if os.path.isfile(specific_path):
                path = specific_path
            elif os.path.isfile(common_path):
                path = common_path
            else:
                raise Exception(_('File {path} not fount').format(path=common_path))

            name = template_data.get('name', _('Untitled'))
            description = template_data.get('description', None)
            gpt_support = template_data.get('gpt_support', False)
            recomended = template_data.get('recomended', False)
            red_book = template_data.get('red_book', False)
            snapshots_support = template_data.get('snapshots_support', False)
            is_default = template_data.get('is_default', False)

            disk_layout_templates.append({'name': name,
                                          'description': description,
                                          'filename': filename,
                                          'path': path,
                                          'gpt_support': gpt_support,
                                          'recomended': recomended,
                                          'red_book': red_book,
                                          'snapshots_support': snapshots_support,
                                          'is_default': is_default})

        return disk_layout_templates
