from datetime import datetime

from typing import Optional, TYPE_CHECKING

from astra_installer.commands import run_command
from astra_installer.logs.logger import Logger

if TYPE_CHECKING:
    from gettext import gettext as _

class DateTimeSetting:
    def __init__(self) -> None:
        self.datetime = None  # type: Optional[datetime]
        self.datetime_changed = False
        self.__error = None  # type: Optional[str]

    @property
    def error(self) -> Optional[str]:
        return self.__error

    def current_datetime(self):
        self.datetime = datetime.now()
        return self.datetime

    def is_valid(self) -> bool:
        return self.datetime is not None

    def apply(self) -> bool:
        dt = DateTimeSetting.datetime_to_str(self.datetime)
        Logger.log('Setting date and time to {datetime}'.format(datetime=dt))
        if self.set_date() is False:
            return False
        if self.set_time() is False:
            return False

        return True

    def set_datetime(self, date, time) -> bool:
        date = datetime(date.year, date.month, date.day,
                        time.hour, time.minute, time.second, time.microsecond)
        commands = ['timedatectl set-ntp false',
                    'timedatectl set-time "{text}"'.format(text=date.strftime('%Y-%m-%d %H:%M:%S'))]
        command = ' && '.join(commands)
        process = run_command(command=command, expect_success=False)
        if process.is_success:
            self.datetime_changed = True
            self.__error = None
            return True
        else:
            output = '. '.join(process.stdout)
            if len(output) > 0:
                self.__error = _('Command "{command}". Output "{output}"').format(command=command, output=output)
            else:
                self.__error = None
            return False

    def set_date(self, dtime=None) -> bool:
        date = self.datetime if dtime is None else dtime
        time = datetime.now().time()
        return self.set_datetime(date, time)

    def set_time(self, dtime=None) -> bool:
        time = self.datetime if dtime is None else dtime
        date = datetime.now().date()
        return self.set_datetime(date, time)

    @staticmethod
    def datetime_to_str(dt) -> str:
        return dt.strftime("%Y-%m-%d-%H.%M.%S")

    def log_selected(self) -> None:
        self.current_datetime()
        dt = DateTimeSetting.datetime_to_str(self.datetime)
        Logger.log('Selected datetime "{datetime}"'.format(datetime=dt))
