import traceback
import re

from typing import Dict, Any, TYPE_CHECKING

from distutils.version import StrictVersion

from astra_installer.commands import run_command
from astra_installer.logs.logger import Logger

if TYPE_CHECKING:
    from gettext import gettext as _

class KernelSetting:
    kernel_package_regexp = re.compile("^linux-\d+.\d+-(generic|hardened)$")
    available_packages = []

    def __init__(self, package: str, features: Dict[str, Any], ram_requirements: Dict[str, Any] = {}) -> None:
        super().__init__()
        self.package = package
        self.features = features
        self.ram_requirements = ram_requirements

    def get_ram_requirement(self, package: str) -> int:
        return int(self.ram_requirements.get(package.replace('-generic', '') \
                                                    .replace('-hardened', ''), 0))

    def is_enough_ram(self, available_ram: int) -> bool:
        if self.package == "linux-image-*-generic":
            return True
        if not self.ram_requirements:
            return True

        ram_req = self.get_ram_requirement(self.package)
        if (ram_req > 0) and (available_ram < ram_req):
            return False
        
        current_version = StrictVersion(self.package.replace('linux-', '').replace('-generic', '').replace('-hardened', ''))
        max_available_req = max(self.ram_requirements,
                                key=lambda x: StrictVersion(x.replace('linux-', '')),
                                default='')
        max_req_version = StrictVersion(max_available_req.replace('linux-', ''))
        if current_version < max_req_version:
            return True
        ram_req = self.get_ram_requirement(max_available_req)
        if (ram_req > 0) and (available_ram < ram_req):
            return False

        return True

    def log_selected(self) -> None:
        Logger.log('Selected kernel is "{kernel}"'.format(kernel=self.package))

    def dump_selected(self) -> Dict[str, Any]:
        return {'kernel-package': self.package}

    @staticmethod
    def find_available_packages() -> None:
        try:
            packages = run_command('apt-cache search linux- | grep Complete').stdout
            for package in packages:
                pkg = package.split(' ', 1)
                if KernelSetting.kernel_package_regexp.match(pkg[0]):
                    KernelSetting.available_packages.append(pkg[0])
            KernelSetting.available_packages.sort()
        except Exception:
            Logger.log(traceback.format_exc())

    @staticmethod
    def get_default_generic() -> str:
        if not KernelSetting.available_packages:
            return ''
        generics = list(filter(lambda x: x.find('-generic') >= 0, KernelSetting.available_packages))
        return max(generics,
                    key=lambda x: StrictVersion(x.replace('linux-', '').replace('-generic', '')),
                    default='')
