import copy
import re
from typing import Set, Optional, List, Dict, Any

from astra_installer.commands import run_command
from astra_installer.logs.logger import Logger


class XKBData:
    default_options = {'grp': set(), 'lv3': set(), 'ctrl': set(), 'grp_led': set(), 'keypad': set(),
                       'kpdl': set(), 'caps': set(), 'altwin': set(), 'compose': set(), 'compat': set(),
                       'numpad': set(), 'srvrkeys': set(), 'apple': set(), 'shift': set(), 'misc': set(),
                       'grab': set(), 'currencysign': set(), 'eurosign': set(), 'rupeesign': set(),
                       'lv5': set(), 'nbsp': set(), 'japan': set(), 'korean': set(), 'esperanto': set(),
                       'solaris': set(), 'terminate': set()}  # type: Dict[str, Set[str]]

    def __init__(self, model: str = 'pc105', layouts: Optional[Set[str]] = None, variant: Optional[Set[str]] = None,
                 options: Optional[Dict[str, Set[str]]] = None, backspace: str = 'guess',
                 kmap: Optional[str] = None) -> None:
        self.model = model
        self.layouts = layouts
        self.variants = variant
        self.options = options
        self.backspace = backspace
        self.kmap = kmap

        if not self.layouts:
            self.layouts = set()

        if not self.variants:
            self.variants = set()

        if not self.options:
            self.options = XKBData.default_options

    def layouts_string(self, priority_layout: Optional[str] = None) -> str:
        # TODO: use ordered set
        if priority_layout and priority_layout in self.layouts:
            layouts = copy.copy(self.layouts)
            layouts.remove(priority_layout)
            return ','.join([priority_layout] + list(layouts))
        else:
            return ','.join(self.layouts)

    def variants_string(self) -> str:
        return ','.join(self.variants)

    def options_string(self) -> str:
        option_strings = []
        for name, values in self.options.items():
            if values:
                for element in values:  # set
                    option_strings.append('{key}:{value}'.format(key=name, value=element))
        return ','.join(option_strings)

    def keyboard_content(self) -> str:
        opts = []
        opts.append('XKBMODEL="{}"'.format(self.model or 'pc105'))
        opts.append('XKBLAYOUT="{}"'.format(self.layouts_string('us') or 'us'))
        opts.append('XKBVARIANT="{}"'.format(self.variants_string() or ','))
        if self.options:
            opts.append('XKBOPTIONS="{}"'.format(self.options_string()))
        if self.backspace:
            opts.append('BACKSPACE="{}"'.format(self.backspace))
        if self.kmap:
            opts.append('KMAP="{}"'.format(self.kmap))
        return '\n'.join(opts)

    def setxkbmap_command(self) -> str:
        args = ['setxkbmap']
        args.append('-model "{}"'.format(self.model or 'pc105'))
        args.append('-layout "{}"'.format(self.layouts_string('us') or 'us'))
        args.append('-variant "{}"'.format(self.variants_string() or ','))
        if self.kmap:
            args.append('-keymap "{}"'.format(self.kmap))
        if self.options:
            args.append('-option -option "{}"'.format(self.options_string()))
        return ' '.join(args)

    @staticmethod
    def from_string(keyboard: str) -> Optional['XKBData']:
        data = XKBData()
        model = re.search(r'XKBMODEL=\"?(\w+)\"?', keyboard)
        layout = re.search(r'XKBLAYOUT=\"?([\w,]+)\"?', keyboard)
        variant = re.search(r'XKBVARIANT=\"?([\w,]+)\"?', keyboard)
        options = re.search(r'XKBOPTIONS=\"?([\w:,]+)\"?', keyboard)
        backspace = re.search(r'BACKSPACE=\"?(\w+)\"?', keyboard)
        kmap = re.search(r'KMAP=\"?(\w+)\"?', keyboard)
        data.model = model.group(1) if model else 'pc105'
        data.layouts = set(layout.group(1).split(',')) if layout else set()
        data.variants = set(variant.group(1).split(',')) if variant else set()
        data.options = XKBData.parse_options(options.group(1)) if options else XKBData.parse_options('')
        data.backspace = backspace.group(1) if backspace else None
        data.kmap = kmap.group(1) if kmap else None
        return data

    @staticmethod
    def parse_options(optionsstr: str) -> Dict[str, Set[str]]:
        options = copy.copy(XKBData.default_options)
        for optionstr in optionsstr.split(','):
            match = re.match(r'^(?P<key>.+?):(?P<value>.+?)$', optionstr)
            if match:
                options[match.group('key')].add(match.group('value'))
        return options


class LayoutSetting:
    additional_layout = 'us'
    keyboard_file = '/etc/default/keyboard'

    def __init__(self, layouts: List[str], default_layout: str,
                 country_default_layouts: Dict[str, List[str]],
                 keys: Dict[str, str], default_key: str) -> None:
        super().__init__()
        self.layouts = layouts
        self.default_layout = default_layout
        self.country_default_layouts = country_default_layouts
        self.selected_layouts = [self.default_layout]
        self.keys = keys
        self.default_key = default_key
        self.selected_keys = [self.default_key]
        self.xkbdata = XKBData()

    def country_layouts(self, country: str) -> List[str]:
        return self.country_default_layouts.get(country, [self.default_layout])

    def select_default_layouts_for_country(self, country: str) -> None:
        self.selected_layouts = self.country_layouts(country)

    def is_valid(self) -> bool:
        return self.selected_layouts is not None and self.selected_keys is not None

    def apply(self) -> None:
        Logger.log('Applying keyboard options')
        run_command(self.export_xkbdata().setxkbmap_command())

    @staticmethod
    def layout_from_locale(locale: str) -> str:
        return locale[3:].lower()

    @staticmethod
    def layouts_from_locales(locales: List[str]) -> List[str]:
        return [LayoutSetting.layout_from_locale(locale) for locale in locales]

    def log_selected(self) -> None:
        Logger.log('Selected layouts are "{layouts}"'.format(layouts=','.join(self.selected_layouts)))
        Logger.log('Selected keys are "{keys}"'.format(keys=','.join(self.selected_keys)))

    def import_xkbdata(self, xkbdata: XKBData) -> None:
        self.xkbdata = xkbdata
        if xkbdata.layouts:
            self.layouts = list(set(self.layouts) | xkbdata.layouts)
            self.selected_layouts = list(xkbdata.layouts)
        grp_options = xkbdata.options.get('grp')
        if grp_options:
            ks = set(self.keys.keys())
            self.selected_keys.clear()
            for grp_option in grp_options:
                if not grp_option in ks:
                    self.keys.update({grp_option: grp_option})
                self.selected_keys.append(grp_option)
            self.default_key = next(iter(self.selected_keys))

    def export_xkbdata(self) -> XKBData:
        self.xkbdata.layouts = set(self.selected_layouts + [self.additional_layout])
        self.xkbdata.options['grp'] = set(self.selected_keys)
        return self.xkbdata

    def dump_selected(self) -> Dict[str, Any]:
        return {'layouts': self.selected_layouts, 'layout_switch_keys': self.selected_keys}
