import os
import re
from typing import Dict, Any, Optional, List, Set, Tuple

from astra_installer import globvars
from astra_installer.settings.archive import Archive
from astra_installer.settings.archives import ArchivesSetting
from astra_installer.settings.component import ComponentId
from astra_installer.settings.components import ComponentsSetting
from astra_installer.settings.date_time import DateTimeSetting
from astra_installer.settings.disk import DiskSetting
from astra_installer.settings.hostname import HostnameSetting
from astra_installer.settings.installer_locale import InstallerLocaleSetting
from astra_installer.settings.layout import LayoutSetting
from astra_installer.settings.repository import Repository
from astra_installer.settings.security import SecuritySetting
from astra_installer.settings.security_feature import Executable
from astra_installer.settings.security_feature import SecurityFeature
from astra_installer.settings.kernel import KernelSetting
from astra_installer.settings.mode import Mode
from astra_installer.settings.mode import ModeSetting
from astra_installer.settings.system_locale import SystemLocaleSetting
from astra_installer.settings.systeminfo import AstraEdition
from astra_installer.settings.systeminfo import SystemInfo
from astra_installer.settings.task import Task
from astra_installer.settings.tasks import TasksSetting
from astra_installer.settings.timezone import TimezoneSetting
from astra_installer.settings.user import UserSetting
from astra_installer.translator import deferredtr as _tr
from astra_installer.utils import Utils

if not globvars.oem_mode:
    from astra_installer.storage.storage import Storage


class SettingsManager:
    installer_locale = None  # type: InstallerLocaleSetting
    system_locale = None  # type: SystemLocaleSetting
    layout = None  # type: LayoutSetting
    timezone = None  # type: TimezoneSetting
    user = None  # type: UserSetting
    hostname = None  # type: HostnameSetting
    kernel = None # type: KernelSetting
    components = None  # type: ComponentsSetting
    archives = None  # type: ArchivesSetting
    security = None  # type: SecuritySetting
    mode = None  # type: ModeSetting
    disk = None  # type: DiskSetting
    storage = None  # type: Storage
    systeminfo = None  # type: SystemInfo
    datetime = None  # type: DateTimeSetting

    @classmethod
    def create_settings(cls, config: Dict[str, Any]) -> None:
        cls.create_systeminfo_setting(config)
        cls.create_installer_locale_setting(config)
        cls.create_system_locale_setting(config)
        cls.create_layout_setting(config)
        cls.create_timezone_setting(config)
        cls.create_user_setting(config)
        cls.create_hostname_setting(config)
        cls.create_kernel_setting(config)
        cls.create_components_setting(config)
        cls.create_security_setting(config)
        cls.create_mode_setting(config)
        cls.create_disk_setting(config)
        cls.create_storage(config)
        cls.create_datetime_settings()

    @classmethod
    def create_systeminfo_setting(cls, config: Dict[str, Any]) -> None:
        efi = Utils.is_efi_supported()
        bios_compat_mode_supported = efi and config.get('bios_compat_mode_supported', True)
        ram = Utils.system_ram_size()
        arch = Utils.value_or_auto(config['info']['arch'], Utils.get_arch_from_environment)
        name = Utils.value_or_auto(config['info']['name'], arch)
        edition = AstraEdition.SE
        release = config['info']['release']
        scripts_path = config['scripts_path']
        oem = config['oem']
        test = config['test']
        SettingsManager.systeminfo = SystemInfo(efi, bios_compat_mode_supported, ram,
                                                arch, name, edition, release,
                                                scripts_path, oem, False, True, set(), test)

    @classmethod
    def create_installer_locale_setting(cls, config: Dict[str, Any]) -> None:
        locales = config['installer_locale']['all']
        default = config['installer_locale']['default']
        SettingsManager.installer_locale = InstallerLocaleSetting(locales, default)

    @classmethod
    def create_system_locale_setting(cls, config: Dict[str, Any]) -> None:
        locales = config['system_locale']['all']
        default = config['system_locale']['default']
        country_default = config['system_locale']['country_default']
        SettingsManager.system_locale = SystemLocaleSetting(locales, default, country_default)

    @classmethod
    def create_layout_setting(cls, config: Dict[str, Any]) -> None:
        layouts = config['layout']['layouts']['all']
        default_layout = config['layout']['layouts']['default']
        country_default_layouts = config['layout']['layouts']['country_default']
        keys = config['layout']['keys']['all']
        default_key = config['layout']['keys']['default']
        SettingsManager.layout = LayoutSetting(layouts, default_layout,
                                               country_default_layouts, keys, default_key)

    @classmethod
    def create_timezone_setting(cls, config: Dict[str, Any]) -> None:
        tz_settings_config = config.get('timezone_settings', {})
        SettingsManager.timezone = TimezoneSetting(tz_settings_config, tz_settings_config.get('default'))

    @classmethod
    def create_user_setting(cls, config: Dict[str, Any]) -> None:
        username = config['user']['username']
        SettingsManager.user = UserSetting(username,
                                           setup_grub_password=(SettingsManager.systeminfo.edition == AstraEdition.SE))
        reserved_names = config['reserved_usernames']
        SettingsManager.user.username_blacklist.update(reserved_names)

    @classmethod
    def create_hostname_setting(cls, config: Dict[str, Any]) -> None:
        hostname = config['host']['hostname']
        SettingsManager.hostname = HostnameSetting(hostname)

    @classmethod
    def create_kernel_setting(cls, config: Dict[str, Any]) -> None:
        kernel_choise_state = 'kernel-choise-available' 
        KernelSetting.find_available_packages()
        default_generic = KernelSetting.get_default_generic()
        if not default_generic:
            default_generic = 'linux-image-*-generic'
            kernel_choise_state = 'kernel-choise-unavailable'
        features = config['kernel'][kernel_choise_state]['features']
        ram_requirements = config['kernel'].get('kernel_ram_requirements', {})
        SettingsManager.kernel = KernelSetting(default_generic, features, ram_requirements)

    @classmethod
    def create_components_setting(cls, config: Dict[str, Any]) -> None:
        if 'archives' in config:
            SettingsManager.components = cls.create_archives_setting(config)
        else:
            SettingsManager.components = cls.create_tasks_setting(config)

    @classmethod
    def create_tasks_setting(cls, config: Dict[str, Any]) -> TasksSetting:
        tasks = []
        for index, task in enumerate(config['tasks']):
            ident = task['id']
            name = task['name'] if 'name' in task else Task.get_name_by_id(ident)
            desc = task.get('desc', '')
            enabled = bool(task.get('enabled', True))
            selected = bool(task.get('selected', False))
            visible = bool(task.get('visible', True))
            priority = float(task.get('priority', float(index)))
            dependencies = []  # type: List[ComponentId]
            enhances = set()  # type: Set[ComponentId]
            if 'enhances' in task:
                enhances = set(task['enhances'])
                selected = False
                visible = False
            hook_path = None
            hooks_path = config['hooks_path']
            if os.path.isfile("{hooks_path}/tasks/{ident}".format(hooks_path=hooks_path, ident=ident)):
                hook_path = "{hooks_path}/tasks/{ident}".format(hooks_path=hooks_path, ident=ident)

            kernel_feature = SettingsManager.kernel.features.get(ident)
            if kernel_feature:
                visible = bool(kernel_feature.get('visible', visible))
                enabled = bool(kernel_feature.get('enabled', enabled))
                selected = bool(kernel_feature.get('selected', selected))

            tasks.append(Task(ComponentId(ident), name, desc, enabled, selected,
                              visible, priority, hook_path, dependencies, enhances))
        excluded_packages = config.get('excluded_packages', {})
        return TasksSetting(tasks, SettingsManager.systeminfo.has_valid_repository,
                            SettingsManager.systeminfo.installedpkgs, excluded_packages)

    @classmethod
    def create_archives_setting(cls, config: Dict[str, Any]) -> ArchivesSetting:
        archives = []
        for archive in config['archives']:
            name = archive['name']
            path = archive['path']
            size = None  # type: Optional[int]
            if 'size' in archive:
                size = int(archive['size'])
            desc = archive.get('description', '')
            enabled = bool(archive.get('enabled', True))
            selected = bool(archive.get('selected', False))
            priority = float(archive.get('priority', 0.0))
            archives.append(Archive(ComponentId(name), name, desc, enabled, selected, True, priority, [], path, size))
        if archives and not any(archive.selected for archive in archives):
            archives[0].selected = True
        return ArchivesSetting(archives)

    @classmethod
    def create_security_setting(cls, config: Dict[str, Any]) -> None:
        features = []
        for index, feature in enumerate(config['security_features']):
            ident = feature['id']
            name = feature['name'] if 'name' in feature else SecurityFeature.get_name_by_id(ident)
            desc = feature.get('desc', '')
            enabled = bool(feature.get('enabled', True))
            selected = bool(feature.get('selected', False))
            visible = bool(feature.get('visible', True))
            priority = float(feature.get('priority', float(index)))
            depends = feature.get('depends', [])
            packages = feature.get('packages', [])
            use_username = feature.get('use_username', False)
            distr_override = config['scripts_path']
            common_override = config['security_scripts_path']

            tasks = []  # type: List[ComponentId]
            if 'tasks' in feature:
                tasks = feature['tasks']

            distr_script = Executable(path='{path}/security/{script}'.format(path=distr_override, script=ident),
                                      use_username=use_username)
            common_script = Executable(path='{path}/{script}'.format(path=common_override, script=ident),
                                       use_username=use_username)
            script = Executable(path='/usr/sbin/{script}'.format(script=ident),
                                chrooted=True,
                                use_username=use_username)

            if os.path.isfile(distr_script.path):
                script = distr_script
            elif os.path.isfile(common_script.path):
                script = common_script

            if len(packages) > 0:
                desc = _tr('additional packages may be installed')

            kernel_feature = SettingsManager.kernel.features.get(ident)
            if kernel_feature:
                visible = bool(kernel_feature.get('visible', visible))
                enabled = bool(kernel_feature.get('enabled', enabled))
                selected = bool(kernel_feature.get('selected', selected))

            features.append(SecurityFeature(ComponentId(ident), name, desc, '',
                                            enabled, selected, visible, priority,
                                            depends, script, packages, tasks))
        SettingsManager.security = SecuritySetting(features, SettingsManager.systeminfo.has_valid_repository,
                                                   SettingsManager.systeminfo.is_kiosk_session)

    @classmethod
    def create_mode_setting(cls, config: Dict[str, Any]) -> None:
        modes = []  # type: List[Mode]
        for index, mode in enumerate(config['modes']):
            ident = mode['id']
            name = mode.get('name', str(ident))
            desc = mode.get('desc', name)
            selected = bool(mode.get('selected', False))
            features = []  # type: List[SecurityFeature]
            for feature in mode['features']:
                feature_ident = feature['id']
                feature_enabled = bool(feature.get('enabled', True))
                feature_selected = bool(feature.get('selected', False))
                features.append(SecurityFeature(ComponentId(feature_ident), name, '', '',
                                                feature_enabled, feature_selected, True,
                                                1.0, [], Executable(''), [], []))
            modes.append(Mode(ident, name, desc, selected, features))
        SettingsManager.mode = ModeSetting(modes)

    @classmethod
    def get_sources_from_sourceslist(cls) -> Tuple[List[Repository], Repository]:
        sources = []
        with open('/etc/apt/sources.list', 'r') as sources_list:
            for line in sources_list:
                source_re = re.compile('\\s*deb\\s+(\\[.*?]\\s+)?(?P<uri>.*?)\\s+.*')
                match = source_re.match(line)
                if match:
                    source = Repository(match.group('uri').strip())
                    sources.append(source)
        if sources:
            return sources, sources[0]
        else:
            raise Exception('No repositories found in sources.list')

    @classmethod
    def get_sources_from_config(cls, config: Dict[str, Any]) -> Tuple[List[Repository], Repository]:
        sources = []
        if 'repos' in config:
            for source in config['repos']['sources'] or []:
                sources.append(Repository(source))
            debootstrap = Repository(config['repos']['debootstrap'])
            return sources, debootstrap
        else:
            raise Exception('No repositories found in config')

    @classmethod
    def find_sources(cls, config: Dict[str, Any], repo_mountpoint: str) -> Tuple[List[Repository], Repository]:
        sources = []
        if os.path.isdir(repo_mountpoint + '/dists') and os.path.isdir(repo_mountpoint + '/pool') and len(
                os.listdir(repo_mountpoint + '/dists')) > 0 and len(os.listdir(repo_mountpoint + '/pool')) > 0:
            debootstrap = Repository('file://' + repo_mountpoint)
        elif 'repos' in config:
            sources, debootstrap = cls.get_sources_from_config(config)
        else:
            sources, debootstrap = cls.get_sources_from_sourceslist()
        return sources, debootstrap

    @classmethod
    def create_disk_setting(cls, config: Dict[str, Any]) -> None:
        sources = []
        repo_source = config.get('repo_source', 'auto')
        repo_mountpoint = config.get('repo_mountpoint', '/run/live/medium')
        target_path = config['disk']['target_path']
        if repo_source == 'auto':
            sources, debootstrap = cls.find_sources(config, repo_mountpoint)
        elif repo_source == 'disk':
            debootstrap = Repository('file://' + repo_mountpoint)
        elif repo_source == 'config':
            sources, debootstrap = cls.get_sources_from_config(config)
        elif repo_source == 'sources.list':
            sources, debootstrap = cls.get_sources_from_sourceslist()
        else:
            raise Exception('Wrong repo_source value. Allowed values: auto | disk | config | sources.list')
        use_eatmydata = config.get('use_eatmydata', False)
        late_command = config.get('late_command', '')
        late_command_intarget = config.get('late_command_intarget', '')
        SettingsManager.disk = DiskSetting(target_path=target_path,
                                           sources=sources,
                                           debootstrap=debootstrap,
                                           use_eatmydata=use_eatmydata,
                                           late_command=late_command,
                                           late_command_intarget=late_command_intarget)

    @classmethod
    def create_storage(cls, config: Dict[str, Any]) -> None:
        storage_layout_templates = config['disk_layout_templates']
        SettingsManager.storage = Storage(storage_layout_templates)

    @classmethod
    def create_datetime_settings(cls) -> None:
        SettingsManager.datetime = DateTimeSetting()

    @classmethod
    def log_selected_settings(cls) -> None:
        settings = [SettingsManager.systeminfo,
                    SettingsManager.installer_locale,
                    SettingsManager.system_locale,
                    SettingsManager.layout,
                    SettingsManager.timezone,
                    SettingsManager.user,
                    SettingsManager.hostname,
                    SettingsManager.kernel,
                    SettingsManager.components,
                    SettingsManager.mode,
                    SettingsManager.security]
        for setting in settings:
            setting.log_selected()

    @classmethod
    def dump_selected_settings(cls) -> Dict[str, Any]:
        selected_settings = {}
        settings = [SettingsManager.system_locale,
                    SettingsManager.layout,
                    SettingsManager.timezone,
                    SettingsManager.user,
                    SettingsManager.hostname,
                    SettingsManager.kernel,
                    SettingsManager.components,
                    SettingsManager.mode,
                    SettingsManager.security,
                    SettingsManager.disk]
        for setting in settings:
            selected_settings.update(setting.dump_selected())
        selected_settings.update(cls.dump_storage())
        return selected_settings

    @classmethod
    def dump_storage(cls) -> Dict[str, Any]:
        kickstart = SettingsManager.storage.current_template.storage_state.partition_module.generate_kickstart()
        return {'storage': {'config': kickstart}}
