import re
import traceback
from typing import Dict, Any

from astra_installer.commands import has_valid_repository
from astra_installer.commands import readfile
from astra_installer.commands import run_command
from astra_installer.logs.logger import Logger
from astra_installer.settings.disk import DiskSetting
from astra_installer.settings.layout import XKBData
from astra_installer.settings.manager import SettingsManager
from astra_installer.settings.system_locale import Locale
from astra_installer.settings.system_locale import SystemLocaleSetting
from astra_installer.utils import Utils, User, Group


class OEMSettingsManager(SettingsManager):
    @classmethod
    def create_settings(cls, config: Dict[str, Any]) -> None:
        cls.create_systeminfo_setting(config)
        cls.create_installer_locale_setting(config)
        cls.create_system_locale_setting(config)
        cls.create_layout_setting(config)
        cls.create_timezone_setting(config)
        cls.create_user_setting(config)
        cls.create_hostname_setting(config)
        cls.create_kernel_setting(config)
        cls.create_components_setting(config)
        cls.create_security_setting(config)
        cls.create_mode_setting(config)
        cls.create_disk_setting(config)
        cls.create_datetime_settings()

    @classmethod
    def create_systeminfo_setting(cls, config: Dict[str, Any]) -> None:
        super().create_systeminfo_setting(config)
        SettingsManager.systeminfo.is_kiosk_session = Utils.is_kiosk_session()
        SettingsManager.systeminfo.has_valid_repository = has_valid_repository()
        SettingsManager.systeminfo.installedpkgs = Utils.installedpkgs()

    @classmethod
    def create_installer_locale_setting(cls, config: Dict[str, Any]) -> None:
        super().create_installer_locale_setting(config)
        try:
            lang = Utils.get_lang()
            if lang:
                locale = Locale.from_string(lang)
                SettingsManager.installer_locale.default = locale.language_and_territory()
        except Exception:
            Logger.log(traceback.format_exc())

    @classmethod
    def create_hostname_setting(cls, config: Dict[str, Any]) -> None:
        super().create_hostname_setting(config)
        try:
            hostname = run_command('hostname').stdout[0]
            SettingsManager.hostname.hostname = hostname
        except Exception:
            Logger.log(traceback.format_exc())

    @classmethod
    def create_timezone_setting(cls, config: Dict[str, Any]) -> None:
        super().create_timezone_setting(config)
        try:
            timezone = readfile('/etc/timezone')[0].strip()
            SettingsManager.timezone.selected = timezone
        except Exception:
            Logger.log(traceback.format_exc())

    @classmethod
    def create_user_setting(cls, config: Dict[str, Any]) -> None:
        super().create_user_setting(config)
        username = Utils.get_current_user()
        assert username is not None, 'Can not get SUDO_USER environment variable'
        reserved_names = config['reserved_usernames']
        system_names = set(User.exising_usernames() + Group.exising_groupnames())
        # Если OEM запущен в киоске, то позволить создать
        # нового пользователя, иначе только менять пароль
        if SettingsManager.systeminfo.is_kiosk_session:
            # Имя пустое т.к. заданное имя могло бы совпасть
            # с уже существующим в системе имененем
            SettingsManager.user.username = ''
        else:
            SettingsManager.user.username = username
            system_names.remove(username)
        SettingsManager.user.username_blacklist.update(reserved_names)
        SettingsManager.user.username_blacklist.update(system_names)

    @classmethod
    def create_kernel_setting(cls, config: Dict[str, Any]) -> None:
        super().create_kernel_setting(config)
        SettingsManager.kernel.package = 'linux-image-*-generic'

    @classmethod
    def create_system_locale_setting(cls, config: Dict[str, Any]) -> None:
        super().create_system_locale_setting(config)
        try:
            result = '\n'.join(run_command('locale').stdout)
            match = re.search(r'LANG=(.*)', result)
            locale = Locale.from_string(match.group(1))
            lnt = locale.language_and_territory()
            SettingsManager.system_locale.default = lnt
            if lnt not in SettingsManager.system_locale.locales:
                SettingsManager.system_locale.locales.append(lnt)
            locales = [lnt, SystemLocaleSetting.additional_locale.language_and_territory()]
            SettingsManager.system_locale.selected = locales
        except Exception:
            Logger.log(traceback.format_exc())

    @classmethod
    def create_layout_setting(cls, config: Dict[str, Any]) -> None:
        super().create_layout_setting(config)
        try:
            keyboard = '\n'.join(readfile('/etc/default/keyboard'))
            data = XKBData.from_string(keyboard)
            SettingsManager.layout.import_xkbdata(data)
        except Exception:
            Logger.log(traceback.format_exc())

    @classmethod
    def create_mode_setting(cls, config: Dict[str, Any]) -> None:
        super().create_mode_setting(config)
        try:
            current_mode = run_command('astra-modeswitch get').stdout[0]
            # Если режим неизвестен, то полагаем, что это Орёл
            current_mode = '0' if current_mode not in {'0', '1', '2'} else current_mode
            for mode in SettingsManager.mode.modes:
                if str(mode.ident) == current_mode:
                    mode.selected = True
                else:
                    mode.selected = False
        except Exception:
            Logger.log(traceback.format_exc())

    @classmethod
    def create_disk_setting(cls, config: Dict[str, Any]) -> None:
        late_command = config.get('late_command', '')
        late_command_intarget = config.get('late_command_intarget', '')
        SettingsManager.disk = DiskSetting('/', None, None, False, late_command=late_command,
                                            late_command_intarget=late_command_intarget)

    @classmethod
    def create_datetime_settings(cls) -> None:
        super().create_datetime_settings()
