from datetime import datetime, timedelta
from typing import Callable, List, Optional


class InstallationStatus:
    def __init__(self) -> None:
        self.error = False
        self.successful_steps = []  # type: List[str]
        self.failed_steps = []  # type: List[str]

    def success(self) -> 'InstallationStatus':
        self.error = False
        return self

    def fail(self) -> 'InstallationStatus':
        self.error = True
        return self

    def __bool__(self) -> bool:
        return self.error


class ProgressManager:
    __max = 100
    __value = 0
    __next = 100
    __description = ''
    __when_started = None  # type: Optional[datetime]
    max_callback = None  # type: Callable[[int], None]
    progress_callback = None  # type: Callable[[int, str], None]
    finished_callback = None  # type: Callable[[InstallationStatus], None]

    @classmethod
    def max(cls) -> int:
        return cls.__max

    @classmethod
    def set_max(cls, value: int) -> None:
        cls.__max = value
        cls.max_callback(value)

    @classmethod
    def value(cls) -> int:
        return cls.__value

    @classmethod
    def set_value(cls, value: int) -> None:
        cls.__value = value

    @classmethod
    def next(cls) -> int:
        return cls.__next

    @classmethod
    def set_next(cls, value: int) -> None:
        cls.__next = value

    @classmethod
    def description(cls) -> str:
        return cls.__description

    @classmethod
    def set_description(cls, value: str) -> None:
        cls.__description = value

    @classmethod
    def increment(cls, description: str) -> None:
        cls.report(cls.value() + 1, description)

    @classmethod
    def change_progress(cls, value: int, next: int, description: str) -> None:
        cls.set_value(value)
        cls.set_next(next)
        cls.set_description(description)
        cls.report(value, description)

    @classmethod
    def report(cls, value: int, description: str) -> None:
        cls.progress_callback(value, description)

    @staticmethod
    def percentage(part: int, whole: int) -> float:
        return 100 * float(part) / float(whole) if whole else 0

    @staticmethod
    def percentage_value(percent: int, whole: int) -> float:
        return (percent * whole) / 100.0

    @classmethod
    def increment_by_percent_of_current_interval(cls, percent: int) -> None:
        interval = cls.next() - cls.value()
        percentage = ProgressManager.percentage_value(percent, interval)
        value = cls.value() + int(percentage)
        cls.report(value, cls.description())

    @classmethod
    def finished(cls, status: InstallationStatus) -> None:
        cls.finished_callback(status)

    @classmethod
    def start_timer(cls) -> None:
        cls.__when_started = datetime.now()

    @classmethod
    def reset_timer(cls) -> None:
        cls.__when_started = None

    @classmethod
    def get_past_time(cls) -> Optional[timedelta]:
        return datetime.now() - cls.__when_started if cls.__when_started is not None else None
