from typing import List, Dict, Any

from astra_installer.logs.logger import Logger
from astra_installer.package import Package
from astra_installer.settings.components import ComponentsSetting
from astra_installer.settings.security_feature import SecurityFeature


class SecuritySetting(ComponentsSetting):
    def __init__(self, features: List[SecurityFeature],
                 repo_available: bool = True,
                 is_kiosk_session: bool = False) -> None:
        self.features = features
        self.__repo_available = repo_available
        self.__is_kiosk_session = is_kiosk_session
        self.__loaded = False

    @property
    def components(self) -> List[SecurityFeature]:
        return self.features

    @components.setter
    def components(self, features: List[SecurityFeature]) -> None:
        self.features = features

    @property
    def exclusive_components(self) -> bool:
        return False

    def load_components(self) -> None:
        for feature in self.features:
            self.set_feature_activation_status(feature)
            SecuritySetting.disable_if_repo_unavailable(feature, self.__repo_available)
        self.__loaded = True

    @property
    def is_components_loaded(self) -> bool:
        return self.__loaded

    def set_feature_activation_status(self, feature: SecurityFeature) -> None:
        if feature.ident == 'astra-autologin-control' and self.__is_kiosk_session:
            feature.selected = False
        if feature.ident == 'astra-installer-oem-control' and self.__is_kiosk_session:
            feature.selected = False
            feature.visible = False
        elif feature.is_active():
            feature.selected = True

    @staticmethod
    def disable_if_repo_unavailable(feature: SecurityFeature, available: bool) -> None:
        """Disable security feature if package installation required and no repo available"""
        if not available:
            for package in feature.packages:
                if Package(package).get_info() is None:
                    feature.enabled = False
                    break

    @property
    def error(self) -> str:
        return ''

    @property
    def is_valid(self) -> bool:
        return True

    def selected_features(self) -> List[SecurityFeature]:
        features = []
        for feature in self.features:
            if feature.selected:
                features.append(feature)
        return features

    def log_selected(self) -> None:
        names = [str(feature) for feature in self.selected_features()]
        Logger.log('Selected security features are "{features}"'.format(features=', '.join(names)))

    def dump_selected(self) -> Dict[str, Any]:
        names = [str(feature) for feature in self.selected_features()]
        return {'features': names}
