from typing import List, TYPE_CHECKING

from astra_installer.commands import is_security_feature_enabled
from astra_installer.settings.component import Component, ComponentId

if TYPE_CHECKING:
    from gettext import gettext as _


class Executable:
    def __init__(self,
                 path: str,
                 chrooted: bool = False,
                 use_username: bool = False) -> None:
        self.path = path
        self.chrooted = chrooted
        self.use_username = use_username


class SecurityFeature(Component):
    name_by_id = {
        'astra-noautonet-control': 'Disable automatic network configuration',
        'astra-bootmenu-control': 'Disable bootloader menu show up',
        'astra-ptrace-lock': 'Disable ptrace capability',
        'astra-console-lock': 'Enable console lock',
        'astra-digsig-control': 'Enable ELF signature check',
        'astra-interpreters-lock': 'Enable interpreters lock',
        'astra-macros-lock': 'Enable macros lock',
        'astra-nochmodx-lock': 'Disable non-execution bit setup',
        'astra-secdel-swapwiper': 'Enable freeing regions cleanup on EXT-partitions',
        'astra-sudo-control': 'Enable sudo password',
        'astra-sumac-lock': 'Disable simultaneous graphics sessions',
        'astra-swapwiper-control': 'Enable swap cleanup',
        'astra-ufw-control': 'Enable ufw firewall',
        'astra-ulimits-control': 'Enable system limits',
        'astra-autologin-control': 'Enable autologin to X session',
        'astra-hardened-control': 'Use Hardened kernel by default',
        'astra-installer-grub32-control': 'Install 32-bit bootloader',
        'astra-installer-oem-control': 'OEM installation',
        'astra-mac-control': 'Mandatory access control',
        'astra-mic-control': 'Mandatory integrity control',
        'astra-installer-localtime-control': 'System clock in local time',
        'astra-nobootmenu-control': 'Disable bootloader menu',
    }

    def __init__(self, ident: ComponentId, name: str, desc: str,
                 full_desc: str, enabled: bool, selected: bool, visible: bool,
                 priority: float, dependencies: List[ComponentId],
                 executable: Executable, packages: List[str],
                 tasks: List[ComponentId]) -> None:
        super().__init__(ident, name, desc, full_desc, enabled, selected, visible,
                         priority, dependencies)
        self.executable = executable
        self.packages = packages
        self.tasks = tasks

    def is_active(self) -> bool:
        return is_security_feature_enabled(self.executable.path, self.executable.chrooted)

    @property
    def full_desc(self) -> str:
        if self.packages:
            return _('If the corresponding OS components are not selected,\nthen the '
                     'following packages with their dependencies will be installed:\n{packages}').format(
                packages=','.join(self.packages))
        else:
            return super().full_desc

    @staticmethod
    def get_name_by_id(ident: str) -> str:
        return SecurityFeature.name_by_id.get(ident, ident)
