import copy
import re
from typing import Dict, Any, List, Optional

from astra_installer.logs.logger import Logger


class Locale:
    def __init__(self, language: str, territory: str,
                 codeset: str = 'UTF-8', modifier: str = None) -> None:
        self.language = language
        self.territory = territory
        self.codeset = codeset
        self.modifier = modifier

    def language_and_territory(self) -> str:
        string = copy.copy(self.language)
        if self.territory:
            string += '_{territory}'.format(territory=self.territory)
        return string

    @staticmethod
    def from_string(localestr: str) -> Optional['Locale']:
        regex = r'^(?P<language>.*?)(_(?P<territory>.+?))?(\.(?P<codeset>.+?))?(@(?P<modifier>.+?))?$'
        m = re.match(regex, localestr)
        if m:
            return Locale(m.group('language'), m.group('territory'), m.group('codeset'), m.group('modifier'))
        else:
            return None


class SystemLocaleSetting:
    additional_locale = Locale('en', 'US')

    def __init__(self, locales: List[str], default: str,
                 country_default: Dict[str, List[str]]) -> None:
        super().__init__()
        self.locales = locales
        self.default = default
        self.country_default = country_default
        self.selected = [default]

    def country_locales(self, country: str) -> List[str]:
        return self.country_default.get(country, [self.default])

    def select_default_for_country(self, country: str) -> None:
        self.selected = self.country_locales(country)

    def selected_locales_with_codeset(self) -> List[str]:
        return ['{}.UTF-8'.format(locale) for locale in self.selected]

    def selected_locales_with_codeset_and_charset(self) -> List[str]:
        return ['{}.UTF-8 UTF-8'.format(locale) for locale in self.selected]

    def is_valid(self) -> bool:
        return self.selected is not None

    def log_selected(self) -> None:
        Logger.log('Selected system locales are "{locales}"'.format(locales=','.join(self.selected)))

    def dump_selected(self) -> Dict[str, Any]:
        return {'locales': self.selected}
