from typing import Set, Optional, List, Dict, Any

from astra_installer.logs.logger import Logger
from astra_installer.settings.components import ComponentsSetting
from astra_installer.settings.task import Task
from astra_installer.translator import deferredtr as _tr


class TasksSetting(ComponentsSetting):
    def __init__(self, tasks: List[Task],
                 repo_available: bool = True,
                 installedpkgs: Set[str] = None,
                 excluded_packages: Set[str] = None) -> None:
        self.tasks = tasks
        self.__error = None  # type: Optional[str]
        self.__loaded = False
        if not repo_available:
            self.__error = _tr('Repository unavailable')
        if installedpkgs is None:
            self.installedpkgs = set()  # type: Set[str]
        else:
            self.installedpkgs = installedpkgs
        if excluded_packages is None:
            self.exclude_packages = set()  # type: Set[str]
        else:
            self.exclude_packages = excluded_packages

    @property
    def components(self) -> List[Task]:
        return self.tasks

    @components.setter
    def components(self, tasks: List[Task]) -> None:
        self.tasks = tasks

    @property
    def exclusive_components(self) -> bool:
        return False

    def load_components(self) -> None:
        for task in self.tasks:
            self.set_task_installation_status(task)
        self.__loaded = True

    @property
    def is_components_loaded(self) -> bool:
        return self.__loaded

    def set_task_installation_status(self, task: Task) -> None:
        if task.is_installed(self.installedpkgs):
            task.enabled = False
            task.selected = True
        else:
            task.selected = False

    @property
    def error(self) -> str:
        return self.__error or ''

    @property
    def is_valid(self) -> bool:
        return not self.__error

    def enhancing_tasks(self, selected_tasks: Set['str']) -> Set['str']:
        return {task.tasksel_name for task in self.tasks if
                task.enhances and task.enhances.difference(selected_tasks) == set()}

    def selected_tasks(self) -> List['str']:
        selected_tasks = set([task.tasksel_name for task in self.tasks if task.selected])
        enhancing_tasks = self.enhancing_tasks(selected_tasks)
        return list(selected_tasks.union(enhancing_tasks))

    def log_selected(self) -> None:
        names = [str(task) for task in self.tasks if task.selected]
        Logger.log('Selected tasks are "{tasks}"'.format(tasks=', '.join(names)))

    def dump_selected(self) -> Dict[str, Any]:
        names = [str(task) for task in self.tasks if task.selected]
        return {'tasks': names}
