import traceback
from typing import TYPE_CHECKING

from astra_installer.logs.logger import Logger
from astra_installer.settings.progress_manager import ProgressManager, InstallationStatus
from astra_installer.setup.archive.disk import DiskInstallArchive
from astra_installer.setup.archive.hostname import HostnameInstallArchive
from astra_installer.setup.archive.layout import LayoutInstallArchive
from astra_installer.setup.archive.security import SecurityInstallArchive
from astra_installer.setup.archive.system_locale import LocaleInstallArchive
from astra_installer.setup.archive.tasks import TasksInstallArchive
from astra_installer.setup.archive.timezone import TimezoneInstallArchive
from astra_installer.setup.archive.user import UserInstallArchive
from astra_installer.setup.common.setup import Setup

if TYPE_CHECKING:
    from gettext import gettext as _


class SetupArchive(Setup):
    def __init__(self) -> None:
        super().__init__()
        self.disk = DiskInstallArchive()
        self.tasks = TasksInstallArchive()
        self.security = SecurityInstallArchive()
        self.timezone = TimezoneInstallArchive()
        self.hostname = HostnameInstallArchive()
        self.layout = LayoutInstallArchive()
        self.locale = LocaleInstallArchive()
        self.user = UserInstallArchive()
        self.status = InstallationStatus()

    def setup(self) -> None:
        try:
            ProgressManager.change_progress(0, 1, _('Disk partitioning'))
            self.disk.create_partitions()
            self.disk.mount_partitions()
            ProgressManager.change_progress(1, 91, _('Unpacking system archive'))
            self.disk.unpack()
            ProgressManager.change_progress(91, 92, _('Creating environment'))
            self.disk.make_chroot()
            self.tasks.install_default_answers()
            ProgressManager.change_progress(92, 94, _('Bootloader installation'))
            self.disk.setup_bootloader()
            self.disk.write_configs()
            ProgressManager.change_progress(94, 95, _('Setting timezone'))
            self.timezone.install()
            ProgressManager.change_progress(95, 96, _('Setting hostname'))
            self.hostname.install()
            ProgressManager.change_progress(96, 97, _('Setting keyboard layout'))
            self.layout.install()
            ProgressManager.change_progress(97, 98, _('Setting system locale'))
            self.locale.install()
            ProgressManager.change_progress(98, 99, _('Creating users'))
            self.user.install()
            ProgressManager.change_progress(99, 100, _('Finalizing installation'))
            self.security.apply_system_mode()
            self.disk.run_postinstall()
            self.tasks.run_hooks()
            self.security.apply_security_features()
            self.disk.finalize()
            ProgressManager.finished(self.status.success())
        except:
            Logger.log(traceback.format_exc())
            ProgressManager.finished(self.status.fail())
