import os
from shutil import copyfile, copytree
from distutils.dir_util import copy_tree

from astra_installer.commands import Chroot, run_command, run_debootstrap, safe_exec
from astra_installer.logs.logger import Logger
from astra_installer.logs.system_log_saver import SystemLogSaver
from astra_installer.settings.manager import SettingsManager
from astra_installer.utils import Utils, Grub


class DiskInstall:
    def __init__(self) -> None:
        self.chroot = Chroot(SettingsManager.disk.target_path)

    def save_logs_at_start(self) -> None:
        Logger.log('Collecting system info and logs')
        run_command('parted -sl', expect_success=False)
        SystemLogSaver.save_logs('/var/log/astra-installer/start')
        SettingsManager.log_selected_settings()

    def create_partitions(self) -> None:
        self.save_logs_at_start()

        Logger.log('Creating partitions')
        SettingsManager.storage.current_template.storage_state.apply()

    def mount_partitions(self) -> None:
        Logger.log('Mounting partitions')
        mount_data = SettingsManager.storage.current_template.storage_state.mounting_partitions()
        mount_data = sorted(mount_data,
                            key=lambda item: item.mountpoint)
        target_path = SettingsManager.disk.target_path.rstrip('/')
        run_command('mkdir -p {dir}'.format(dir=target_path))

        for data in mount_data:
            mountpoint = '{target_path}/{mountpoint}'.format(target_path=target_path,
                                                             mountpoint=data.mountpoint.lstrip('/'))
            run_command('mkdir -p {dir}'.format(dir=mountpoint))
            run_command('mount {partition_path} {dir}'.format(partition_path=data.target,
                                                              dir=mountpoint))

    def debootstrap(self) -> None:
        Logger.log('Running debootstrap')
        arguments = ['--no-check-gpg',
                     '--arch={arch}'.format(arch=SettingsManager.systeminfo.arch),
                     '--components=main,contrib,non-free',
                     SettingsManager.systeminfo.name,
                     SettingsManager.disk.target_path,
                     SettingsManager.disk.debootstrap.url]
        run_debootstrap(arguments, SettingsManager.disk.use_eatmydata)

    def mount_filesystems(self) -> None:
        Logger.log('Mounting host filesystems inside chroot')
        run_command('mount --types proc /proc {target}/proc'.format(target=SettingsManager.disk.target_path))
        run_command('mount --rbind /sys {target}/sys'.format(target=SettingsManager.disk.target_path))
        run_command('mount --make-rslave {target}/sys'.format(target=SettingsManager.disk.target_path))
        run_command('mount --rbind /dev {target}/dev'.format(target=SettingsManager.disk.target_path))
        run_command('mount --make-rslave {target}/dev'.format(target=SettingsManager.disk.target_path))

    @staticmethod
    def make_sources_entry(url: str, name: str) -> str:
        return 'deb {url} {name} main contrib non-free'.format(url=url, name=name)

    def copy_overrides(self, dirname: str) -> None:
        Logger.log('Copying files')
        overrides_dir = os.path.join(SettingsManager.systeminfo.scripts_path, dirname)
        if os.path.isdir(overrides_dir):
            copy_tree(overrides_dir, self.chroot.path('/'))

    def write_sources(self) -> None:
        Logger.log('Writing sources.list')
        # TODO: implement support for both raw sources.list strings and url-only strings
        sources = [source.url for source in SettingsManager.disk.sources]
        if sources:
            self.chroot.create_file('/etc/apt/sources.list', '\n'.join(sources))

    def add_cdrom(self) -> None:
        if SettingsManager.disk.debootstrap.is_local_and_exists:
            Logger.log('Adding cdrom')
            run_command('mkdir -p {target}/media/cdrom'.format(target=SettingsManager.disk.target_path))
            run_command('mount --bind {repo} {target}/media/cdrom'.format(repo=SettingsManager.disk.debootstrap.path,
                                                                          target=SettingsManager.disk.target_path))
            self.chroot.run_command('apt-cdrom add --no-mount')

    def prepare_apt(self) -> None:
        Logger.log('Preparing apt')
        self.chroot.apt_get_update()
        self.chroot.apt_get_dist_upgrade()
        # busybox нужен для генерации initramfs, проявляется при установке с базовыми средствами
        self.chroot.apt_get_install(['locales',
                                     'busybox',
                                     'tasksel'])

    def prepare_eatmydata(self) -> None:
        Logger.log('Preparing eatmydata')
        if SettingsManager.disk.use_eatmydata and \
                SettingsManager.systeminfo.arch == 'amd64' and not SettingsManager.systeminfo.oem:
            run_command('mkdir -p /eatmydata')
            run_command('cp /usr/lib/x86_64-linux-gnu/libeatmydata* /eatmydata')
            run_command('mkdir -p {target}/eatmydata'.format(target=SettingsManager.disk.target_path))
            run_command('mount --rbind /eatmydata {target}/eatmydata'.format(target=SettingsManager.disk.target_path))
            run_command('mount --make-rslave {target}/eatmydata'.format(target=SettingsManager.disk.target_path))
            os.environ['LD_LIBRARY_PATH'] = '/eatmydata'
            os.environ['LD_PRELOAD'] = 'libeatmydata.so'

    def make_chroot(self) -> None:
        Logger.log('Prepairing chrooted environment')
        self.mount_filesystems()
        self.copy_overrides('preinst_overrides')
        self.write_sources()
        self.prepare_eatmydata()
        self.add_cdrom()
        self.prepare_apt()

    def setup_bootloader(self) -> None:
        device = SettingsManager.storage.current_template.storage_state.bootable_device()
        Logger.log('Installing grub to {path}'.format(path=device.path))
        self.chroot.apt_get_install(['lvm2', 'cryptsetup'])  # TODO
        if SettingsManager.storage.current_template.use_efi(SettingsManager.systeminfo.efi):
            target = ''
            if SettingsManager.systeminfo.arch == 'amd64':
                self.chroot.set_selections(
                    ['grub-efi-amd64 grub-efi-amd64/install_devices multiselect {path}'.format(path=device.path)])
                self.chroot.apt_get_install(['grub-efi-amd64'])
                features = {feature.ident for feature in SettingsManager.security.selected_features()}
                if 'astra-installer-grub32-control' in features:
                    self.chroot.run_command('grub-install --bootloader-id=astralinux {path}'.format(path=device.path))
                    self.chroot.apt_get_purge(['grub-efi-amd64', 'grub-efi-amd64-bin'])
                    self.chroot.set_selections(
                        ['grub-efi-ia32 grub-efi-ia32/install_devices multiselect {path}'.format(path=device.path)])
                    self.chroot.apt_get_install(['grub-efi-ia32'])
                    target = ' --target=i386-efi'
            elif SettingsManager.systeminfo.arch == 'arm64':
                self.chroot.set_selections(
                    ['grub-efi-arm64 grub-efi-arm64/install_devices multiselect {path}'.format(path=device.path)])
                self.chroot.apt_get_install(['grub-efi-arm64'])
                features = {feature.ident for feature in SettingsManager.security.selected_features()}
            else:
                self.chroot.set_selections(
                    ['grub-efi grub-efi/install_devices multiselect {path}'.format(path=device.path)])
                self.chroot.apt_get_install(['grub-efi'])
            self.chroot.run_command('grub-install{target} --bootloader-id=astralinux {path}'.format(target=target, path=device.path))
        else:
            if SettingsManager.systeminfo.arch != "mips" and SettingsManager.systeminfo.arch != "e2k-8c":
                self.chroot.set_selections(
                    ['grub-pc grub-pc/install_devices multiselect {path}'.format(path=device.path)])
                self.chroot.apt_get_install(['grub-pc'])
                if SettingsManager.kernel.package != "linux-image-*-generic":
                    self.chroot.run_command('grub-install {path}'.format(path=device.path))
            else:
                # archive installation style
                self.chroot.run_command('grub-install {path}'.format(path=device.path))
        self.setup_grub_password()
    
    def setup_grub_password(self) -> None:
        grb = Grub(self.chroot)
        if SettingsManager.user.setup_grub_password:
            # if something wrong with gui, need to use the administrator password
            grb.set_password(SettingsManager.user.username,
                                SettingsManager.user.grub_password if SettingsManager.user.grub_password else SettingsManager.user.password,
                                SettingsManager.user.grub_password_encrypted)
        else:
            if SettingsManager.systeminfo.oem:
                grb.drop_password()

    def write_configs(self) -> None:
        Logger.log('Writing storage configs')
        SettingsManager.storage.current_template.storage_state.write_configs()

    def finalize(self) -> None:
        Logger.log('Finalizing installation')
        self.copy_overrides('postinst_overrides')
        self.write_resume()
        self.update_grub()
        self.update_initramfs()
        self.run_late_command()
        self.save_logs_at_finish()
        self.cleanup()
        self.setfsilev()
        self.umount()
        self.reboot()

    def run_postinstall(self) -> None:
        Logger.log('Running postinstall scripts')
        target = SettingsManager.disk.target_path
        adduser_script = os.path.join(SettingsManager.systeminfo.scripts_path, '94astra_adduser_tuning')
        postinst_script = os.path.join(SettingsManager.systeminfo.scripts_path, '94astra_extra_postinst')
        run_command('{script} {target}'.format(script=adduser_script, target=target))
        run_command('{script} {target} {user}'.format(script=postinst_script, target=target,
                                                      user=SettingsManager.user.username))

    def write_resume(self) -> None:
        Logger.log('Writing resume device')
        swap_devices = SettingsManager.storage.current_template.storage_state.partition_module.storage.fsset.swap_devices
        if swap_devices:
            resume_device = max(swap_devices, key=lambda x: x.size)
            resume_conf = 'RESUME={device}'.format(device=resume_device.fstab_spec)
            self.chroot.create_file('/etc/initramfs-tools/conf.d/resume', resume_conf)

    def run_late_command(self) -> None:
        Logger.log('Running late command')
        late_command_intarget = SettingsManager.disk.late_command_intarget
        safe_exec(self.chroot.run_command, late_command_intarget)
        late_command = SettingsManager.disk.late_command
        safe_exec(run_command, late_command)

    def update_grub(self) -> None:
        Logger.log('Updating grub')
        if SettingsManager.systeminfo.arch != 'mips' and SettingsManager.systeminfo.arch != 'e2k-8c':
            self.chroot.run_command('update-grub')

    def update_initramfs(self) -> None:
        Logger.log('Updating initramfs')
        self.chroot.run_command('update-initramfs -u -k all')

    def save_logs_at_finish(self) -> None:
        Logger.log('Collecting system info and logs')
        target = SettingsManager.disk.target_path
        SystemLogSaver.save_logs('/var/log/astra-installer/finish')
        copyfile(Logger.file, '{target}/var/log/astra-installer.log'.format(target=target))
        copytree('/var/log/astra-installer', '{target}/var/log/astra-installer'.format(target=target))
        safe_exec(copyfile, '/var/log/astra-installer-preseed.yaml',
                  '{target}/var/log/astra-installer-preseed.yaml'.format(target=target))

    def cleanup(self) -> None:
        Logger.log('Cleaning up installed system')
        self.chroot.run_command('apt-get clean')

    def setfsilev(self) -> None:
        Logger.log('Setup filesystem integrity')
        selected_features = {feature.ident for feature in SettingsManager.security.selected_features()}
        if 'astra-mic-control' in selected_features:
            self.chroot.run_command('astra-mic-control enable')

    def umount(self) -> None:
        Logger.log('Unmounting filesystems')
        target = SettingsManager.disk.target_path
        if SettingsManager.disk.debootstrap.is_local_and_exists:
            run_command('umount {target}/media/cdrom'.format(target=target))
        run_command('umount -R {target}'.format(target=target))

    def reboot(self) -> None:
        if SettingsManager.disk.reboot_after_install:
            Utils.reboot()
