from astra_installer.commands import Chroot, run_command
from astra_installer.logs.logger import Logger
from astra_installer.package import Package
from astra_installer.settings.manager import SettingsManager
from astra_installer.settings.security_feature import SecurityFeature


class SecurityInstall:
    def __init__(self) -> None:
        self.chroot = Chroot(SettingsManager.disk.target_path)

    def apply_security_feature(self, feature: SecurityFeature) -> None:
        action = 'enable' if feature.selected else 'disable'

        if feature.executable.chrooted:
            arguments = [action]
        else:
            arguments = [self.chroot.chroot, action]

        if feature.executable.use_username:
            arguments.append(SettingsManager.user.username)

        command = feature.executable.path + ' ' + ' '.join(arguments)

        if feature.executable.chrooted:
            self.chroot.run_command(command)
        else:
            run_command(command)

    def is_feature_dependencies_satisfied(self, feature: SecurityFeature) -> bool:
        if feature.packages:
            try:
                return all(Package(package).get_info(self.chroot) for package in feature.packages)
            except Exception:
                return False
        else:
            return True

    def apply_system_mode(self) -> None:
        Logger.log('Applying security level')
        # Вызов astra-modeswitch должен происходить всегда.
        # В текущей версии он просто выключает лишние рубильники.
        # Фичи безопасности устанавливаются вызовом соотв. скриптов (astra-mc/mic и т.д.).
        # В debian-инсталляторе выбор режима реализован так же.
        selected_mode = next(mode for mode in SettingsManager.mode.modes if mode.selected)
        astra_license = "MODE={mode}\n" \
                        "DESCRIPTION={desc}\n" \
                        "URL=https://astralinux.ru/information/licenses".format(mode=selected_mode.ident,
                                                                                desc=selected_mode.desc)
        self.chroot.create_file('/etc/astra_license', astra_license)
        if SettingsManager.systeminfo.arch != "e2k-8c":
            self.chroot.run_command("astra-modeswitch set {}".format(selected_mode.ident))

    def apply_security_features(self) -> None:
        Logger.log('Applying security features')
        for feature in SettingsManager.security.features:
            Logger.log('Processing security feature "{name}". '
                       'Selected: {selected}'.format(name=feature.name,
                                                     selected=feature.selected))
            if self.is_feature_dependencies_satisfied(feature):
                try:
                    self.apply_security_feature(feature)
                except Exception as e:
                    Logger.log('Failed to apply security feature: {name}'.format(name=feature.name))
                    Logger.log(str(e))
            else:
                Logger.log('Skipping "{name}". Dependencies are not satisfied.\n'
                           'Dependencies: {packages}'.format(name=feature.name,
                                                             packages=', '.join(feature.packages)))
