import traceback
from typing import TYPE_CHECKING

from astra_installer.logs.logger import Logger
from astra_installer.settings.progress_manager import ProgressManager, InstallationStatus
from astra_installer.setup.common.disk import DiskInstall
from astra_installer.setup.common.hostname import HostnameInstall
from astra_installer.setup.common.layout import LayoutInstall
from astra_installer.setup.common.security import SecurityInstall
from astra_installer.setup.common.system_locale import LocaleInstall
from astra_installer.setup.common.tasks import TasksInstall
from astra_installer.setup.common.timezone import TimezoneInstall
from astra_installer.setup.common.user import UserInstall

if TYPE_CHECKING:
    from gettext import gettext as _


class Setup:
    def __init__(self) -> None:
        self.disk = DiskInstall()
        self.tasks = TasksInstall()
        self.security = SecurityInstall()
        self.timezone = TimezoneInstall()
        self.hostname = HostnameInstall()
        self.layout = LayoutInstall()
        self.locale = LocaleInstall()
        self.user = UserInstall()
        self.status = InstallationStatus()

    def setup(self) -> None:
        try:
            ProgressManager.change_progress(0, 1, _('Disk partitioning'))
            self.disk.create_partitions()
            self.disk.mount_partitions()
            ProgressManager.change_progress(1, 30, _('Preparing for installation'))
            self.disk.debootstrap()
            ProgressManager.change_progress(30, 31, _('Preparing the environment'))
            self.disk.make_chroot()
            self.tasks.install_default_answers()
            ProgressManager.change_progress(31, 33, _('Bootloader installation'))
            self.disk.setup_bootloader()
            self.disk.write_configs()
            ProgressManager.change_progress(33, 34, _('Setting timezone'))
            self.timezone.install()
            ProgressManager.change_progress(34, 35, _('Setting hostname'))
            self.hostname.install()
            ProgressManager.change_progress(35, 36, _('Setting keyboard layout'))
            self.layout.install()
            ProgressManager.change_progress(36, 37, _('Setting language and regional standards'))
            self.locale.install()
            ProgressManager.change_progress(37, 98, _('Installing selected components'))
            self.tasks.install()
            ProgressManager.change_progress(98, 99, _('Creating users'))
            self.user.install()
            ProgressManager.change_progress(99, 100, _('Finalizing installation'))
            self.security.apply_system_mode()
            self.disk.run_postinstall()
            self.tasks.run_hooks()
            self.security.apply_security_features()
            self.disk.finalize()
            ProgressManager.finished(self.status.success())
        except:
            Logger.log(traceback.format_exc())
            ProgressManager.finished(self.status.fail())
