import os
from typing import List, Sequence

from astra_installer.commands import Chroot
from astra_installer.logs.logger import Logger
from astra_installer.settings.manager import SettingsManager
from astra_installer.settings.tasks import TasksSetting
from astra_installer.commands import run_command


class TasksInstall:
    def __init__(self) -> None:
        self.chroot = Chroot(SettingsManager.disk.target_path)

    def install_default_answers(self) -> None:
        Logger.log('Installing default answers')
        answers = os.path.join(SettingsManager.systeminfo.scripts_path, 'install_answers.cfg')
        with open(answers, 'r') as file:
            self.chroot.set_selections([line.rstrip('\n') for line in file.readlines()
                                        if len(line.rstrip('\n')) > 0])

    def install(self) -> None:
        tasks = SettingsManager.components.trycast(TasksSetting).selected_tasks()
        tasks.extend(self.get_additional_tasks())
        Logger.log('Installing tasks {tasks}'.format(tasks=' '.join(tasks)))
        packages = self.chroot.get_packages_from_tasks(tasks)
        packages = [package for package in packages if
                    package not in SettingsManager.components.trycast(TasksSetting).exclude_packages]
        packages.extend(self.get_additional_packages())
        if SettingsManager.kernel.package != 'linux-image-*-generic':
            kernel_packages = [SettingsManager.kernel.package]
            if 'Hardened' in tasks:
                kernel_packages.append(SettingsManager.kernel.package.replace('generic', 'hardened'))
            self.chroot.apt_get_install(kernel_packages, report_progress=False)
        self.chroot.apt_get_install(packages, report_progress=True)

    def get_additional_tasks(self) -> Sequence[str]:
        tasks = []
        for feature in SettingsManager.security.selected_features():
            if feature.tasks:
                tasks.extend(feature.tasks)
        return tasks

    def get_additional_packages(self) -> List[str]:
        packages = ['astra-safepolicy']
        if not SettingsManager.systeminfo.oem:
            for partition in SettingsManager.storage.current_template.storage_state.mountpoints.values():
                if partition.type == "mdarray":
                    if 'mdadm' not in packages:
                        packages.append('mdadm')
                    break
                for disk in partition.disks:
                    if disk.type == "mdarray":
                        if 'mdadm' not in packages:
                            packages.append('mdadm')
                        break
        if SettingsManager.kernel.package == 'linux-image-*-generic':
            if not SettingsManager.systeminfo.oem:
                if SettingsManager.systeminfo.arch != 'e2k-8c':
                    packages.append(SettingsManager.kernel.package)
        for feature in SettingsManager.security.selected_features():
            packages.extend(feature.packages)
        return packages

    def run_hooks(self) -> None:
        tasks = SettingsManager.components.trycast(TasksSetting).tasks
        for task in tasks:
            if task.selected and task.hook_path is not None:
                Logger.log('Running hook for task: {task}'.format(task=task.name))
                try:
                    run_command('{script} {target} {username}'.format(script=task.hook_path,
                                                                      target=SettingsManager.disk.target_path,
                                                                      username=SettingsManager.user.username))
                except Exception as e:
                    Logger.log('Failed hook for task: {task}'.format(task=task.name))
                    Logger.log(str(e))
