from typing import List

from astra_installer.commands import Chroot, safe_exec
from astra_installer.logs.logger import Logger
from astra_installer.settings.manager import SettingsManager
from astra_installer.utils import DesktopUtils, User


class UserInstall:
    def __init__(self) -> None:
        self.chroot = Chroot(SettingsManager.disk.target_path)

    def create_user(self) -> None:
        username = SettingsManager.user.username
        Logger.log('Creating user {username}'.format(username=username))
        self.chroot.run_command('groupadd -f -g 1000 {user}'.format(user=username))
        self.chroot.run_command('useradd -M -N -d /home/{user} '
                                '-s /bin/bash -u 1000 -g 1000 {user}'.format(user=username))

    def setup_user_password(self) -> None:
        arguments = []  # type: List[str]
        if SettingsManager.user.encrypted:
            arguments = ['--encrypted', *arguments]
        command = 'chpasswd {arguments}'.format(arguments=' '.join(arguments))
        self.chroot.run_command(command, stdin='{user}:{password}'.format(user=SettingsManager.user.username,
                                                                          password=SettingsManager.user.password))

    def create_groups(self) -> None:
        self.chroot.run_command('groupadd astra-admin')
        self.chroot.run_command('addgroup --gid 333 astra-console')

    def add_user_to_group(self, user: str, group: str) -> None:
        self.chroot.run_command('adduser {user} {group}'.format(user=user, group=group), expect_success=False)

    def add_user_to_groups(self, user: str, groups: List[str]) -> None:
        for group in groups:
            self.add_user_to_group(user, group)

    def add_user_to_default_groups(self):
        username = SettingsManager.user.username
        groups = ['astra-admin',
                  'astra-console',
                  'weston-launch',
                  'lpadmin',
                  'cdrom',
                  'floppy',
                  'audio',
                  'video',
                  'plugdev',
                  'dip',
                  'netdev',
                  'scanner']
        self.add_user_to_groups(username, groups)

    def create_user_home(self) -> None:
        self.chroot.run_command('cp -r /etc/skel /home/{user}'.format(user=SettingsManager.user.username))
        self.chroot.run_command('chown -R {user}:{user} /home/{user}'.format(user=SettingsManager.user.username))

    def enable_oem_integration(self, selected_features) -> None:
        available_features = {feature.ident for feature in SettingsManager.security.features}
        utils = DesktopUtils(self.chroot, self.chroot.run_command)
        safe_exec(utils.install_kiosk_profile, 'astra-admin')
        # Если выбора автологина нет, включаем автологин и выключаем блокировку экрана
        if 'astra-autologin-control' not in available_features:
            safe_exec(utils.enable_autologin, User(SettingsManager.user.username))
            safe_exec(utils.disable_kiosk_screenlocker)
            safe_exec(utils.disable_mac_prompt)
        else:
            # Если выбор есть и автологин выбран, выключаем блокировку экрана
            if 'astra-autologin-control' in selected_features:
                safe_exec(utils.disable_kiosk_screenlocker)
                safe_exec(utils.disable_mac_prompt)

    def install(self) -> None:
        self.create_user()
        self.setup_user_password()
        self.create_user_home()

        selected_features = {feature.ident for feature in SettingsManager.security.selected_features()}
        if 'astra-installer-oem-control' in selected_features:
            self.enable_oem_integration(selected_features)
