import os
import sys
import traceback
from shutil import copyfile
from typing import TYPE_CHECKING

from astra_installer.commands import Chroot
from astra_installer.logs.logger import Logger
from astra_installer.settings.manager import SettingsManager
from astra_installer.settings.progress_manager import ProgressManager, InstallationStatus
from astra_installer.setup.common.setup import Setup
from astra_installer.setup.e2k8c.disk import DiskInstallE2K8C
from astra_installer.setup.e2k8c.hostname import HostnameInstallE2K8C
from astra_installer.setup.e2k8c.layout import LayoutInstallE2K8C
from astra_installer.setup.e2k8c.security import SecurityInstallE2K8C
from astra_installer.setup.e2k8c.system_locale import LocaleInstallE2K8C
from astra_installer.setup.e2k8c.tasks import TasksInstallE2K8C
from astra_installer.setup.e2k8c.timezone import TimezoneInstallE2K8C
from astra_installer.setup.e2k8c.user import UserInstallE2K8C

if TYPE_CHECKING:
    from gettext import gettext as _


class SetupE2K8C(Setup):
    def __init__(self) -> None:
        super().__init__()
        self.chroot = Chroot(SettingsManager.disk.target_path)
        self.disk = DiskInstallE2K8C()
        self.tasks = TasksInstallE2K8C()
        self.security = SecurityInstallE2K8C()
        self.timezone = TimezoneInstallE2K8C()
        self.hostname = HostnameInstallE2K8C()
        self.layout = LayoutInstallE2K8C()
        self.locale = LocaleInstallE2K8C()
        self.user = UserInstallE2K8C()
        self.status = InstallationStatus()

    def fix_tmpdir(self) -> None:
        Logger.log('Fixing temporary directory')
        self.chroot.run_command('mkdir /root/tmp')

    def apply_optimization_hacks(self) -> None:
        Logger.log('Applying optimization hacks')
        self.chroot.run_command('mv /usr/sbin/update-initramfs /usr/sbin/update-initramfs.bak')
        self.chroot.run_command('ln -s /bin/true /usr/sbin/update-initramfs')

    def fix_console_setup_broken_fonts(self) -> None:
        Logger.log('Fixing broken console fonts')
        tty = os.ttyname(sys.stdout.fileno())
        if tty.startswith('/dev/tty'):
            self.chroot.run_command('mount --bind /dev/null {tty}'.format(tty=tty), expect_success=False)

    def remove_optimization_hacks(self) -> None:
        Logger.log('Removing optimization hacks')
        self.chroot.run_command('mv /usr/sbin/update-initramfs.bak /usr/sbin/update-initramfs')

    def fix_graphics_drivers(self, harware_revision: str) -> None:
        if harware_revision == '1c' and self.chroot.is_installed('libgl1:e2k-8c'):
            Logger.log('Fixing graphics settings on 1c elbrus')
            self.chroot.apt_get_install(['vivante', 'vivante-dev'])
            xorg_conf = os.path.join(SettingsManager.systeminfo.scripts_path, 'xorg.conf')
            copyfile(xorg_conf, '{target}/etc/X11/xorg.conf'.format(target=SettingsManager.disk.target_path))
            self.chroot.create_file('/etc/modules-load.d/vivante.conf', 'vivante\ngalcore')
            self.chroot.create_file('/etc/ld.so.conf.d/vivante.conf', '/usr/lib/vivante/')
            self.chroot.append_file('/etc/environment', 'QMLSCENE_DEVICE=softwarecontext')
            self.chroot.run_command('ln -sf /usr/lib/vivante/libGL.so /usr/lib/e2k-linux-gnu/libGL.so')
            self.chroot.run_command('ln -sf /usr/lib/vivante/libGL.so.1 /usr/lib/e2k-linux-gnu/libGL.so.1')
            self.chroot.run_command('ln -sf /usr/lib/vivante/libGL.so.1.2 /usr/lib/e2k-linux-gnu/libGL.so.1.2')
            self.chroot.run_command('ln -sf /usr/lib/vivante/libGL.so.1.2 /usr/lib/e2k-linux-gnu/libGL.so.1.2.0')
            self.chroot.run_command('ln -sf /usr/lib/vivante/libGL.so.1.2 /usr/lib/e2k-linux-gnu/libGL.so.1.7.0')
            self.chroot.run_command(
                'ln -sf /usr/lib/vivante/dri/vivante_dri.so /usr/lib/e2k-linux-gnu/dri/vivante_dri.so')
            self.chroot.run_command('ldconfig')
            self.chroot.run_command("sed -e \"s|^ReserveServers=.*|ReserveServers=|\" -i /etc/X11/fly-dm/fly-dmrc",
                                    expect_success=False)

    def setup(self) -> None:
        try:
            ProgressManager.change_progress(0, 1, _('Disk partitioning'))
            self.disk.create_partitions()
            self.disk.mount_partitions()
            ProgressManager.change_progress(1, 30, _('Base system installation'))
            self.disk.debootstrap()
            ProgressManager.change_progress(30, 31, _('Creating environment'))
            self.fix_tmpdir()
            self.disk.make_chroot()
            self.tasks.install_default_answers()
            ProgressManager.change_progress(31, 33, _('Bootloader installation'))
            self.disk.setup_bootloader()
            self.disk.write_configs()
            ProgressManager.change_progress(33, 34, _('Setting timezone'))
            self.timezone.install()
            ProgressManager.change_progress(34, 35, _('Setting hostname'))
            self.hostname.install()
            ProgressManager.change_progress(35, 36, _('Setting keyboard layout'))
            self.layout.install()
            ProgressManager.change_progress(36, 37, _('Setting system locale'))
            self.locale.install()
            self.apply_optimization_hacks()
            ProgressManager.change_progress(37, 98, _('Installing packages'))
            self.fix_console_setup_broken_fonts()
            self.tasks.install()
            self.fix_graphics_drivers(self.disk.hardware_revision())
            ProgressManager.change_progress(98, 99, _('Creating users'))
            self.user.install()
            ProgressManager.change_progress(99, 100, _('Finalizing installation'))
            self.security.apply_system_mode()
            self.disk.run_postinstall()
            self.tasks.run_hooks()
            self.security.apply_security_features()
            self.remove_optimization_hacks()
            self.disk.finalize()
            ProgressManager.finished(self.status.success())
        except:
            Logger.log(traceback.format_exc())
            ProgressManager.finished(self.status.fail())
