import traceback
from typing import TYPE_CHECKING

from astra_installer.logs.logger import Logger
from astra_installer.settings.progress_manager import ProgressManager, InstallationStatus
from astra_installer.setup.archive.setup import SetupArchive
from astra_installer.setup.mips.disk import DiskInstallMIPS
from astra_installer.setup.mips.hostname import HostnameInstallMIPS
from astra_installer.setup.mips.layout import LayoutInstallMIPS
from astra_installer.setup.mips.security import SecurityInstallMIPS
from astra_installer.setup.mips.system_locale import LocaleInstallMIPS
from astra_installer.setup.mips.tasks import TasksInstallMIPS
from astra_installer.setup.mips.timezone import TimezoneInstallMIPS
from astra_installer.setup.mips.user import UserInstallMIPS

if TYPE_CHECKING:
    from gettext import gettext as _


class SetupMIPS(SetupArchive):
    def __init__(self) -> None:
        super().__init__()
        self.disk = DiskInstallMIPS()
        self.tasks = TasksInstallMIPS()
        self.security = SecurityInstallMIPS()
        self.timezone = TimezoneInstallMIPS()
        self.hostname = HostnameInstallMIPS()
        self.layout = LayoutInstallMIPS()
        self.locale = LocaleInstallMIPS()
        self.user = UserInstallMIPS()
        self.status = InstallationStatus()

    def setup(self) -> None:
        try:
            ProgressManager.change_progress(0, 1, _('Disk partitioning'))
            self.disk.create_partitions()
            self.disk.mount_partitions()
            ProgressManager.change_progress(1, 91, _('Unpacking system archive'))
            self.disk.unpack()
            ProgressManager.change_progress(91, 92, _('Creating environment'))
            self.disk.make_chroot()
            self.tasks.install_default_answers()
            ProgressManager.change_progress(92, 94, _('Bootloader installation'))
            self.disk.setup_bootloader()
            self.disk.write_configs()
            ProgressManager.change_progress(94, 95, _('Setting timezone'))
            self.timezone.install()
            ProgressManager.change_progress(95, 96, _('Setting hostname'))
            self.hostname.install()
            ProgressManager.change_progress(96, 97, _('Setting keyboard layout'))
            self.layout.install()
            ProgressManager.change_progress(97, 98, _('Setting system locale'))
            self.locale.install()
            ProgressManager.change_progress(98, 99, _('Creating users'))
            self.user.install()
            ProgressManager.change_progress(99, 100, _('Finalizing installation'))
            self.security.apply_system_mode()
            self.disk.run_postinstall()
            self.tasks.run_hooks()
            self.security.apply_security_features()
            self.disk.finalize()
            ProgressManager.finished(self.status.success())
        except:
            Logger.log(traceback.format_exc())
            ProgressManager.finished(self.status.fail())
