import traceback
from typing import Callable, TYPE_CHECKING

from astra_installer.logs.logger import Logger
from astra_installer.settings.manager import SettingsManager
from astra_installer.settings.progress_manager import ProgressManager, InstallationStatus
from astra_installer.setup.common.setup import Setup
from astra_installer.setup.oem.disk import DiskInstallOEM
from astra_installer.setup.oem.hostname import HostnameInstallOEM
from astra_installer.setup.oem.layout import LayoutInstallOEM
from astra_installer.setup.oem.security import SecurityInstallOEM
from astra_installer.setup.oem.system_locale import LocaleInstallOEM
from astra_installer.setup.oem.tasks import TasksInstallOEM
from astra_installer.setup.oem.timezone import TimezoneInstallOEM
from astra_installer.setup.oem.user import UserInstallOEM

if TYPE_CHECKING:
    from gettext import gettext as _


class SetupOEM(Setup):
    def __init__(self) -> None:
        super().__init__()
        self.disk = DiskInstallOEM()
        self.tasks = TasksInstallOEM()
        self.security = SecurityInstallOEM()
        self.timezone = TimezoneInstallOEM()
        self.hostname = HostnameInstallOEM()
        self.layout = LayoutInstallOEM()
        self.locale = LocaleInstallOEM()
        self.user = UserInstallOEM()
        self.status = InstallationStatus()

    def safely_exec(self, f: Callable, description: str) -> None:
        try:
            f()
            self.status.successful_steps.append(description)
        except Exception:
            self.status.failed_steps.append(description)
            Logger.log(_('Failed to install {step}').format(step=description))
            Logger.log(traceback.format_exc())

    def setup(self) -> None:
        try:
            ProgressManager.change_progress(0, 1, _('Setup GRUB password'))
            self.disk.setup_grub_password()
            ProgressManager.change_progress(1, 2, _('Setting timezone'))
            self.safely_exec(self.timezone.install, _('timezone'))
            ProgressManager.change_progress(2, 3, _('Setting hostname'))
            self.safely_exec(self.hostname.install, _('hostname'))
            ProgressManager.change_progress(3, 4, _('Setting keyboard layout'))
            self.safely_exec(self.layout.install, _('keyboard layout'))
            ProgressManager.change_progress(4, 5, _('Setting system locale'))
            self.safely_exec(self.locale.install, _('system locale'))
            ProgressManager.change_progress(5, 6, _('Installing packages'))
            self.safely_exec(self.tasks.update, _('packages installation'))
            if SettingsManager.systeminfo.is_kiosk_session:
                ProgressManager.change_progress(6, 25, _('Installing packages'))
                self.safely_exec(self.tasks.dist_upgrade, _('packages installation'))
                ProgressManager.change_progress(25, 89, _('Installing packages'))
                self.safely_exec(self.tasks.install, _('packages installation'))
            else:
                ProgressManager.change_progress(6, 89, _('Installing packages'))
                self.safely_exec(self.tasks.install, _('packages installation'))
            ProgressManager.change_progress(89, 90, _('Tuning users'))
            self.safely_exec(self.user.install, _('users tuning'))
            ProgressManager.change_progress(90, 99, _('Tuning additional OS settings'))
            self.safely_exec(self.tasks.run_hooks, _('additional tweaks'))
            self.security.setup(self.status)
            ProgressManager.change_progress(99, 100, _('Finalizing setup'))
            self.safely_exec(self.disk.finalize, _('setup finalizing'))
            if not self.status.failed_steps:
                self.disk.reboot()
            ProgressManager.finished(self.status.success())
        except:
            Logger.log(traceback.format_exc())
            ProgressManager.finished(self.status.fail())
