import os
import pwd
import re

from astra_installer.commands import run_command, safe_exec, create_file
from astra_installer.logs.logger import Logger
from astra_installer.settings.manager import SettingsManager
from astra_installer.settings.systeminfo import AstraEdition
from astra_installer.setup.common.user import UserInstall
from astra_installer.utils import Utils, DesktopUtils, User, ConfigWriter, FilePath


class UserInstallOEM(UserInstall):
    def __init__(self) -> None:
        super().__init__()

    def create_user(self) -> None:
        username = SettingsManager.user.username
        Logger.log('Creating user {username}'.format(username=username))
        self.chroot.run_command('adduser --gecos "" --disabled-login {username}'.format(username=username))

    def lock_current_user(self) -> None:
        username = Utils.get_current_user()
        run_command('usermod --lock --expiredate 0 --shell /sbin/nologin {username}'.format(username=username))

    def hide_locked_user_in_dm(self) -> None:
        username = Utils.get_current_user()
        regex = r'^.*?HiddenUsers\s*=\s*(?P<value>[-a-z0-9_,]*)$'

        def modify_line(line: str):
            match = re.match(regex, line)
            if match:
                users = set(match.group('value').strip().split(','))
                if len(users) == 0:
                    users.update({'root'})
                users.update({username})
                value = ','.join(users)
                return 'HiddenUsers={value}'.format(value=value)
            else:
                return None

        default = 'HiddenUsers=root,{user}'.format(user=username)
        ConfigWriter(FilePath('/etc/X11/fly-dm/fly-dmrc'), self.chroot).read().modify(modify_line, default).write()

    def change_user_password(self) -> None:
        username = SettingsManager.user.username
        Logger.log('Change {username} password'.format(username=username))
        if len(SettingsManager.user.password) > 0:
            self.setup_user_password()

    def mandatize_user(self) -> None:
        if SettingsManager.systeminfo.edition == AstraEdition.SE:
            username = SettingsManager.user.username
            uid = pwd.getpwnam(username).pw_uid
            gid = pwd.getpwnam(username).pw_gid
            content = '{username}:3f'.format(username=username)
            filepath = '/etc/parsec/micdb/{uid}'.format(uid=uid)
            create_file(filepath, content)
            os.chmod(filepath, mode=0o640)
            os.chown(filepath, uid=0, gid=gid)

    def enable_oem_integration(self, selected_features) -> None:
        available_features = {feature.ident for feature in SettingsManager.security.features}
        utils = DesktopUtils(self.chroot, run_command)
        safe_exec(utils.install_kiosk_profile, 'astra-admin')
        # Если выбора автологина нет, включаем автологин и выключаем блокировку экрана
        if 'astra-autologin-control' not in available_features:
            safe_exec(utils.enable_autologin, User(SettingsManager.user.username))
            safe_exec(utils.disable_kiosk_screenlocker)
            safe_exec(utils.disable_mac_prompt)
        else:
            # Если выбор есть и автологин выбран, выключаем блокировку экрана
            if 'astra-autologin-control' in selected_features:
                safe_exec(utils.disable_kiosk_screenlocker)
                safe_exec(utils.disable_mac_prompt)

    def disable_oem_integration(self) -> None:
        available_features = {feature.ident for feature in SettingsManager.security.features}
        utils = DesktopUtils(self.chroot, run_command)
        safe_exec(utils.uninstall_kiosk_profile, 'astra-admin')
        # Если выбора автологина нет, выключаем автологин
        if 'astra-autologin-control' not in available_features:
            safe_exec(utils.disable_autologin, User(Utils.get_current_user()))
        # Оставляем поведение по умолчанию для блокировки и запроса уровня - включаем их
        safe_exec(utils.enable_kiosk_screenlocker)
        safe_exec(utils.enable_mac_prompt)

    def delete_user(self) -> None:
        username = Utils.get_current_user()
        Logger.log('Deleting user {username}'.format(username=username))
        self.chroot.run_command('userdel --force --remove {username}'.format(username=username))
        self.chroot.run_command('sed -i \'s/%3A0=.*/%3A0=/\' /var/lib/fly-dm/fly-dmsts')

    def install(self) -> None:
        selected_features = {feature.ident for feature in SettingsManager.security.selected_features()}
        if 'astra-installer-oem-control' in selected_features:
            self.enable_oem_integration(selected_features)
        else:
            self.disable_oem_integration()

        if SettingsManager.systeminfo.is_kiosk_session:
            self.create_user()
            self.setup_user_password()
            self.add_user_to_default_groups()
            self.mandatize_user()
            #self.lock_current_user()
            #safe_exec(self.hide_locked_user_in_dm)
            self.delete_user()
        else:
            self.change_user_password()
