from typing import List, Dict, Optional, Tuple, TYPE_CHECKING
from functools import cmp_to_key

from PyQt5.QtCore import QObject, pyqtSignal
from pyudev import Device, Context

from astra_installer.storage.config.config_file import ConfigFile
from astra_installer.storage.storage_state import StorageState
from astra_installer.storage.template.storage_template import StorageTemplate
from astra_installer.storage.utils.storage import partition_compare
from astra_installer_kickstart.module.partitioning_module import PartitioningModule
from astra_installer_kickstart.module.storage_module import KickstartStorageModule
from astra_installer_kickstart.report import KickstartReport, KickstartMessage

if TYPE_CHECKING:
    from gettext import gettext as _


StorageTemplatesData = List[Dict[str, str]]


class Storage(QObject):
    current_template_changed = pyqtSignal(object)  # type: Optional[StorageTemplate]
    storage_templates_changed = pyqtSignal(list)  # type: List[StorageTemplate]

    def __init__(self, templates_data: StorageTemplatesData, parent: QObject = None) -> None:
        super().__init__(parent=parent)
        self.__kickstart_storage_module = KickstartStorageModule()
        self.__storage_templates = []  # type: List[StorageTemplate]
        self.__current_template = None  # type: Optional[StorageTemplate]
        self.__all_physical_devices_cache = []  # type: List[Device]
        self.__setup(templates_data)

    @property
    def kickstart_storage_module(self) -> KickstartStorageModule:
        return self.__kickstart_storage_module

    @property
    def templates(self) -> List[StorageTemplate]:
        return self.__storage_templates

    @templates.setter
    def templates(self, templates: List[StorageTemplate]) -> None:
        self.__storage_templates = templates
        self.storage_templates_changed.emit(self.__storage_templates)

    @property
    def current_template(self) -> Optional[StorageTemplate]:
        return self.__current_template

    @current_template.setter
    def current_template(self, template: Optional[StorageTemplate]) -> None:
        self.__current_template = template
        self.current_template_changed.emit(self.__current_template)

    @staticmethod
    def create_template_text(template: StorageTemplate) -> str:
        device_path = template.usable_device.device_node if template.usable_device else None
        partition_table = 'gpt' if template.config_file.gpt_support and template.gpt_is_used else 'msdos'
        from astra_installer.settings.manager import SettingsManager
        reqpart = "reqpart" if template.use_efi(SettingsManager.systeminfo.efi) else ''
        if partition_table == 'gpt' and not reqpart:
            text = 'part biosboot --fstype=biosboot --size=2\n'
        else:
            text = ''
        text += template.config_file.source_text \
            .format(table=partition_table,
                    device=device_path)
        text += reqpart
        return text

    def config_text_state(self, text: str, passphrase: str = None) -> Tuple[KickstartReport, Optional[StorageState]]:
        report = self.__kickstart_storage_module.read_kickstart(text)
        if not report.is_valid():
            return report, None
        partition_module = self.__kickstart_storage_module.created_partitioning
        partition_module.set_passphrase(passphrase)
        task = partition_module.configure_with_task()
        try:
            task.run()
        except Exception as exception:
            task_report = KickstartReport()
            task_report.error_messages = [str(exception)]
            return task_report, None
        state = StorageState(partition_module=partition_module)
        error = state.validate(validate_efi=False)
        if error:
            task_report = KickstartReport()
            task_report.error_messages = [error]
            return task_report, None
        return report, state

    def init_template(self, template: StorageTemplate, passphrase: str = None) -> Tuple[KickstartReport,
                                                                                        Optional[StorageState]]:
        if template.config_file:
            if template.usable_device is None:
                report = KickstartReport()
                message = KickstartMessage()
                message.message = _('Storage device not specified')
                report.error_messages = [message]
                return report, None

            text = self.create_template_text(template)
            report, state = self.config_text_state(text=text, passphrase=passphrase)
            template.config_text = text
            if state:
                state.based_on_device = template.usable_device
                state.based_on_gpt = template.gpt_is_used
                state.bios_is_used = template.install_into_bios
                template.storage_state = state
            return report, state

        partition_module = PartitioningModule()
        partition_module.on_storage_changed(self.__kickstart_storage_module.storage)
        partition_module.storage.reset()
        state = StorageState(partition_module=partition_module)
        state.is_custom = True
        template.storage_state = state
        return KickstartReport(), state

    def all_physical_devices(self, update: bool = False) -> List[Device]:
        if update or not self.__all_physical_devices_cache:
            self.__kickstart_storage_module.storage.reset()
            paths = [device.path for device in self.__kickstart_storage_module.storage.disks
                     if device.type != "mdarray"]
            self.__all_physical_devices_cache = [device for device in Context().list_devices(subsystem="block")
                                                 if device.device_node in paths]

        return self.__all_physical_devices_cache

    def validate(self) -> Optional[str]:
        if not self.current_template:
            return

        return self.current_template.storage_state.validate()

    def warnings(self) -> List[str]:
        if not self.current_template:
            return []

        return self.current_template.storage_state.warnings()

    def __setup(self, templates: StorageTemplatesData) -> None:
        import blivet.partitioning
        blivet.partitioning._partition_compare_key = cmp_to_key(partition_compare)

        self.__kickstart_storage_module.storage.reset()
        self.__init_storage_templates(templates)
        if self.__storage_templates:
            self.__current_template = self.__storage_templates[0]
            devices = self.all_physical_devices()
            if devices:
                self.__current_template.usable_device = devices[0]

    def __init_storage_templates(self, templates: StorageTemplatesData) -> None:
        for template_data in templates:
            config_file = ConfigFile.from_data(template_data)
            layout = StorageTemplate(name=config_file.name, config_file=config_file)
            devices = self.all_physical_devices()
            if devices:
                layout.usable_device = devices[0]
            self.__storage_templates.append(layout)

    def getTemplateNameCount(self, name: str) -> int:
        count = 0
        for template in self.templates:
            if template.name.startswith(name):
                count += 1
        return count
