from enum import IntEnum
from typing import Optional

from PyQt5.QtCore import QObject, pyqtSignal
from pyudev import Device

from astra_installer.storage.config.config_file import ConfigFile
from astra_installer.storage.storage_state import StorageState


class TemplateType(IntEnum):
    Auto = 0
    Custom = 1


class StorageTemplate(QObject):
    usable_device_changed = pyqtSignal(object)  # type: Optional[Device]
    gpt_is_used_changed = pyqtSignal(bool)
    install_into_bios_changed = pyqtSignal(bool)
    storage_state_changed = pyqtSignal()
    config_text_changed = pyqtSignal(str)

    def __init__(self, name: str = None, config_file: ConfigFile = None, parent: QObject = None) -> None:
        super().__init__(parent=parent)
        self.__name = name or (config_file.name if config_file else None)
        self.__config_file = config_file
        self.__usable_device = None  # type: Optional[Device]
        self.__gpt_is_used = False
        self.__install_into_bios = False
        self.__type = TemplateType.Auto if config_file else TemplateType.Custom
        self.__storage_state = None  # type: Optional[StorageState]
        self.__config_text = None  # type: Optional[str]

    def __repr__(self) -> str:
        return self.name

    @property
    def name(self) -> str:
        return self.__name

    @property
    def config_file(self) -> Optional[ConfigFile]:
        return self.__config_file

    @property
    def type(self) -> TemplateType:
        return self.__type

    @property
    def usable_device(self) -> Optional[Device]:
        return self.__usable_device

    @usable_device.setter
    def usable_device(self, device: Device) -> None:
        self.__usable_device = device
        self.usable_device_changed.emit(self.__usable_device)

    @property
    def gpt_is_used(self) -> bool:
        return self.__gpt_is_used

    @gpt_is_used.setter
    def gpt_is_used(self, used: bool) -> None:
        self.__gpt_is_used = used
        self.gpt_is_used_changed.emit(self.__gpt_is_used)

    @property
    def description(self) -> Optional[str]:
        return self.__config_file.description if self.__config_file else None

    @property
    def recomended(self) -> bool:
        return self.__config_file.recomended if self.__config_file else False

    @property
    def red_book(self) -> bool:
        return self.__config_file.red_book if self.__config_file else False

    @property
    def snapshots_support(self) -> bool:
        return self.__config_file.snapshots_support if self.__config_file else False

    @property
    def is_default(self) -> bool:
        return self.__config_file.is_default if self.__config_file else False

    @property
    def storage_state(self) -> Optional[StorageState]:
        return self.__storage_state

    @storage_state.setter
    def storage_state(self, storage_state: StorageState):
        self.__storage_state = storage_state
        self.storage_state_changed.emit()

    @property
    def config_text(self) -> Optional[str]:
        return self.__config_text

    @config_text.setter
    def config_text(self, text: str) -> None:
        self.__config_text = text
        self.config_text_changed.emit(self.__config_text)

    @property
    def install_into_bios(self) -> bool:
        return self.__install_into_bios

    @install_into_bios.setter
    def install_into_bios(self, bios: bool) -> None:
        self.__install_into_bios = bios
        self.install_into_bios_changed.emit(self.__install_into_bios)

    @property
    def storage_state_based_on_actual_data(self) -> bool:
        if not self.storage_state:
            return False

        if self.storage_state.is_custom:
            return True

        return bool(self.storage_state.based_on_device == self.usable_device
                    and self.storage_state.based_on_gpt == self.gpt_is_used
                    and self.storage_state.bios_is_used == self.install_into_bios)

    def use_efi(self, efi: bool = False) -> bool:
        return (efi and not self.__install_into_bios)
