import os
from re import match
from typing import Optional, Set, List

from pyudev import Device

from blivet.util import compare

from astra_installer.commands import run_command


def system_ram_size() -> int:
    return os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')


def get_device_uuid(path: str) -> Optional[str]:
    command_result = run_command(command='blkid {path}'.format(path=path),
                                 expect_success=False)
    if command_result.is_success:
        math_result = match(r'.* UUID=\"(?P<uuid>.+?)".*', next(iter(command_result.stdout), ''))
        return math_result.group('uuid') if math_result else None
    else:
        return None


def mounted_devices() -> Set[str]:
    devices = set()  # type: Set[str]
    with open('/proc/mounts') as file:
        for line in file:
            device = line.split()[0]
            result = match(r'^(?P<device>/dev/.*?\d+)$', device)
            if result:
                devices.add(result.group('device'))
            # Find the mounted LVM-devices
            result = match(r'^(?P<device>/dev/mapper/.*?)$', device)
            if result:
                devices.add(result.group('device'))

    return devices


def sort_devices(devices: List[Device]) -> List[Device]:
    mounted_paths = mounted_devices()
    used_devices = [device for device in devices
                    if device.device_node in mounted_paths
                    or {children.device_node for children in device.children} & mounted_paths]

    used_devices.sort(key=lambda device: device.device_node)
    unused_devices = list(set(devices) ^ set(used_devices))
    unused_devices.sort(key=lambda device: device.device_node)

    return unused_devices + used_devices


def partition_compare(part1, part2):
    """ More specifically defined partitions come first.

        < 1 => x < y
          0 => x == y
        > 1 => x > y

        :param part1: the first partition
        :type part1: :class:`devices.PartitionDevice`
        :param part2: the other partition
        :type part2: :class:`devices.PartitionDevice`
        :return: see above
        :rtype: int
    """
    ret = 0

    # start sector overrides all other sorting factors
    part1_start = part1.req_start_sector
    part2_start = part2.req_start_sector
    if part1_start is not None and part2_start is None:
        return -1
    elif part1_start is None and part2_start is not None:
        return 1
    elif part1_start is not None and part2_start is not None:
        return compare(part1_start, part2_start)

    if part1.weight:
        ret -= part1.weight

    if part2.weight:
        ret += part2.weight

    if ret > 0:
        ret = 1
    elif ret < 0:
        ret = -1

    return ret
