import grp
import os
import pwd
import re
import subprocess
from argparse import ArgumentParser
from os import environ
from re import match
from shlex import quote
from signal import signal, SIGINT, SIGTSTP
from subprocess import call, CompletedProcess
from typing import List, Tuple, Optional, Dict, Set, NewType, Callable, TYPE_CHECKING

from astra_installer.commands import Chroot, run_command, readfile
from astra_installer.package import Package
from astra_installer.settings.task import Component

if TYPE_CHECKING:
    from gettext import gettext as _


class Utils:
    __test_environment_key = 'test_environment'
    __test_environment_true_value = '1'
    __test_environment_false_value = '0'

    astra_edition_regex = r'^(?P<edition>\w+)\s+(?P<release>.*?)\s+\(?(?P<name>\w+)\)?$'

    @staticmethod
    def is_test_environment() -> bool:
        if environ.get(Utils.__test_environment_key, Utils.__test_environment_false_value) \
                == Utils.__test_environment_true_value:
            return True
        else:
            return False

    @staticmethod
    def set_test_environment(test: bool = True) -> None:
        if test:
            environ[Utils.__test_environment_key] = Utils.__test_environment_true_value
        else:
            environ[Utils.__test_environment_key] = Utils.__test_environment_false_value

    @staticmethod
    def file_path(*paths) -> str:
        package_directory = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(package_directory, *paths)

    @staticmethod
    def bytes_to_gb(count_bytes: int) -> float:
        return round(count_bytes / 1024 / 1024 / 1024, 2)

    @staticmethod
    def system_ram_size() -> int:
        return os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')

    @staticmethod
    def size_with_dimension(count_bytes: int) -> Tuple[int, str]:
        dimensions = ['MiB', 'GiB', 'TiB']
        dimension_size = 1024

        value = count_bytes // (dimension_size * dimension_size)
        dimension = dimensions[0]

        for current_dimension in dimensions[1:]:
            if value % dimension_size == 0:
                value = value // dimension_size
                dimension = current_dimension
            else:
                break

        return value, dimension

    @staticmethod
    def poweroff() -> None:
        call(['systemctl', 'poweroff'])

    @staticmethod
    def reboot() -> None:
        call(['systemctl', 'reboot'])

    @staticmethod
    def override_system_signals():
        signal(SIGINT, Utils.do_nothing)
        signal(SIGTSTP, Utils.do_nothing)

    @staticmethod
    def do_nothing(*args):
        pass

    @staticmethod
    def is_console_session() -> bool:
        result = run_command('echo $TERM')
        return result.is_success and bool(match(r'linux', '\n'.join(result.stdout)))

    @staticmethod
    def is_kiosk_session() -> bool:
        return os.path.isdir('/etc/fly-kiosk/astra-admin')

    @staticmethod
    def installedpkgs() -> Set[str]:
        command = "dpkg-query -W -f='${db:Status-Abbrev}|${binary:Package}\n'"
        result = run_command(command, expect_success=False)
        if result.is_success:
            packages = set()
            for line in result.stdout:
                m = re.match(r'^ii.*?\|(?P<name>.*?)(:.*)?$', line)
                if m:
                    packages.add(m.group('name'))
            return packages
        else:
            return set()

    @staticmethod
    def get_arch_from_environment() -> str:
        arch = run_command('uname --machine', expect_success=False).stdout[0].strip()
        if arch == 'x86_64':
            return 'amd64'
        elif arch == 'aarch64':
            return 'arm64'
        elif arch == 'e2k':
            return 'e2k-8c'
        elif arch == 'mips64':
            return 'mips'
        else:
            return arch

    @staticmethod
    def value_or_auto(value, f) -> str:
        return f() if value == 'auto' else value

    @staticmethod
    def get_command_output(command: str) -> List[str]:
        process = run_command(command=command, expect_success=False)
        if process.is_fail:
            return process.stdout
        else:
            return process.stdout

    @staticmethod
    def get_current_user() -> Optional[str]:
        return os.environ.get('SUDO_USER')

    @staticmethod
    def get_sudo_uid() -> Optional[str]:
        return os.environ.get('SUDO_UID')

    @staticmethod
    def get_lang() -> Optional[str]:
        return os.environ.get('LANG', 'en_US.UTF-8')

    @staticmethod
    def check_dependencies() -> bool:
        info = Package('astra-installer-meta').get_info()
        if info is None:
            print('astra-installer-meta package not found. Please install astra-installer-meta package.')
        return info is not None

    @staticmethod
    def is_efi_supported() -> bool:
        # Are we running under with efi ?
        run_command('modprobe efivars >/dev/null 2>&1 ; /bin/true')
        return os.path.exists("/proc/efi") or os.path.exists("/sys/firmware/efi")

    @staticmethod
    def guess_file_location(possible_locations: [str], filename: str) -> Optional[str]:
        for location in possible_locations:
            path = location + '/' + filename
            if os.path.isfile(path):
                return path
        return None

    @staticmethod
    def find_preseed_path() -> Optional[str]:
        return Utils.guess_file_location(['/etc', '/lib/live/mount/medium',
                                          '/usr/lib/live/mount/medium', '/run/live/medium'],
                                         'astra-installer-preseed.yaml')

    @staticmethod
    def preseed_file_available() -> bool:
        return Utils.find_preseed_path() is not None

    @staticmethod
    def preseed_cmdline_option() -> Optional[str]:
        try:
            cmdline = readfile('/proc/cmdline')[0]
            options = cmdline.split()
            for option in options:
                match = re.match(r'astra-installer-preseed=(?P<url>.*)', option)
                if match:
                    return match.group('url')
        except:
            return None

    @staticmethod
    def locale_cmdline_option() -> Optional[str]:
        cmdline = readfile('/proc/cmdline')[0]
        options = cmdline.split()
        for option in options:
            match = re.match(r'debian-installer/locale=(?P<locale>.*)', option)
            if match:
                return match.group('locale')
        else:
            return None

    @staticmethod
    def preseed_argument() -> Optional[str]:
        parser = ArgumentParser()
        parser.add_argument('--preseed', type=str, nargs='?')
        arguments = parser.parse_args()
        return arguments.preseed

    @staticmethod
    def preseed_available() -> bool:
        parser = ArgumentParser()
        parser.add_argument('--preseed', type=str, nargs='?')
        arguments = parser.parse_args()
        return bool(arguments.preseed or Utils.preseed_file_available() or Utils.preseed_cmdline_option())

    @staticmethod
    def download_preseed(url: str) -> bool:
        cmd = "curl '{url}' -o /tmp/preseed.yaml --max-time 60".format(url=quote(url))
        return run_command(cmd).is_success

    @staticmethod
    def disable_x_screenlocker() -> None:
        if not run_command("xset dpms 0 0 0").is_success:
            raise RuntimeError("Command 'xset dpms 000' fault")
        if not run_command("xset s off").is_success:
            raise RuntimeError("Command 'xset s off' fault")


class ComponentsReader:
    @staticmethod
    def checked_components(components: List[Component]) -> List[Component]:
        return [component for component in components if component.selected]

    @staticmethod
    def showable_checked_components(components: List[Component]) -> List[Component]:
        return [component for component in ComponentsReader.checked_components(components) if component.visible]

    @staticmethod
    def checked_components_string(components: List[Component]) -> str:
        components = ComponentsReader.showable_checked_components(components)
        names = [_(component.name) for component in components]
        if len(names) > 0:
            return ', '.join(list(names))
        else:
            return _('Not selected')


FilePath = NewType('FilePath', str)
Command = NewType('Command', str)


class ConfigWriter:
    def __init__(self, file: FilePath, chroot: Chroot) -> None:
        self.file = file
        self.chroot = chroot
        self.content = []  # type: List[str]

    def read(self) -> 'ConfigWriter':
        if self.chroot.isfile(self.file):
            self.content = self.chroot.readfile(self.file)
        return self

    def modify(self, modify_line: Callable[[str], Optional[str]], default: Optional[str]) -> 'ConfigWriter':
        lines = []
        found = False
        for line in self.content:
            modified_line = modify_line(line)
            if modified_line:
                lines.append(modified_line + '\n')
                found = True
            else:
                lines.append(line)
        if not found and default:
            lines.append(default + '\n')
        self.content = lines
        return self

    def replace(self, regex: str, option: str, add_if_not_found: bool = True) -> 'ConfigWriter':
        lines = []
        found = False
        for line in self.content:
            match = re.match(regex, line)
            if match:
                lines.append(option + '\n')
                found = True
            else:
                lines.append(line)
        if not found and add_if_not_found:
            lines.append(option + '\n')
        self.content = lines
        return self

    def write(self, create_if_not_exist: bool = False) -> 'ConfigWriter':
        if self.chroot.isfile(self.file) or create_if_not_exist:
            self.chroot.create_file(self.file, ''.join(self.content).rstrip())
        return self


class User:
    def __init__(self, name: str, homedir: FilePath = '') -> None:
        self.name = name
        self.homedir = homedir or FilePath('/home/{user}'.format(user=name))

    @classmethod
    def get_homedir(cls, username: str) -> FilePath:
        return FilePath(pwd.getpwnam(username).pw_dir)

    @classmethod
    def from_username(cls, name: str) -> 'User':
        return User(name, cls.get_homedir(name))

    @classmethod
    def exising_usernames(cls) -> List[str]:
        return [entry.pw_name for entry in pwd.getpwall()]

    @classmethod
    def all(cls) -> List['User']:
        return [User(entry.pw_name, FilePath(entry.pw_dir)) for entry in pwd.getpwall()]


class Group:
    def __init__(self, name: str, users: List[User]) -> None:
        self.name = name
        self.users = users

    @classmethod
    def get_users(cls, groupname: str) -> List[User]:
        return list(map(lambda user: User.from_username(user), grp.getgrnam(groupname).gr_mem))

    @classmethod
    def from_groupname(cls, name: str) -> 'Group':
        return Group(name, cls.get_users(name))

    @classmethod
    def exising_groupnames(cls) -> List[str]:
        return [entry.gr_name for entry in grp.getgrall()]

    @classmethod
    def all(cls) -> List['Group']:
        return [Group(e.gr_name, list(map(lambda m: User.from_username(m.gr_name), e.gr_mem))) for e in grp.getgrall()]


class DesktopUtils:
    OEM_DESKTOP_FILE = FilePath('astra-installer-qt-oem.desktop')
    OEM_DESKTOP_FILE_PATH = FilePath('/usr/share/applications/astra-installer-qt-oem.desktop')
    KIOSK_PROFILE_PATH = FilePath('/usr/lib/python3/dist-packages/astra_installer_qt_oem/kiosk')
    KIOSK_AUTOSTART_SOURCE = FilePath(
        '/usr/lib/python3/dist-packages/astra_installer_qt_oem/kiosk_astra-installer-qt-oem-autostart.desktop')
    KIOSK_AUTOSTART_DEST = FilePath(
        '/etc/xdg/autostart/kiosk_astra-installer-qt-oem-autostart.desktop')
    KIOSKTHEME_PATH = '/usr/share/fly-wm/theme/default.themerc.fly-kiosk'
    OEM_DESKTOP_AUTOSTART_FILE = FilePath('kiosk_astra-installer-qt-oem-autostart.desktop')
    BIN_NAME = 'astra-installer-qt-oem'


    def __init__(self, chroot: Chroot, command_runner: Callable = None) -> None:
        self.chroot = chroot
        self.run_command = command_runner or DesktopUtils.execute

    @staticmethod
    def execute(command: Command) -> CompletedProcess:
        result = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        if result.returncode != 0:
            message = "Command '{command}' exited with code {code}".format(command=command, code=result.returncode)
            raise RuntimeError(message)
        return result

    def copy_kiosk_profile(self, name: str) -> None:
        profile = DesktopUtils.KIOSK_PROFILE_PATH
        self.run_command(Command('mkdir -p /etc/fly-kiosk'))
        cmd = Command('cp -R {profile} /etc/fly-kiosk/{name}'.format(profile=profile, name=name))
        self.run_command(cmd)

    def rm_kiosk_profile(self, name: str) -> None:
        cmd = Command('rm -R /etc/fly-kiosk/{name}'.format(name=name))
        self.run_command(cmd)

    def install_kiosk_autostart(self) -> None:
        self.run_command(Command('cp {source} {dest}'.format(source=DesktopUtils.KIOSK_AUTOSTART_SOURCE,
                                                             dest=DesktopUtils.KIOSK_AUTOSTART_DEST)))

    def remove_kiosk_autostart(self) -> None:
        self.run_command(Command('rm {dest}'.format(dest=DesktopUtils.KIOSK_AUTOSTART_DEST)))

    def rewrite_kiosk_screenlocker(self, options: Dict[str, str]) -> None:
        cw = ConfigWriter(FilePath(DesktopUtils.KIOSKTHEME_PATH), self.chroot).read()
        for option, value in options.items():
            line = '{option}={value}'.format(option=option, value=value)
            cw.replace(r'^(?!\s*;)\s*{option}\s*=.*$'.format(option=option), line)
        cw.write()

    def enable_kiosk_screenlocker(self) -> None:
        options = {'ScreenSaverDelay': '600', 'LockerOnSleep': 'true',
                   'LockerOnDPMS': 'true', 'LockerOnLid': 'true', 'LockerOnSwitch': 'true'}
        self.rewrite_kiosk_screenlocker(options)

    def disable_kiosk_screenlocker(self) -> None:
        options = {'ScreenSaverDelay': '0', 'LockerOnSleep': 'false',
                   'LockerOnDPMS': 'false', 'LockerOnLid': 'false', 'LockerOnSwitch': 'false'}
        self.rewrite_kiosk_screenlocker(options)

    def enable_mac_prompt(self) -> None:
        cw = ConfigWriter(FilePath('/etc/X11/fly-dm/fly-dmrc'), self.chroot).read()
        cw.replace(r'^.*?AutoLoginMAC\s*=\s*(?P<value>.*)$', '#AutoLoginMAC=')
        cw.write()

    def disable_mac_prompt(self) -> None:
        cw = ConfigWriter(FilePath('/etc/X11/fly-dm/fly-dmrc'), self.chroot).read()
        cw.replace(r'^.*?AutoLoginMAC\s*=\s*(?P<value>.*)$', 'AutoLoginMAC=0:63:0x0:0x0!:')
        cw.write()

    def install_kiosk_profile(self, name: str, hide_eula_page: bool) -> None:
        self.copy_kiosk_profile(name)
        self.install_kiosk_autostart()
        if hide_eula_page:
            file_autostart_path = '/etc/fly-kiosk/{profile_name}/autostart/{file_name}'.format(profile_name=name, file_name=DesktopUtils.OEM_DESKTOP_AUTOSTART_FILE)
            file_desktop_path = '/etc/fly-kiosk/{profile_name}/desktop/{file_name}'.format(profile_name=name, file_name=DesktopUtils.OEM_DESKTOP_FILE)
            command = 'sed -i \'s/{bin_name}/{bin_name} --hide_eula_page/\' {file}'
            self.run_command(Command(command.format(bin_name=DesktopUtils.BIN_NAME, file=file_autostart_path)))
            self.run_command(Command(command.format(bin_name=DesktopUtils.BIN_NAME, file=file_desktop_path)))
            self.run_command(Command(command.format(bin_name=DesktopUtils.BIN_NAME, file=DesktopUtils.KIOSK_AUTOSTART_DEST)))

    def uninstall_kiosk_profile(self, name: str) -> None:
        self.rm_kiosk_profile(name)
        self.remove_kiosk_autostart()

    def enable_autologin(self, user: User) -> None:
        cmd = Command('astra-autologin-control enable {user}'.format(user=user.name))
        self.run_command(cmd)

    def disable_autologin(self, user: User) -> None:
        cmd = Command('astra-autologin-control disable {user}'.format(user=user.name))
        self.run_command(cmd)


class Grub:
    PASSWORD_FILE = FilePath('/etc/grub.d/07_password')

    def __init__(self, chroot: Chroot) -> None:
        self.chroot = chroot

    @staticmethod
    def encrypt_password(password: str) -> str:
        if not password:
            raise RuntimeError("Cannot encrypt empty password")

        passwords = "%s\n%s\n" % (password, password)
        result = run_command("grub-mkpasswd-pbkdf2", stdin=passwords)
        encrypted_password = result.stdout[-1].split()[-1].strip()

        if not encrypted_password.startswith("grub.pbkdf2."):
            raise RuntimeError("Failed to encrypt boot loader password")

        return encrypted_password

    def set_password(self, user: str, password: str, is_encrypted: bool = False) -> None:
        if not user:
            raise RuntimeError("Cannot set empty GRUB user")

        encrypted_password = Grub.encrypt_password(password) if not is_encrypted else password
        text = "cat << EOF\nset superusers=\"%s\"\npassword_pbkdf2 %s %s\nEOF\n" % (user, user, encrypted_password)
        self.chroot.create_file(Grub.PASSWORD_FILE, text)

    def drop_password(self) -> None:
        self.chroot.create_file(Grub.PASSWORD_FILE, '')
