from typing import Callable, Optional, TYPE_CHECKING

from npyscreen import NPSAppManaged, setTheme, wgmultiline, ActionFormV2, MiniButton, Pager
from npyscreen.wgwidget import Widget

from astra_installer.commands import run_command
from astra_installer.settings.config import ConfigLoader
from astra_installer.settings.manager import SettingsManager
from astra_installer.tui.overrides_fixes_methods import WidgetOverride, MiniButtonOverride, PagerOverride
from astra_installer.tui.theme import DefaultTheme
from astra_installer.utils import Utils
from astra_installer_console.ui.installation_form import InstallationForm
from astra_installer_console.ui.license_form import LicenseForm
from astra_installer_console.ui.settings_form import SettingsForm
from astra_installer_console.ui.storage_config_editor_form import StorageConfigEditorForm

if TYPE_CHECKING:
    from gettext import gettext as _


class ConsoleInstaller(NPSAppManaged):
    STARTING_FORM = 'license'

    def __init__(self, locale: str):
        self.locale = locale
        self.last_command = None  # type: Optional[Callable]
        super().__init__()

    def onStart(self) -> None:
        Utils.override_system_signals()
        self.setup_backend()
        self.set_core_messages_level(1)
        SettingsManager.installer_locale.change_locale(self.locale)
        self.translate_global_variables()
        self.override_base_methods()
        setTheme(DefaultTheme)
        self.addForm('license', LicenseForm)
        self.addForm('regional_settings', SettingsForm)
        self.addForm('installation', InstallationForm)
        self.addForm('partition_config', StorageConfigEditorForm)

    def switch_form(self, form_id) -> None:
        self.setNextForm(form_id)
        self.switchFormNow()

    def exit(self, command: Callable = None) -> None:
        self.last_command = command
        self.switch_form(None)

    def run_last_command(self):
        if callable(self.last_command):
            self.last_command()

    @staticmethod
    def set_core_messages_level(level: int):
        run_command('dmesg -n {level}'.format(level=level))

    @staticmethod
    def translate_global_variables():
        wgmultiline.MORE_LABEL = '- {more} -'.format(more=_('more'))
        ActionFormV2.OK_BUTTON_TEXT = _('OK')
        ActionFormV2.CANCEL_BUTTON_TEXT = _('Cancel')

    @staticmethod
    def setup_backend() -> None:
        config = ConfigLoader().load()
        SettingsManager.create_settings(config)

    @staticmethod
    def override_base_methods():
        Widget.set_up_handlers = WidgetOverride.set_up_handlers
        Widget.get_and_use_key_press = WidgetOverride.get_and_use_key_press
        MiniButton.set_up_handlers = MiniButtonOverride.set_up_handlers
        Pager.h_scroll_line_down = PagerOverride.h_scroll_line_down
