import curses
from typing import Dict, Optional, TYPE_CHECKING

from npyscreen import FormBaseNew, ButtonPress, Popup, ActionPopup, FixedText
from npyscreen.fmForm import util_viewhelp

from astra_installer.utils import Utils as BackendUtils
from astra_installer_console.ui.selection_dialog import SelectionDialog
from astra_installer_console.utils import Utils as ConsoleUtils
from astra_installer.settings.manager import SettingsManager

if TYPE_CHECKING:
    from gettext import gettext as _


class BaseMainForm(FormBaseNew):
    def __init__(self,
                 title: str = None,
                 sub_title: str = None,
                 help: str = None,
                 previous_form_id: str = None,
                 next_form_id: str = None,
                 *args, **kwargs):
        title = _('Astra Linux Installer') if title is None else kwargs.get('name', None)
        if sub_title is not None:
            title += ' - ' + sub_title
        if help is None:
            help = self._read_main_help()
        kwargs.pop('name', None)
        self.help_label_1 = None  # type: Optional[FixedText]
        self.help_label_2 = None  # type: Optional[FixedText]
        self.back_button = None  # type: Optional[ButtonPress]
        self.next_button = None  # type: Optional[ButtonPress]
        self.previous_form_id = previous_form_id
        self.next_form_id = next_form_id
        super().__init__(name=title, help=help, cycle_widgets=True, *args, **kwargs)

    def set_up_handlers(self):
        super().set_up_handlers()
        self.handlers.update({curses.KEY_F10: self.shutdown_pressed})
        self.handlers.pop('^O', None)

    def create(self) -> None:
        self.set_default_sizes()

        self.help_label_1 = self.add(FixedText,
                                     value='←, →, ↑, ↓, TAB - {navigation_text} | '
                                           'SPACE - {change_text}'
                                     .format(navigation_text=_('Navigation'),
                                             change_text=_('To change')),
                                     relx=2,
                                     rely=-3,
                                     use_max_space=True,
                                     editable=False)

        self.help_label_2 = self.add(FixedText,
                                     value='ESC - {cancel_text} | '
                                           'F10 - {shutdown_text}'.format(cancel_text=_('Cancel'),
                                                                          shutdown_text=_('Shutdown')),
                                     relx=2,
                                     rely=-2,
                                     use_max_space=True,
                                     editable=False)

        height, width = self.useable_space()

        next_button_text = _('Next')
        self.next_button = self.add(ButtonPress,
                                    name=next_button_text,
                                    when_pressed_function=self.switch_next_form,
                                    rely=-2,
                                    relx=width - len(next_button_text) - 4,
                                    use_max_space=True)

        previous_button_text = _('Back')
        self.back_button = self.add(ButtonPress,
                                    name=previous_button_text,
                                    when_pressed_function=self.switch_previous_form,
                                    rely=-2,
                                    relx=self.next_button.relx - len(previous_button_text) - 3,
                                    use_max_space=True)

        self.nextrelx = 2
        self.nextrely = 2

        self.create_widgets()

        self.swap_widget_on_end(self.help_label_1)
        self.swap_widget_on_end(self.help_label_2)
        self.swap_widget_on_end(self.back_button)
        self.swap_widget_on_end(self.next_button)

    def switch_form(self, form_id: str) -> None:
        self.parentApp.switch_form(form_id)

    def switch_next_form(self) -> None:
        self.switch_form(self.next_form_id)

    def switch_previous_form(self) -> None:
        if self.previous_form_id is None:
            self.parentApp.switchFormPrevious()
        else:
            self.switch_form(self.previous_form_id)

    def create_widgets(self) -> None:
        pass

    def sizes_for_dialogs(self) -> Dict:
        return {}

    def pre_edit_loop(self) -> None:
        self.set_default_sizes()

    def post_edit_loop(self) -> None:
        self.set_default_sizes({})

    def swap_widget_on_end(self, widget) -> None:
        temp = ''
        self._widgets__.append(temp)
        first, second = self._widgets__.index(widget), len(self._widgets__) - 1
        self._widgets__[second], self._widgets__[first] = self._widgets__[first], self._widgets__[second]
        self._widgets__.pop(self._widgets__.index(temp))

    def set_control_button_name(self, button: ButtonPress, name: str) -> None:
        button.name = button.safe_string(name)
        button.label_width = len(button.name) + 2
        if button is self.next_button:
            height, width = self.useable_space()
            self.next_button.set_relyx(y=self.next_button.rely,
                                       x=width - self.next_button.label_width - 2)
        if button is self.back_button or self.next_button:
            self.back_button.set_relyx(y=self.back_button.rely,
                                       x=self.next_button.relx - self.back_button.label_width - 1)

    def set_default_sizes(self, new_sizes: Dict = None) -> None:
        sizes = self.sizes_for_dialogs() if new_sizes is None else new_sizes
        x = sizes.get('x', None)
        y = sizes.get('y', None)
        height = sizes.get('height', None)
        width = sizes.get('width', None)

        Popup.SHOW_ATX = x if x is not None else 10
        Popup.SHOW_ATY = y if y is not None else 2
        Popup.DEFAULT_LINES = height if height is not None else 12
        Popup.DEFAULT_COLUMNS = width if width is not None else 60

        ActionPopup.SHOW_ATX = Popup.SHOW_ATX
        ActionPopup.SHOW_ATY = Popup.SHOW_ATY
        ActionPopup.DEFAULT_LINES = Popup.DEFAULT_LINES
        ActionPopup.DEFAULT_COLUMNS = Popup.DEFAULT_COLUMNS

    def shutdown_pressed(self, key) -> None:
        items = [(_('Shutdown computer'), BackendUtils.poweroff),
                 (_('Reboot computer'), BackendUtils.reboot)]

        dialog = SelectionDialog(name=_('Select action:'),
                                 items=items)
        dialog.edit()
        if dialog.value is not None:
            command = dialog.value[1]
            if callable(command):
                self.parentApp.exit(command)

    def h_display_help(self, input) -> Optional[bool]:
        if self.help is None:
            return
        if self.name:
            help_name = '{name_text} - {help_text}'.format(name_text=self.name,
                                                           help_text=_('Help'))
        else:
            help_name = None

        self._display_form_with_text(title=help_name, text=str(self.help) if isinstance(self.help, str) else None)
        self.display()

        return True

    def _display_form_with_text(self, title: str, text: Optional[str]) -> None:
        curses.flushinp()
        util_viewhelp.view_help(text, title=title, autowrap=self.WRAP_HELP)
        self.display()

    def draw_title_and_help(self) -> None:
        self._draw_title_line()

    def _draw_title_line(self) -> None:
        self._draw_title()
        self._draw_help()

    def _draw_title(self) -> None:
        try:
            if self.name:
                _title = self.name[:(self.columns - 4)]
                _title = ' ' + str(_title) + ' '
                if isinstance(_title, bytes):
                    _title = _title.decode('utf-8', 'replace')
                self.add_line(0, 1,
                              _title,
                              self.make_attributes_list(_title, curses.A_NORMAL),
                              self.columns - 4)
        except:
            pass

    def _draw_help(self) -> None:
        if self.help and self.editing:
            self._draw_help_text('{help_text}: F1'.format(help_text=_('Help')))

    def _draw_help_text(self, text: str):
        text = ' {text} '.format(text=text)
        try:
            if isinstance(text, bytes):
                text = text.decode('utf-8', 'replace')
            self.add_line(0,
                          self.curses_pad.getmaxyx()[1] - len(text) - 2,
                          text,
                          self.make_attributes_list(text, curses.A_NORMAL),
                          len(text))
        except:
            pass

    @staticmethod
    def _read_main_help() -> str:
        locale = SettingsManager.installer_locale.selected
        filename = 'main_help_{locale}.txt'.format(locale=locale)
        file = open(ConsoleUtils.file_path('resources', 'help', filename), 'r')
        return file.read()
