import curses
from copy import deepcopy
from typing import List, Callable, TYPE_CHECKING

from npyscreen import TitleText, FixedText, ActionPopup, MultiSelect, SelectOne

from astra_installer.settings.manager import SettingsManager
from astra_installer.settings.task import Component
from astra_installer.utils import ComponentsReader
from astra_installer_console.ui.base_main_form import BaseMainForm

if TYPE_CHECKING:
    from gettext import gettext as _


class ComponentMixin:
    def display_value(self, component: Component) -> str:
        return _(component.name)

    def update(self, clear=True):
        super().update()
        if self.start_display_at > 0:
            self.parent.above_text.hidden = False
        else:
            self.parent.above_text.hidden = True
        self.parent.above_text.display()


class ComponentsMultiSelect(ComponentMixin, MultiSelect):
    def h_select_toggle(self, input) -> None:
        super().h_select_toggle(input)
        self.values[self.cursor_line].selected = not self.values[self.cursor_line].selected


class ComponentsSelectOne(ComponentMixin, SelectOne):
    def h_select(self, ch) -> None:
        super().h_select(ch)
        for index, component in enumerate(self.values):
            if component.enabled and component.visible:
                component.selected = index == self.cursor_line


class ComponentsMultiSelectDialog(ActionPopup):
    widget = ComponentsMultiSelect

    def __init__(self, values: List[Component] = None, dialog_accepted_callback: Callable = None, *args, **kwargs):
        self.values = sorted(values, key=lambda component: component.priority)
        self.dialog_accepted_callback = dialog_accepted_callback
        self.checkboxes = None  # type: ComponentsMultiSelect
        self.above_text = None  # type: FixedText
        super().__init__(*args, **kwargs)

    def create(self) -> None:
        self.nextrely = 1
        self.above_text = self.add(FixedText,
                                   value='- {text} -'.format(text=_('above')),
                                   editable=False,
                                   hidden=True)
        self.above_text.color = 'CONTROL'

        self.checkboxes = self.add(self.widget,
                                   values=self.values,
                                   value=[index for index, component in enumerate(self.values)
                                          if component.selected],
                                   scroll_exit=True)

    def on_ok(self) -> None:
        self.values = self.checkboxes.values
        if callable(self.dialog_accepted_callback):
            self.dialog_accepted_callback(self.values)


class ComponentsSelectOneDialog(ComponentsMultiSelectDialog):
    widget = ComponentsSelectOne


class ComponentsCheckBoxCombo(FixedText):
    dialog = ComponentsMultiSelectDialog

    def __init__(self, *args, **kwargs):
        self.components = kwargs.get('values', None)  # type: List[Component]
        self.values_changed_callback = kwargs.get('values_changed_callback', None)  # type: Callable
        kwargs.pop('values_changed_callback', None)
        self.hidden_components = None  # type: List[Component]
        super().__init__(*args, **kwargs)

    def set_up_handlers(self) -> None:
        super().set_up_handlers()
        self.handlers.update({curses.ascii.SP: self.h_change_value,
                              curses.ascii.NL: self.h_change_value})

    def h_change_value(self, input) -> None:
        component_to_show = []
        self.hidden_components = []
        for component in self.components:
            if component.visible and component.enabled:
                component_to_show.append(component)
            else:
                self.hidden_components.append(component)
        form = self.dialog(name=self.name,
                           values=deepcopy(component_to_show),
                           dialog_accepted_callback=self.dialog_accepted)
        form.edit()

    def dialog_accepted(self, components: List[Component]) -> None:
        self.components = components + self.hidden_components
        if callable(self.values_changed_callback):
            self.values_changed_callback(self.components)


class ComponentsRadioButtonCombo(ComponentsCheckBoxCombo):
    dialog = ComponentsSelectOneDialog


class TitleComponentsCheckBoxCombo(TitleText):
    _entry_type = ComponentsCheckBoxCombo

    def __init__(self, *args, **kwargs):
        self.values_changed_callback = kwargs.get('values_changed_callback', None)  # type: Callable
        kwargs['values_changed_callback'] = self.values_changed
        super().__init__(*args, **kwargs)
        self.change_text(kwargs.get('values', None))

    def values_changed(self, values: List[Component]) -> None:
        self.change_text(values)
        if callable(self.values_changed_callback):
            self.values_changed_callback(values)

    def change_text(self, components: List[Component]) -> None:
        self.value = ComponentsReader.checked_components_string(components)
        max_text_len = self.width - self.text_field_begin_at - 4
        if len(self.value) > max_text_len:
            self.value = self.value[:max_text_len] + '...'


class TitleComponentsRadioButtonCombo(TitleComponentsCheckBoxCombo):
    _entry_type = ComponentsRadioButtonCombo


class ComponentsManager:
    def __init__(self, form: BaseMainForm, select_widgets_at: int, max_width: int):
        self.form = form
        self.os_backend = SettingsManager.components
        self.security_backend = SettingsManager.security
        self.setup(select_widgets_at, max_width)

    def setup(self, select_widgets_at: int, max_width: int) -> None:
        combo_class = TitleComponentsCheckBoxCombo
        if self.os_backend.exclusive_components:
            combo_class = TitleComponentsRadioButtonCombo

        self.form.add(combo_class,
                      name=_('OS components:'),
                      begin_entry_at=select_widgets_at,
                      max_width=max_width,
                      values=self.os_backend.components,
                      values_changed_callback=self.components_changed)

        self.form.add(TitleComponentsCheckBoxCombo,
                      name=_('Additional settings:'),
                      begin_entry_at=select_widgets_at,
                      max_width=max_width,
                      values=self.security_backend.components,
                      values_changed_callback=self.security_components_changed)

    def components_changed(self, components: List[Component]) -> None:
        self.os_backend.components = components

    def security_components_changed(self, components: List[Component]) -> None:
        self.security_backend.components = components
