from datetime import datetime
from typing import Optional, TYPE_CHECKING

from babel.dates import get_timezone_location
from npyscreen import notify_confirm

from astra_installer.settings.manager import SettingsManager
from astra_installer_console.ui.title_combobox_pair import TitleComboBoxPair
from astra_installer_console.ui.time_widget import TitleTimeBox
from astra_installer_console.ui.title_date_box import TitleDate

if TYPE_CHECKING:
    from gettext import gettext as _


class DatetimeManager:
    def __init__(self, form, select_widgets_at, max_width):
        self.form = form
        self.backend = SettingsManager
        self.timezone_combobox = Optional[None]  # type: TitleComboBoxPair
        self.date_combobox = Optional[None]  # type: TitleDate
        self.time_box = Optional[None]  # type: TitleTimeBox
        self.setup(select_widgets_at, max_width)

    def setup(self, select_widgets_at, max_width):
        if self.backend.installer_locale.selected:
            current_locale = self.backend.installer_locale.selected[:2].lower()
        else:
            current_locale = None
        available_timezones = self.backend.timezone.available_timezones(current_locale)
        default_timezone = self.backend.timezone.default_timezone(current_locale)

        for index, timezone in enumerate(available_timezones):
            if timezone == default_timezone:
                current_index = index
                break
        else:
            current_index = 0

        timezones_list = self.create_timezone_items(available_timezones, current_locale)
        self.timezone_combobox = self.form.add(TitleComboBoxPair,
                                               name=_('Time zone:'),
                                               begin_entry_at=select_widgets_at,
                                               max_width=max_width,
                                               value=current_index,
                                               values=timezones_list,
                                               selection_changed_callback=self.timezone_changed)

        self.date_combobox = self.form.add(TitleDate,
                                           name=_('Date:'),
                                           begin_entry_at=select_widgets_at,
                                           max_width=max_width,
                                           value=datetime.now(),
                                           use_datetime=True,
                                           selection_changed_callback=self.date_changed)

        self.time_box = self.form.add(TitleTimeBox,
                                      name=_('Time:'),
                                      show_time_now=True,
                                      auto_update_time=True,
                                      begin_entry_at=select_widgets_at,
                                      max_width=max_width,
                                      time_changed_callback=self.time_changed)

    def timezone_changed(self, value) -> None:
        self.backend.timezone.selected = value

    def date_changed(self, value):
         if self.backend.datetime.set_date(value) is False:
             self.date_combobox.value = datetime.now()
             notify_confirm(_('Date "{date}" not supported').format(date=value.strftime('%d.%m.%Y')),
                            title=_('Error changing date'),
                            editable=False)

    def time_changed(self, value):
         if self.backend.datetime.set_time(value) is False:
             self.time_box.value = datetime.now()
             notify_confirm(_('Time "{time}" not supported').format(time=value.strftime('%H:%M:%S')),
                            title=_('Error changing time'),
                            editable=False)

    def setup_backend(self):
        self.backend.timezone.selected = self.timezone_combobox.get_current_value()

    @staticmethod
    def create_timezone_items(timezones, locale=None) -> list:
        items = []
        for timezone in timezones:
            time_now = datetime.now(timezone)
            offset = time_now.strftime('%z')
            offset_str = '{}:{}'.format(offset[:-2], offset[-2:])
            timezone_str = get_timezone_location(timezone.__str__(), locale if locale is not None else 'en')
            item_str = 'UTC{offset} {timezone}'.format(offset=offset_str,
                                                       timezone=timezone_str)
            items.append((item_str, timezone.__str__()))
        return items
