from time import strftime, gmtime
from typing import List, Optional, TYPE_CHECKING

from npyscreen import BoxTitle
from npyscreen import FixedText
from npyscreen import Pager
from npyscreen import SliderPercent

from astra_installer.logs.logger import Logger
from astra_installer.settings.progress_manager import ProgressManager, InstallationStatus
from astra_installer.setup.setup_creator import SetupCreator
from astra_installer.utils import Utils
from astra_installer_console.ui.base_main_form import BaseMainForm

if TYPE_CHECKING:
    from gettext import gettext as _


class InfoBox(BoxTitle):
    _contained_widget = Pager


class InstallationForm(BaseMainForm):
    def __init__(self, *args, **kwargs):
        self.setup = SetupCreator.create()
        self.progress = None  # type: Optional[SliderPercent]
        self.progress_desc = None  # type: Optional[FixedText]
        self.infobox = None  # type: Optional[InfoBox]
        self.log = []  # type: List[str]
        super().__init__(sub_title=_('Installation'), *args, **kwargs)

    def create_widgets(self) -> None:
        height, width = self.useable_space()

        ProgressManager.max_callback = self.on_backend_max_changed
        ProgressManager.progress_callback = self.on_backend_progress_changed
        ProgressManager.finished_callback = self.on_backend_finished
        Logger.log_callback = self.on_backend_log_appended

        self.progress_desc = self.add(FixedText,
                                      value=_('Progress description...'),
                                      editable=False)

        self.progress = self.add(SliderPercent,
                                 out_of=ProgressManager.max(),
                                 accuracy=0,
                                 editable=False)

        self.nextrely += 1
        self.infobox = self.add(InfoBox,
                                name=_('Installation info'),
                                height=height - 8,
                                hidden=True,
                                contained_widget_arguments={'autowrap': True})

        self.set_control_button_name(self.next_button, _('Finish'))
        self.next_button.when_pressed_function = self.when_next_pressed

        self.next_button.hidden = True
        self.back_button.hidden = True
        self.help_label_1.hidden = True
        self.help_label_2.hidden = True

    def when_next_pressed(self) -> None:
        Utils.reboot()

    def on_backend_max_changed(self, value: int) -> None:
        self.progress.out_of = value
        super().display()

    def on_backend_progress_changed(self, value: int, description: str) -> None:
        self.progress.set_value(value)
        self.progress_desc.value = description
        super().display()

    def on_backend_finished(self, error: InstallationStatus) -> None:
        self.progress.set_value(self.progress.out_of)
        if error:
            progress_desc = _('Installation completed with error.')
            self.progress_desc.value = progress_desc
            self.log.append('')
            self.log.append(progress_desc)
            self.infobox.set_values(self.log)
            self.infobox.entry_widget.h_show_end(None)
        else:
            progress_desc = _('Installation completed successfully.')
            text = '{result_text}\n' \
                   '{installation_time_text} - {installation_time}\n' \
                   '{hint_about_reboot}' \
                .format(result_text=progress_desc,
                        installation_time_text=_('Installation time'),
                        installation_time=strftime('%H:%M:%S', gmtime(ProgressManager.get_past_time().total_seconds())),
                        hint_about_reboot=_('After pressing the finish button, the system will restarted.'))
            self.progress_desc.value = progress_desc
            self.infobox.set_values(text.splitlines())
        self.infobox.hidden = False
        self.next_button.hidden = False
        super().display()

    def on_backend_log_appended(self, message: str) -> None:
        self.log.append(message)

    def run(self) -> None:
        ProgressManager.start_timer()
        self.setup.setup()

    def pre_edit_loop(self) -> None:
        self.run()
