from typing import Dict, Optional, TYPE_CHECKING

from npyscreen import BoxTitle
from npyscreen import Checkbox
from npyscreen import Pager
from npyscreen import notify_confirm
from npyscreen import FixedText

from astra_installer.license import License
from astra_installer.settings.manager import SettingsManager
from astra_installer_console.ui.base_main_form import BaseMainForm
from astra_installer_console.ui.radiobutton import RadioButtonsDataList
from astra_installer.settings.mode import Mode

if TYPE_CHECKING:
    from gettext import gettext as _


class LicenseBox(BoxTitle):
    _contained_widget = Pager


class LicenseForm(BaseMainForm):
    def __init__(self, *args, **kwargs):
        self.nextrelx = None  # type: Optional[int]
        self.nextrely = None  # type: Optional[int]
        self.license = None  # type: Optional[LicenseBox]
        self.accept_license_checkbox = None  # type: Optional[Checkbox]
        self.mode_label = None # type: Optional[FixedText]
        self.mode_radiobuttons_list = None  # type: Optional[RadioButtonsDataList]
        super().__init__(sub_title=_('License'), next_form_id='regional_settings', *args, **kwargs)

    @property
    def selected_mode(self) -> Optional[Mode]:
        try:
            for index in self.mode_radiobuttons_list.value:
                if index == -1:
                    return None
            text, mode = self.mode_radiobuttons_list.get_selected_value()
            return mode
        except TypeError:
            return None

    def create_widgets(self) -> None:
        height, width = self.useable_space()

        license_lines = License.load(SettingsManager).splitlines()

        self.license = self.add(LicenseBox, name=_('License:'), height=height - 12,
                                contained_widget_arguments={'values': license_lines,
                                                            'autowrap': True,
                                                            'scroll_exit': True})

        is_e2k_8c = (SettingsManager.systeminfo.arch == "e2k-8c")
        selected_mode_index = -1
        maximum_mode_index = 0
        mode_pairs_data = []
        for index, mode in enumerate(SettingsManager.mode.modes):
            mode_pairs_data.append((_(mode.name), mode))
            #if mode.selected:
            #    selected_mode_index = index
            if mode.name == "Maximum":
                maximum_mode_index = index

        self.mode_label = self.add(FixedText,
                                   value=_('Select the security level provided by the license terms:'),
                                   editable=False)
        self.mode_radiobuttons_list = self.add(RadioButtonsDataList,
                                               values_changed_callback=self.selected_mode_changed,
                                               values=mode_pairs_data,
                                               value=[selected_mode_index] if not is_e2k_8c else [maximum_mode_index],
                                               max_height=3,
                                               scroll_exit=True)
        if is_e2k_8c:
            self.mode_label.hidden = True
            self.mode_radiobuttons_list.hidden = True

        self.nextrely += 1
        self.accept_license_checkbox = self.add(Checkbox, name=_('I accept the License agreement terms'))

        self.back_button.hidden = True
        self.next_button.when_pressed_function = self.when_next_pressed

        self.selected_mode_changed()

    def when_next_pressed(self) -> None:
        if self.selected_mode:
            if self.accept_license_checkbox.value:
                self.when_license_accepted()
            else:
                self.when_license_not_accepted()
        else:
            notify_confirm(_('Please select the security level first.'),
                           title=_('The security level is not selected'),
                           editable=False)

    def when_license_accepted(self) -> None:
        self.switch_next_form()

    def sizes_for_dialogs(self) -> Dict:
        height, width = self.useable_space()

        dialog_width = width // 2
        dialog_height = height // 2

        sizes = {
            'x': (width - dialog_width) // 2,
            'y': (height - dialog_height) // 2,
            'width': dialog_width,
            'height': dialog_height
        }
        return sizes

    @staticmethod
    def when_license_not_accepted() -> None:
        notify_confirm(_('Please accept the license first.'),
                       title=_('License not accepted'),
                       editable=False)

    def selected_mode_changed(self) -> None:
        if not self.selected_mode:
            return
        for mode in SettingsManager.mode.modes:
            mode.selected = True if mode.ident == self.selected_mode.ident else False
        updated = {component.ident: component for component in self.selected_mode.features}
        for component in SettingsManager.security.components:
            if component.ident in updated:
                component.enabled = updated[component.ident].enabled
                component.selected = updated[component.ident].selected
