import traceback
from typing import Tuple, Dict, Optional, TYPE_CHECKING

from npyscreen import BoxBasic, notify_confirm

from astra_installer.logs.logger import Logger
from astra_installer.settings.autoinstall_config import AutoinstallConfigSaver
from astra_installer.settings.manager import SettingsManager
from astra_installer.storage.utils.device_action import DeviceActionUtils
from astra_installer_console.ui.base_main_form import BaseMainForm
from astra_installer_console.ui.components_manager import ComponentsManager
from astra_installer_console.ui.datetime_manager import DatetimeManager
from astra_installer_console.ui.hostname_manager import HostnameManager
from astra_installer_console.ui.layout_manager import LayoutManager
from astra_installer_console.ui.question_dialog import QuestionDialog
from astra_installer_console.ui.reboot_manager import RebootManager
from astra_installer_console.ui.storage_manager import StorageManager
from astra_installer_console.ui.system_locale_manager import SystemLocaleManager
from astra_installer_console.ui.user_manager import UserManager
from astra_installer_console.ui.kernel_manager import KernelManager

if TYPE_CHECKING:
    from gettext import gettext as _


class SettingsForm(BaseMainForm):
    def __init__(self, *args, **kwargs):
        self.backend = SettingsManager
        self.layout_manager = None  # type: Optional[LayoutManager]
        self.system_locale_manager = None  # type: Optional[SystemLocaleManager]
        self.datetime_manager = None  # type: Optional[DatetimeManager]
        self.components_manager = None  # type: Optional[ComponentsManager]
        self.kernel_manager = None # type: Optional[KernelManager]
        self.disks_layout_manager = None  # type: Optional[StorageManager]
        self.hostname_manager = None  # type: Optional[HostnameManager]
        self.user_manager = None  # type: Optional[UserManager]
        self.reboot_manager = None  # type: Optional[RebootManager]
        super().__init__(sub_title=_('Settings'), next_form_id='installation', *args, **kwargs)

    def sizes_for_dialogs(self) -> Dict:
        height, width = self.useable_space()
        box_width = 70
        count_widgets_line = 21

        sizes = {
            'x': (width - box_width) // 2,
            'y': (height - count_widgets_line) // 2,
            'height': count_widgets_line,
            'width': box_width
        }

        return sizes

    def create_widgets(self) -> None:
        sizes = self.sizes_for_dialogs()
        select_widgets_at = 31
        max_widgets_width = 68
        box_width = sizes['width']

        self.nextrelx = sizes['x']
        self.nextrely = sizes['y']

        first_box = self.add(BoxBasic,
                             name=_('Regional settings'),
                             editable=False,
                             height=7,
                             width=box_width)
        self.nextrelx += 2
        self.nextrely -= 6
        self.layout_manager = LayoutManager(self, select_widgets_at, max_widgets_width)
        self.system_locale_manager = SystemLocaleManager(self, select_widgets_at, max_widgets_width)
        self.datetime_manager = DatetimeManager(self, select_widgets_at, max_widgets_width)
        self.swap_widget_on_end(first_box)

        self.nextrelx -= 2
        self.nextrely += 1
        self.add(BoxBasic,
                 name=_('Authorization settings'),
                 editable=False,
                 height=7 if SettingsManager.systeminfo.arch != "e2k-8c" else 6,
                 width=box_width)

        self.nextrelx += 2
        self.nextrely -= 6 if SettingsManager.systeminfo.arch != "e2k-8c" else 5
        self.hostname_manager = HostnameManager(self, select_widgets_at, max_widgets_width)
        self.user_manager = UserManager(self, select_widgets_at, max_widgets_width)

        self.set_control_button_name(self.next_button, _('Install'))
        self.next_button.when_pressed_function = self.when_next_pressed

        self.nextrelx -= 2
        self.nextrely += 1
        self.add(BoxBasic,
                 name=_('Other settings'),
                 editable=False,
                 height=6 if SettingsManager.kernel.package.find('*') >= 0 else 7,
                 width=box_width)

        self.nextrelx += 2
        self.nextrely -= 5 if SettingsManager.kernel.package.find('*') >= 0 else 6
        self.components_manager = ComponentsManager(self, select_widgets_at, max_widgets_width)
        if SettingsManager.kernel.package.find('*') < 0:
            self.kernel_manager = KernelManager(self, select_widgets_at, max_widgets_width)
        self.disks_layout_manager = StorageManager(self, select_widgets_at, max_widgets_width)

        self.reboot_manager = RebootManager(self, max_widgets_width)

    def when_next_pressed(self) -> None:
        valid, message = self.settings_is_valid()
        if valid:
            warnings = self.backend.storage.current_template.storage_state.warnings()
            if not SettingsManager.kernel.is_enough_ram(SettingsManager.systeminfo.ram):
                warnings.append(_('The available amount of RAM may not be enough for the correct work of the system'))
            if warnings:
                self.display_message(title=_('Warnings'), message='\n'.join(warnings))
            self.show_confirmation_dialog()
            self.save_preseed_config()
        else:
            self.display_message(title=_('Validation failed'), message=message)

    def settings_is_valid(self) -> Tuple[bool, str]:
        validation_results = [self.hostname_manager.validate(),
                              self.user_manager.validate(),
                              self.validate_disks_settings()]
        for valid, message in validation_results:
            if not valid:
                break
        else:
            valid = True
            message = str()

        return valid, message

    def validate_disks_settings(self) -> Tuple[bool, Optional[str]]:
        current_template = self.backend.storage.current_template
        if not current_template.storage_state or not current_template.storage_state_based_on_actual_data:
            report, state = SettingsManager.storage.init_template(current_template)
            if not report.is_valid():
                return False, '\n'.join([str(message) for message in report.get_messages()])
            self.backend.storage.current_template.storage_state = state

        if self.backend.storage.current_template.storage_state.partition_module.requires_passphrase():
            self.backend.storage.current_template.storage_state = None
            return False, _('Passphrase not entered for encrypted devices')

        error = self.backend.storage.validate()
        if error:
            return False, error
        else:
            return True, None

    def show_confirmation_dialog(self):
        lines = [_('Do you really want to start installing the operating system?'),
                 _('Changes will be made to disk:')]
        lines.extend([DeviceActionUtils.action_info(action) for action
                      in self.backend.storage.current_template.storage_state.actions])

        dialog = QuestionDialog(title=_('Install confirmation'),
                                text='\n'.join(lines),
                                ok_button_text=_('Install'),
                                cancel_button_text=_('Back'),
                                confirmed_callback=self.switch_next_form)
        dialog.edit()

    @staticmethod
    def display_message(title: str, message: str) -> None:
        if len(message) > 0:
            message = message[:1].upper() + message[1:]
            if message[:-1] != '.':
                message += '.'
        notify_confirm(message,
                       title=title,
                       editable=False)

    @staticmethod
    def save_preseed_config() -> None:
        try:
            config = SettingsManager.dump_selected_settings()
            AutoinstallConfigSaver.save(config)
        except Exception:
            Logger.log('Failed to log preseed config based on selected settings')
            Logger.log(traceback.print_exc())
