from curses import KEY_F2
from typing import TYPE_CHECKING

from npyscreen import notify_confirm

from astra_installer.settings.manager import SettingsManager
from astra_installer.storage.template.storage_template import StorageTemplate, TemplateType
from astra_installer.utils import Utils
from astra_installer_console.ui.text_editor_form import TextEditorForm

if TYPE_CHECKING:
    from gettext import gettext as _


class StorageConfigEditorForm(TextEditorForm):
    def __init__(self, *args, **kwargs):
        super().__init__(sub_title=_('Partition device config editor'),
                         help=self.__get_help_text(),
                         *args, **kwargs)
        self.__storage = SettingsManager.storage

    def pre_edit_loop(self) -> None:
        self.set_text(self.__storage.current_template.config_text or '')

    def ok_pressed(self) -> None:
        text = self.get_text()
        report, state = self.__storage.config_text_state(text)
        if not report.is_valid():
            notify_confirm('\n'.join([str(message) for message in report.get_messages()]),
                           title=_('Configuration error'),
                           editable=False)
            return

        warnings = state.warnings()
        if warnings:
            notify_confirm('\n'.join(warnings),
                           title=_('Configuration warnings'),
                           editable=False)

        state.is_custom = True

        if self.__storage.current_template.type == TemplateType.Custom:
            self.__storage.current_template.config_text = text
            self.__storage.current_template.storage_state = state
        else:
            count = self.__storage.getTemplateNameCount(_('Custom disk layout'))
            name = _('Custom disk layout') if count == 0 else _('Custom disk layout ({})').format(count)
            template = StorageTemplate(name=name)
            template.config_text = text
            template.storage_state = state
            template.install_into_bios = self.__storage.current_template.install_into_bios
            self.__storage.templates = self.__storage.templates + [template]
            self.__storage.current_template = template

        self.switch_previous_form()

    def set_up_handlers(self):
        super().set_up_handlers()
        self.handlers.update({KEY_F2: self.__current_layout_shortcut_pressed})

    def _draw_help(self) -> None:
        layout_text = '{layout_text}:F2'.format(layout_text=_('Current disks layout'))

        if self.help and self.editing:
            all_text = '{help_text}: F1'.format(help_text=_('Help')) + ' | ' + layout_text
        else:
            all_text = layout_text

        self._draw_help_text(all_text)

    def __current_layout_shortcut_pressed(self, __: int) -> None:
        lines = [_('parted output:') + '\n']
        lines.extend(Utils.get_command_output('parted -l'))
        lines.append(_('fdisk output:') + '\n')
        lines.extend(Utils.get_command_output('fdisk -l'))

        title = self.name + ' - ' + _('Current disks layout')

        self._display_form_with_text(title=title, text='\n'.join(lines))

    @classmethod
    def __get_help_text(cls) -> str:
        locale = SettingsManager.installer_locale.selected
        filename = 'storage_kickstart_config_help_{locale}.txt'.format(locale=locale)
        file = open(Utils.file_path('resources', 'help', filename), 'r')
        return file.read()
