import curses
from typing import Callable, Optional, Tuple, List, TYPE_CHECKING

from npyscreen import ActionPopup, FixedText, ButtonPress
from npyscreen import notify_confirm
from pyudev import Device

from astra_installer.settings.manager import SettingsManager
from astra_installer.storage.template.storage_template import StorageTemplate, TemplateType
from astra_installer.storage.utils.pyudev_utils import device_size
from astra_installer.storage.utils.storage import sort_devices
from astra_installer_console.ui.checkbox_callback import CheckBoxCallback
from astra_installer_console.ui.passphrase_dialog import PassphraseDialog
from astra_installer_console.ui.question_dialog import QuestionDialog
from astra_installer_console.ui.radiobutton import RadioButtonsDataList

if TYPE_CHECKING:
    from gettext import gettext as _


class StorageSettingsDialog(ActionPopup):
    def __init__(self,
                 current_template_changed_callback: Callable = None,
                 template_usable_device_changed_callback: Callable = None,
                 *args, **kwargs):
        self.nextrelx = None  # type: Optional[int]
        self.nextrely = None  # type: Optional[int]
        self.editw = None  # type: Optional[int]
        self.__storage = SettingsManager.storage
        self.__current_template_changed_callback = current_template_changed_callback
        self.__template_usable_device_changed_callback = template_usable_device_changed_callback
        self.__template_radio_buttons = None  # type: Optional[RadioButtonsDataList]
        self.__devices_label = None  # type: Optional[FixedText]
        self.__devices_radio_buttons = None  # type: Optional[RadioButtonsDataList]
        self.__use_gpt_checkbox = None  # type: Optional[CheckBoxCallback]
        self.__use_bios_checkbox = None  # type: Optional[CheckBoxCallback]
        self.__edit_template_button = None  # type: Optional[ButtonPress]
        super().__init__(*args, **kwargs)

    @staticmethod
    def device_string(device: Device) -> str:
        count_bytes = device_size(device)
        gb_size = round(count_bytes / 1024 / 1024 / 1024, 2)

        text = '{size} {gb_text}'.format(size=gb_size, gb_text=_('Gb'))
        model = device.properties.get('ID_MODEL')
        if model:
            text += ' {model}'.format(model=model)
        text += ' ' + device.device_node

        return text

    @property
    def current_template(self) -> Optional[StorageTemplate]:
        try:
            text, template = self.__template_radio_buttons.get_selected_value()
            return template
        except TypeError:
            return None

    @property
    def current_device(self) -> Optional[Device]:
        if not self.__devices_radio_buttons.editable:
            return None

        try:
            text, device = self.__devices_radio_buttons.get_selected_value()
            return device
        except TypeError:
            return None

    @property
    def use_gpt(self) -> bool:
        return not self.__use_gpt_checkbox.hidden and self.__use_gpt_checkbox.value

    @property
    def use_bios(self) -> bool:
        return not self.__use_bios_checkbox.hidden and self.__use_bios_checkbox.value

    def edit(self) -> None:
        self.editing = True
        self.__init_template(self.current_template)
        super().edit()

    def create_control_buttons(self):
        close_button_text = _('Close')
        self._add_button('ok_button',
                         self.__class__.OKBUTTON_TYPE,
                         close_button_text,
                         0 - self.__class__.OK_BUTTON_BR_OFFSET[0],
                         0 - self.__class__.OK_BUTTON_BR_OFFSET[1] - len(close_button_text),
                         None)

    def create(self):
        self.add(FixedText,
                 value=_('Disk partitioning config:'),
                 editable=False)

        self.__template_radio_buttons = self.add(RadioButtonsDataList,
                                                 values_changed_callback=self.__template_radio_buttons_value_changed,
                                                 max_height=3,
                                                 scroll_exit=True)

        self.nextrely += 1
        self.__devices_label = self.add(FixedText,
                                        value=_('Disk on which the system will be installed:'),
                                        editable=False)

        self.__devices_radio_buttons = self.add(RadioButtonsDataList,
                                                values_changed_callback=self.__devices_radio_buttons_value_changed,
                                                max_height=3,
                                                scroll_exit=True)

        self.nextrely += 1
        self.__use_gpt_checkbox = self.add(CheckBoxCallback,
                                           name=_('Use GPT partition table'),
                                           value=False,
                                           toggled_callback=self.__use_gpt_checkbox_toggled)

        self.__use_bios_checkbox = self.add(CheckBoxCallback,
                                            name=_('Install system into BIOS compatibility mode'),
                                            hidden=(not SettingsManager.systeminfo.bios_compat_mode_supported),
                                            value=False,
                                            toggled_callback=self.__use_bios_checkbox_toggled)

        self.nextrelx = 1
        self.nextrely = -3
        edit_config_button_text = '{hotkey} : {text}'.format(hotkey='F2',
                                                             text=_('Edit selected disk partitioning config'))
        self.__edit_template_button = self.add(ButtonPress,
                                               name=edit_config_button_text,
                                               when_pressed_function=self.__edit_template_button_pressed,
                                               editable=False,
                                               use_max_space=True)

        self.add_handlers({curses.KEY_F2: self.__edit_template_shortcut_pressed})

        self.__storage.current_template_changed.connect(self.__storage_current_template_changed)
        self.__storage.storage_templates_changed.connect(self.__storage_storage_templates_changed)

        self.__update_devices()
        self.__update_templates()

    def __update_devices(self) -> None:
        devices_data = []  # type: List[Tuple[str, Device]]
        for device in sort_devices(self.__storage.all_physical_devices()):
            devices_data.append((self.device_string(device), device))
        self.__devices_radio_buttons.values = devices_data

    def __update_templates(self) -> None:
        templates_data = []  # type: List[Tuple[str, StorageTemplate]]
        for template in self.__storage.templates:
            if template.type == TemplateType.Auto:
                data = (_('AUTO: {template_name}').format(template_name=_(template.name)), template)
            else:
                data = (_(template.name), template)
            templates_data.append(data)

        self.__template_radio_buttons.values = templates_data
        self.__storage_current_template_changed(self.__storage.current_template)

    def __storage_current_template_changed(self, template: Optional[StorageTemplate]) -> None:
        for index, data in enumerate(self.__template_radio_buttons.values):
            text, radio_button_template = data
            if radio_button_template == template:
                self.__template_radio_buttons.set_current_index(index)
                break

    def __storage_storage_templates_changed(self, _: List[StorageTemplate]) -> None:
        self.__update_templates()

    def find_cancel_button(self) -> None:
        self.editw = len(self._widgets__) - 1

    def __template_radio_buttons_value_changed(self) -> None:
        if self.current_template:
            self.__devices_radio_buttons.editable = not (self.current_template.type != TemplateType.Auto)
        else:
            self.__devices_radio_buttons.editable = True
        self.__devices_radio_buttons.set_boxes_visible(self.__devices_radio_buttons.editable)

        if self.__devices_radio_buttons.editable:
            self.__devices_label.value = _('Disk on which the system will be installed:')
        else:
            self.__devices_label.value = _('Available disks:')
        self.__devices_label.display()

        if self.current_template:
            self.__use_gpt_checkbox.hidden = not (self.current_template.config_file
                                                  and self.current_template.config_file.gpt_support)
        else:
            self.__use_gpt_checkbox.hidden = True
        self.__use_gpt_checkbox.display()

        self.__storage.current_template = self.current_template

        if not self.current_device:  # TODO
            if self.__devices_radio_buttons.editable:
                self.__devices_radio_buttons.set_current_index(0)
            else:
                self.__devices_radio_buttons.set_current_index(-1)
        else:
            self.__devices_radio_buttons_value_changed()
        self.__devices_radio_buttons.display()

    def __devices_radio_buttons_value_changed(self) -> None:
        if self.current_template:
            self.current_template.usable_device = self.current_device
            self.__init_template(self.current_template)

    def __use_gpt_checkbox_toggled(self, _: bool) -> None:
        self.current_template.gpt_is_used = self.use_gpt
        self.__init_template(template=self.current_template, force=True)

    def __use_bios_checkbox_toggled(self, _: bool) -> None:
        self.current_template.install_into_bios = self.use_bios
        self.__init_template(template=self.current_template, force=True)

    def __edit_template_button_pressed(self) -> None:
        self.__handle_edit_template_action()

    def __edit_template_shortcut_pressed(self, *key) -> None:
        self.__handle_edit_template_action()

    def __handle_edit_template_action(self) -> None:
        text, selected_template = self.__template_radio_buttons.get_selected_value()
        if selected_template.type == TemplateType.Custom:
            self.__open_config_editor()
            return

        # Uncomment if necessary only one custom template
        # for template in self.__storage.templates:
        #    if template.type == TemplateType.Custom:
        #        dialog = QuestionDialog(title=_('Custom configuration found'),
        #                                text=_('Changing this template will overwrite the manual configuration. '
        #                                       'Do you want to continue?'))
        #        dialog.edit()
        #        if not dialog.confirmed:
        #            return
        #        else:
        #            break

        self.__open_config_editor()

    def __open_config_editor(self) -> None:
        self.exit_editing()
        self.parentApp.switch_form('partition_config')

    def __init_template(self, template: StorageTemplate, force: bool = False) -> None:
        #  TODO
        if force and self.current_template.storage_state and self.current_template.storage_state.is_custom:
            return

        if not force:
            if not self.editing \
                    or self.current_template.storage_state_based_on_actual_data:
                return

        self.current_template.storage_state = None
        if not self.__devices_radio_buttons.values:
            notify_confirm(_('No storage devices found'),
                           title=_('Configuration error'),
                           editable=False)
            return

        report, state = self.__storage.init_template(template)
        if not report.is_valid():
            notify_confirm('\n'.join([str(message) for message in report.get_messages()]),
                           title=_('Template error'),
                           editable=False)
            return

        warnings = state.warnings()
        if warnings:
            notify_confirm('\n'.join(warnings),
                           title=_('Template warnings'),
                           editable=False)

        if state.partition_module.requires_passphrase():
            self.current_template.storage_state = None
            dialog = PassphraseDialog(title=_('Enter passphrase for encrypt partitions'))
            dialog.edit()
            if not dialog.confirmed:
                return

            report, state = self.__storage.init_template(template=template, passphrase=dialog.passphrase)
            if not report.is_valid():
                notify_confirm('\n'.join([str(message) for message in report.get_messages()]),
                               title=_('Template error'),
                               editable=False)
