from typing import Optional, TYPE_CHECKING

from npyscreen import TitleText
from pyudev import Device

from astra_installer.settings.manager import SettingsManager
from astra_installer.storage.template.storage_template import StorageTemplate, TemplateType
from astra_installer.storage.utils.pyudev_utils import device_size
from astra_installer_console.ui.storage_button import StorageButton

if TYPE_CHECKING:
    from gettext import gettext as _


class StorageTitle(TitleText):
    _entry_type = StorageButton

    def __init__(self, *args, **kwargs):
        self.__storage = SettingsManager.storage
        super().__init__(current_template_changed_callback=self.__current_template_changed,
                         template_usable_device_changed_callback=self.__template_usable_device_changed,
                         *args, **kwargs)
        self.__setup()

    def __setup(self) -> None:
        self.__storage.current_template_changed.connect(self.__current_template_changed)
        if self.__storage.current_template:
            self.__storage.current_template.usable_device_changed.connect(self.__template_usable_device_changed)
        self.__update_info_label()

    def __current_template_changed(self, template: Optional[StorageTemplate]) -> None:
        if template:
            template.usable_device_changed.connect(self.__template_usable_device_changed)
        self.__update_info_label()

    def __template_usable_device_changed(self, __: Optional[Device]) -> None:
        self.__update_info_label()

    def __update_info_label(self) -> None:
        if not self.__storage.current_template:
            text = _('Storage layout unset')
        elif self.__storage.current_template.usable_device:
            text = _(self.__storage.current_template.name)
            count_bytes = device_size(self.__storage.current_template.usable_device)
            gb_size = round(count_bytes / 1024 / 1024 / 1024, 2)
            device_info = '{size} {gb_text} {node}' \
                .format(size=gb_size,
                        gb_text=_('Gb'),
                        node=self.__storage.current_template.usable_device.device_node)
            text += ' - ' + device_info
        else:
            if self.__storage.current_template.type == TemplateType.Custom:
                text = self.__storage.current_template.name
            else:
                text = _('Storage device not found')

        max_text_len = self.width - self.text_field_begin_at - 4
        if len(text) > max_text_len:
            text = text[:max_text_len] + '...'

        self.value = text
