from re import match
from typing import Optional, TYPE_CHECKING

from astra_installer.commands import run_command
from astra_installer.settings.manager import SettingsManager
from astra_installer.translator import Translator
from astra_installer.tui.wait_dialog import WaitDialog
from astra_installer.tui.wait_launcher import WaitLauncher
from astra_installer.utils import Utils
from astra_installer_console.ui.title_combobox_pair import TitleComboBoxPair

if TYPE_CHECKING:
    from gettext import gettext as _


class SystemLocaleManager:
    def __init__(self, form, select_widgets_at, max_width):
        self.form = form
        self.system_locale_backend = SettingsManager.system_locale
        self.installer_locale_backend = SettingsManager.installer_locale
        self.layout_backend = SettingsManager.layout
        self.locales_combobox = Optional[None]  # type: TitleComboBoxPair
        self.wait_dialog_enabled = False  # TODO: del
        self.setup(select_widgets_at, max_width)

    def setup(self, select_widgets_at: int, max_width: int) -> None:
        current_index = 0
        items = []
        for index, locale in enumerate(self.system_locale_backend.locales):
            text = Translator.locale_name(locale, self.installer_locale_backend.selected)
            if locale != self.installer_locale_backend.selected:
                text += ' - ' + Translator.locale_name(locale, locale)
            else:
                current_index = index
            items.append((text, locale))

        self.locales_combobox = self.form.add(TitleComboBoxPair,
                                              name=_('System language:'),
                                              begin_entry_at=select_widgets_at,
                                              value=current_index,
                                              max_width=max_width,
                                              values=items,
                                              selection_changed_callback=self.locale_changed)

        self.locale_changed(self.installer_locale_backend.selected)
        self.wait_dialog_enabled = True

    def locale_changed(self, locale: str) -> None:
        launcher = WaitLauncher(time=1.0, main_task=self.change_locale, wait_task=self.show_wait_dialog)
        launcher.set_main_args(locale)
        launcher.run()

    def show_wait_dialog(self):
        if self.wait_dialog_enabled:
            WaitDialog(screen=self.form, message=_('Applying system locale...')).show()

    def change_locale(self, locale: str) -> None:
        self.setup_backend()
        if Utils.is_console_session():
            self.change_layout_locale(locale)
        else:
            self.layout_backend.apply()

    def setup_layouts(self, locale: str) -> None:
        layout = self.layout_backend.layout_from_locale(locale)
        self.layout_backend.selected_layouts = [layout]

    def setup_backend(self) -> None:
        locale = self.locales_combobox.get_current_value()
        self.system_locale_backend.selected = [locale]
        self.setup_layouts(locale)

    def change_layout_locale(self, locale: str) -> None:
        lines = []
        for line in open(self.layout_backend.keyboard_file, 'r').read().splitlines():
            if match(r'XKBLAYOUT\s*=\s*".+"', line):
                lines.append('XKBLAYOUT="us,{locale}"'.format(locale=self.layout_backend.layout_from_locale(locale)))
            else:
                lines.append(line)
        open(self.layout_backend.keyboard_file, 'w').write('\n'.join(lines))

        run_command('setupcon --save')
