import curses
from datetime import datetime
from typing import Callable, TYPE_CHECKING

from npyscreen import Textfield, FixedText, ActionPopup

from astra_installer.tui.overrides_fixes_methods import AsciiCode
from astra_installer_console.ui.title_widget import TitleWidget

if TYPE_CHECKING:
    from gettext import gettext as _


class DigitEdit(Textfield):
    def __init__(self, screen, max_value: int = 99, apply_callback: Callable = None, *args, **kwargs):
        self.max_value = max_value
        self.apply_callback = apply_callback
        self.handlers = {}
        self.complex_handlers = []
        kwargs['max_height'] = 1
        max_value_len = len(str(self.max_value))
        kwargs['max_width'] = max_value_len
        super().__init__(screen, *args, **kwargs)
        self.maximum_string_length = max_value_len

    def set_up_handlers(self) -> None:
        self.handlers = {
            curses.KEY_BTAB: self.h_exit_up,
            curses.KEY_DOWN: self.h_reduce_value,
            curses.KEY_UP: self.h_increase_value,
            curses.KEY_LEFT: self.h_exit_left,
            curses.KEY_RIGHT: self.h_exit_right,
            curses.KEY_MOUSE: self.h_exit_mouse}

        self.complex_handlers.extend(((self.t_input_isprint, self.h_addch),))

    def h_increase_value(self, key) -> None:
        self.set_value(str(int(self.value) + 1) if int(self.value) < self.max_value else str(0))

    def h_reduce_value(self, key) -> None:
        self.set_value(str(int(self.value) - 1) if int(self.value) > 0 else str(self.max_value))

    def set_value(self, value: str) -> None:
        max_len = len(str(self.max_value))
        value_len = len(value)
        if value_len < max_len:
            new_value = ''.join(['0' for count in range(max_len - value_len)]) + value
        else:
            new_value = value[:max_len]

        if int(new_value) > self.max_value:
            self.value = str(self.max_value)
        else:
            self.value = new_value

    def display(self) -> None:
        if self.hidden:
            self.clear()
            self.parent.refresh()
        else:
            self.update(cursor=False)
            self.parent.refresh()

    def h_addch(self, inp) -> None:
        if not self.editable:
            return

        try:
            ch_adding = chr(inp)
        except TypeError:
            ch_adding = input

        if ch_adding == ' ':
            if callable(self.apply_callback):
                self.apply_callback()

        try:
            ch_adding = str(int(ch_adding))
        except ValueError:
            return

        value = str(int(self.value))
        if len(value) < len(str(self.max_value)):
            self.set_value(value + ch_adding)
        else:
            self.set_value(ch_adding)

        if len(str(int(self.value))) >= len(str(self.max_value)):
            self.h_exit_right(None)


class TimeForm(ActionPopup):
    def __init__(self, value=datetime.now(), value_changed_callback: Callable = None, *args, **kwargs):
        self.value = value
        self.value_changed_callback = value_changed_callback
        self.hours = None  # type: DigitEdit
        self.minutes = None  # type: DigitEdit
        self.seconds = None  # type: DigitEdit
        super().__init__(cycle_widgets=True, *args, **kwargs)

    def create_control_buttons(self):
        pass

    def create(self) -> None:
        height, width = self.useable_space()

        x = (width - 8) // 2
        y = (height - 3) // 2

        self.hours = self.add(DigitEdit,
                              value=self.value.strftime('%H'),
                              max_value=23,
                              relx=x,
                              rely=y,
                              apply_callback=self.time_apply)

        x += 2
        self.add(FixedText, value=':', editable=False, relx=x, rely=y, use_max_space=True)
        x += 1
        self.minutes = self.add(DigitEdit,
                                value=self.value.strftime('%M'),
                                max_value=59,
                                relx=x,
                                rely=y,
                                apply_callback=self.time_apply)

        x += 2
        self.add(FixedText, value=':', editable=False, relx=x, rely=y, use_max_space=True)
        x += 1
        self.seconds = self.add(DigitEdit,
                                value=self.value.strftime('%S'),
                                max_value=59,
                                relx=x,
                                rely=y,
                                apply_callback=self.time_apply)

        self.add(FixedText, editable=False,
                 value='↑, ↓ - {text}'.format(text=_('Change value')),
                 relx=2,
                 rely=-3,
                 use_max_space=True)

        self.add(FixedText, editable=False,
                 value='<{numbers_text}> - {help_text}'.format(numbers_text=_('numbers'),
                                                               help_text=_('Enter value')),
                 relx=2,
                 rely=-2,
                 use_max_space=True)

    def set_up_handlers(self) -> None:
        super().set_up_handlers()
        self.handlers.update({curses.ascii.SP: self.time_apply,
                              curses.ascii.NL: self.time_apply,
                              curses.ascii.TAB: self.setup_exit_from_widget,
                              curses.ascii.ESC: self.setup_exit_from_widget,
                              AsciiCode.ALT_ESC: self.setup_exit_from_widget})

    def time_apply(self, key=None) -> None:
        self.setup_exit_from_widget()
        time_string = '{hours}:{minutes}:{seconds}'.format(hours=self.hours.value,
                                                           minutes=self.minutes.value,
                                                           seconds=self.seconds.value)
        self.value = datetime.strptime(time_string, '%H:%M:%S')
        if callable(self.value_changed_callback):
            self.value_changed_callback(self.value)

    def setup_exit_from_widget(self, key=None) -> None:
        self.hours.editing = False
        self.minutes.editing = False
        self.seconds.editing = False
        self.editing = False


class TimeBox(Textfield):
    def __init__(self, screen, name: str = None, time_changed_callback: Callable = None, *args, **kwargs):
        self.name = name
        self.time_changed_callback = time_changed_callback
        super().__init__(screen, *args, **kwargs)

    def display_value(self, value: datetime) -> str:
        return value.strftime('%H:%M') if isinstance(value, datetime) else _('-unset-')

    def update(self, **keywords) -> None:
        keywords.update({'cursor': False})
        super().update(**keywords)

    def edit(self) -> None:
        super(Textfield, self).edit()

    def set_up_handlers(self) -> None:
        super(Textfield, self).set_up_handlers()
        self.handlers.update({curses.ascii.SP: self.h_change_value,
                              curses.ascii.NL: self.h_change_value})

    def h_change_value(self, *arg) -> None:
        form = TimeForm(name=self.name,
                        value=self.value,
                        return_exit=True,
                        select_exit=True,
                        scroll_exit=False,
                        lines=21,
                        columns=70,
                        value_changed_callback=self.time_changed)
        form.edit()

    def time_changed(self, value: datetime) -> None:
        self.value = value
        if callable(self.time_changed_callback):
            self.time_changed_callback(value)


class TitleTimeBox(TitleWidget):
    _entry_type = TimeBox

    def __init__(self, *args, **kwargs):
        kwargs['value'] = datetime.now() if kwargs.get('show_time_now', None) is True else None
        kwargs.pop('show_time_now', None)
        super().__init__(*args, **kwargs)
