import calendar
import curses
from datetime import datetime
from typing import Callable, TYPE_CHECKING

from npyscreen import DateCombo, Popup, MonthBox

from astra_installer_console.ui.title_widget import TitleWidget

if TYPE_CHECKING:
    from gettext import gettext as _


class MonthBoxCallback(MonthBox):
    def __init__(self, *args, **kwargs):
        self.selection_changed_callback = kwargs.get('selection_changed_callback', None)  # type: Callable
        kwargs.pop('selection_changed_callback', None)
        super().__init__(*args, **kwargs)

    def set_up_handlers(self) -> None:
        super().set_up_handlers()
        self.handlers.update({curses.ascii.SP: self.date_apply,
                              curses.ascii.NL: self.date_apply})

    def date_apply(self, key) -> None:
        self.h_exit_down(key)
        if callable(self.selection_changed_callback):
            self.selection_changed_callback(self.value)

    def update(self, clear=True):
        calendar.setfirstweekday(self.firstWeekDay)
        if clear:
            self.clear()
        if self.hidden:
            self.clear()
            return False

        # Title line
        if not self.value:
            _title_line = _('Value not set')
        else:
            year = self.value.year
            month = self.value.month
            monthname = self.mouth_to_str(self.value.month)
            day = self.value.day

            _title_line = "%s, %s" % (monthname, year)

        if isinstance(_title_line, bytes):
            _title_line = _title_line.decode(self.encoding, 'replace')

        if self.do_colors():
            title_attribute = self.parent.theme_manager.findPair(self)
        else:
            title_attribute = curses.A_NORMAL

        self.add_line(self.rely, self.relx,
                      _title_line,
                      self.make_attributes_list(_title_line, title_attribute),
                      self.width - 1
                      )

        if self.value:
            # Print the days themselves
            try:
                cal_data = calendar.monthcalendar(year, month)
                do_cal_print = True
            except OverflowError:
                do_cal_print = False
                self.parent.curses_pad.addstr(self.rely + 1, self.relx, _('Unable to display'))
                self.parent.curses_pad.addstr(self.rely + 2, self.relx, _('calendar for date'))
            if do_cal_print:
                # Print the day names
                # weekheader puts an extra space at the end of each name

                cal_header = calendar.weekheader(self.__class__.DAY_FIELD_WIDTH - 1)
                if isinstance(cal_header, bytes):
                    cal_header = cal_header.decode(self.encoding, 'replace')

                if self.do_colors():
                    cal_title_attribute = self.parent.theme_manager.findPair(self, 'LABEL')
                else:
                    cal_title_attribute = curses.A_NORMAL
                self.add_line(self.rely + 1, self.relx,
                              cal_header,
                              self.make_attributes_list(cal_header, cal_title_attribute),
                              self.width,
                              )

                print_line = self.rely + 2

                for calrow in cal_data:
                    print_column = self.relx

                    for thisday in calrow:
                        if thisday is 0:
                            pass
                        elif day == thisday:
                            if self.do_colors():
                                self.parent.curses_pad.addstr(print_line, print_column, str(thisday),
                                                              curses.A_STANDOUT | self.parent.theme_manager.findPair(
                                                                  self, self.color))
                            else:
                                self.parent.curses_pad.addstr(print_line, print_column, str(thisday), curses.A_STANDOUT)
                        else:
                            if self.do_colors():
                                self.parent.curses_pad.addstr(print_line, print_column, str(thisday),
                                                              self.parent.theme_manager.findPair(self, self.color))
                            else:
                                self.parent.curses_pad.addstr(print_line, print_column, str(thisday))
                        print_column += self.__class__.DAY_FIELD_WIDTH

                    print_line += 1

            self.print_help()

    def print_help(self):
        mouth_text = 'm(SHIFT + m) - {text}'.format(text=_('Next(previous) mouth'))
        year_text = 'y(SHIFT + y) - {text}'.format(text=_('Next(previous) year'))

        if self.do_colors():
            self.parent.curses_pad.addstr(self.rely + 8,
                                          self.relx,
                                          mouth_text,
                                          self.parent.theme_manager.findPair(self, 'DEFAULT'))
            self.parent.curses_pad.addstr(self.rely + 9,
                                          self.relx,
                                          year_text,
                                          self.parent.theme_manager.findPair(self, 'DEFAULT'))
        else:
            self.parent.curses_pad.addstr(self.rely + 8,
                                          self.relx,
                                          mouth_text)
            self.parent.curses_pad.addstr(self.rely + 9,
                                          self.relx,
                                          year_text)

    @staticmethod
    def mouth_to_str(mouth: int) -> str:
        mouths = [
            _('January'),
            _('February'),
            _('March'),
            _('April'),
            _('May'),
            _('June'),
            _('July'),
            _('August'),
            _('September'),
            _('October'),
            _('November'),
            _('December')]
        try:
            return mouths[mouth - 1]
        except IndexError:
            return _('Unknown mouth')


class TitleDateBox(DateCombo):
    def __init__(self, widget=None, selection_changed_callback: Callable = None, *args, **kwargs):
        self.selection_changed_callback = selection_changed_callback
        self._old_value = kwargs.get('value', None)
        super().__init__(widget, *args, **kwargs)

    def h_change_value(self, *arg) -> None:
        _old_date = self.value
        form = Popup(name=self.name,
                     columns=(MonthBox.DAY_FIELD_WIDTH * 7) + 42,
                     lines=21, )
        monthbox = form.add(MonthBoxCallback,
                            allowPastDate=self.allow_date_in_past,
                            allowTodaysDate=self.allow_todays_date,
                            use_max_space=True,
                            use_datetime=self.use_datetime,
                            allowClear=self.allow_clear,
                            selection_changed_callback=self.selection_changed)
        try:
            self.value.isoformat()
            monthbox.value = self.value
        except:
            monthbox.value = self.date_or_datetime().today()
            monthbox._check_today_validity()
        form.display()
        monthbox.edit()

    def display_value(self, value) -> str:
        if self.value:
            return '{day} {mouth}, {year}'.format(day=self.value.strftime('%d'),
                                                  mouth=MonthBoxCallback.mouth_to_str(self.value.month),
                                                  year=self.value.strftime('%Y'))
        else:
            return '-{text}-'.format(text=_('unset'))

    def selection_changed(self, value: datetime) -> None:
        self.value = value
        if callable(self.selection_changed_callback):
            self.selection_changed_callback(self.value)


class TitleDate(TitleWidget):
    _entry_type = TitleDateBox
