from typing import Tuple, Optional, TYPE_CHECKING

from npyscreen import TitleText

from astra_installer.settings.manager import SettingsManager
from astra_installer_console.ui.grub_title import GrubTitle
from astra_installer_console.ui.password_text import TitlePasswordEntry

if TYPE_CHECKING:
    from gettext import gettext as _


class UserManager:
    def __init__(self, form, select_widgets_at, max_width):
        self.form = form
        self.backend = SettingsManager.user
        self.username = None  # type: TitleText
        self.password = None  # type: TitlePasswordEntry
        self.password_confirmation = None  # type: TitlePasswordEntry
        self.grub_title = None  # type GrubTitle
        self.setup(select_widgets_at, max_width)

    def setup(self, select_widgets_at, max_width) -> None:
        self.username = self.form.add(TitleText,
                                      name=_('Username:'),
                                      begin_entry_at=select_widgets_at,
                                      max_width=max_width,
                                      value=self.backend.username)

        self.password = self.form.add(TitlePasswordEntry,
                                      name=_('Password:'),
                                      begin_entry_at=select_widgets_at,
                                      max_width=max_width,
                                      value=self.backend.password)

        self.password_confirmation = self.form.add(TitlePasswordEntry,
                                                   name=_('Password confirmation:'),
                                                   begin_entry_at=select_widgets_at,
                                                   max_width=max_width,
                                                   value=self.backend.password)

        if SettingsManager.systeminfo.arch != "e2k-8c":
            self.grub_title = self.form.add(GrubTitle,
                                            name=_('GRUB password settings:'),
                                            begin_entry_at=select_widgets_at,
                                            max_width=max_width)

    def validate(self) -> Tuple[bool, Optional[str]]:
        valid, error = self.backend.is_valid_username(self.username.value)
        if not valid:
            return False, error
        valid, error = self.backend.is_valid_password(self.password.value)
        if not valid:
            return False, error
        elif not self.is_equal_passwords():
            return False, _('Passwords do not match')
        else:
            valid, error = self.validate_grub_password()
            if not valid:
                return False, error
            self.backend.username = self.username.value
            self.backend.password = self.password.value
            if self.grub_title:
                self.backend.setup_grub_password = self.grub_title.setup_grub_password
                if self.grub_title.grub_password:
                    self.backend.grub_password = self.grub_title.grub_password
                else:
                    self.backend.grub_password = self.backend.password
            return True, None

    def validate_grub_password(self) -> Tuple[bool, Optional[str]]:
        if not self.grub_title:
            return True, None
        if not self.grub_title.setup_grub_password:
            return True, None
        if self.grub_title.grub_password != self.grub_title.grub_password_confirmation:
            return False, _('GRUB passwords do not match')
        if self.grub_title.grub_password:
            valid, error = self.backend.is_valid_password(self.grub_title.grub_password)
            if not valid:
                text = _('GRUB password settings:') + '\n' + error
                return False, text
        return True, None

    def is_equal_passwords(self):
        return self.password.value == self.password_confirmation.value
