# Copyright (C) 2009-2011
# Red Hat, Inc.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors: Chris Lumens <clumens@redhat.com>
#

from typing import TYPE_CHECKING

from blivet.size import Size
from blivet.devicelibs import raid

from astra_installer_kickstart.specification import PartSpec

if TYPE_CHECKING:
    from gettext import gettext as _


class Platform:
    """Platform

       A class containing platform-specific information and methods for use
       during installation.  The intent is to eventually encapsulate all the
       architecture quirks in one place to avoid lots of platform checks
       throughout anaconda."""
    _packages = []

    # requirements for bootloader stage1 devices
    _boot_stage1_device_types = []
    _boot_stage1_format_types = []
    _boot_stage1_mountpoints = []
    _boot_stage1_max_end = None
    _boot_stage1_raid_levels = []
    _boot_stage1_raid_metadata = []
    _boot_stage1_raid_member_types = []
    _boot_stage1_description = _("boot loader device")
    _boot_stage1_missing_error = ""
    _boot_raid_description = _("RAID Device")
    _boot_partition_description = _("First sector of boot partition")
    _boot_descriptions = {}

    _non_linux_format_types = []

    def __init__(self):
        """Creates a new Platform object.  This is basically an abstract class.
           You should instead use one of the platform-specific classes as
           returned by get_platform below.  Not all subclasses need to provide
           all the methods in this class."""

        # self.update_from_flags()
        pass

    # def update_from_flags(self):
    #     if conf.storage.gpt:
    #         disklabel_class = get_device_format_class("disklabel")
    #         disklabel_types = disklabel_class.get_platform_label_types()
    #         if "gpt" not in disklabel_types:
    #             log.warning("GPT is not a supported disklabel on this platform. Using default "
    #                         "disklabel %s instead.", disklabel_types[0])
    #         else:
    #             disklabel_class.set_default_label_type("gpt")

    def __call__(self):
        return self

    @property
    def boot_stage1_constraint_dict(self):
        d = {"device_types": self._boot_stage1_device_types,
             "format_types": self._boot_stage1_format_types,
             "mountpoints": self._boot_stage1_mountpoints,
             "max_end": self._boot_stage1_max_end,
             "raid_levels": self._boot_stage1_raid_levels,
             "raid_metadata": self._boot_stage1_raid_metadata,
             "raid_member_types": self._boot_stage1_raid_member_types,
             "descriptions": dict((k, _(v)) for k, v in self._boot_descriptions.items())}
        return d

    @property
    def packages(self):
        _packages = self._packages
        return _packages

    def set_platform_bootloader_reqs(self):
        """Return the required platform-specific bootloader partition
           information.  These are typically partitions that do not get mounted,
           like biosboot or prepboot, but may also include the /boot/efi
           partition."""
        return []

    def set_platform_boot_partition(self):
        """Return the default /boot partition for this platform."""
        return [PartSpec(mountpoint="/boot", size=Size("1GiB"))]

    def set_default_partitioning(self):
        """Return the default platform-specific partitioning information."""
        return self.set_platform_bootloader_reqs() + self.set_platform_boot_partition()

    @property
    def stage1_missing_error(self):
        """A platform-specific error message to be shown if stage1 target
           selection fails."""
        return self._boot_stage1_missing_error


class CommonPlatform(Platform):
    pass


class EFI(Platform):
    _boot_stage1_format_types = ["efi"]
    _boot_stage1_device_types = ["partition", "mdarray"]
    _boot_stage1_mountpoints = ["/boot/efi"]
    _boot_stage1_raid_levels = [raid.RAID1]
    _boot_stage1_raid_metadata = ["1.0"]
    _boot_efi_description = _("EFI System Partition")
    _boot_descriptions = {"partition": _boot_efi_description,
                          "mdarray": Platform._boot_raid_description}

    # XXX hpfs, if reported by blkid/udev, will end up with a type of None
    _non_linux_format_types = ["vfat", "ntfs", "hpfs"]
    _boot_stage1_missing_error = _("For a UEFI installation, you must include "
                                   "an EFI System Partition on a GPT-formatted "
                                   "disk, mounted at /boot/efi.")

    def set_platform_bootloader_reqs(self):
        ret = super().set_platform_bootloader_reqs()
        ret.append(PartSpec(mountpoint="/boot/efi", fstype="efi",
                            size=Size("200MiB"), max_size=Size("600MiB"),
                            grow=True))
        return ret
