#
# Copyright (C) 2009  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Chris Lumens <clumens@redhat.com>
#

from blivet.util import stringize, unicodeize
from pykickstart.base import KickstartHandler
from pykickstart.parser import KickstartParser
from pykickstart.version import DEVEL as VERSION

from astra_installer_kickstart.addon import AddonSection, AddonRegistry
from astra_installer_kickstart import commands as COMMANDS
from astra_installer_kickstart.commands.commands import AutoPart, BTRFS, ClearPart, Fcoe, Iscsi
from astra_installer_kickstart.commands.commands import Snapshot, ZFCP, IgnoreDisk, IscsiName, Nvdimm


class PartSpec:
    def __init__(self, mountpoint=None, fstype=None, size=None, max_size=None,
                 grow=False, btr=False, lv=False, thin=False, weight=0,
                 required_space=0, encrypted=False):
        """ Create a new storage specification.  These are used to specify
            the default partitioning layout as an object before we have the
            storage system up and running.  The attributes are obvious
            except for the following:

            btr -- Should this be allocated as a btrfs subvolume?  If not,
                   it will be allocated as a partition.
            lv -- Should this be allocated as a logical volume?  If not,
                  it will be allocated as a partition.
            thin -- Should this be allocated as a thin logical volume if it is
                    being allocated as a logical volume?
            weight -- An integer that modifies the sort algorithm for partition
                      requests.  A larger value means the partition will end up
                      closer to the front of the disk.  This is mainly used to
                      make sure /boot ends up in front, and any special (PReP,
                      appleboot, etc.) partitions end up in front of /boot.
                      This value means nothing unless lv and btr are both False.
            required_space -- This value is only taken into account if
                             lv=True, and specifies the size in MiB that the
                             containing VG must be for this PartSpec to even
                             get used.  The VG's size is calculated before any
                             other LVs are created inside it.  If not enough
                             space exists, this PartSpec will never get turned
                             into an LV.
            encrypted -- Should this request be encrypted? For logical volume
                         requests, this is satisfied if the PVs are encrypted
                         as in the case of encrypted LVM autopart.
        """

        self.mountpoint = mountpoint
        self.fstype = fstype
        self.size = size
        self.max_size = max_size
        self.grow = grow
        self.lv = lv
        self.btr = btr
        self.thin = thin
        self.weight = weight
        self.required_space = required_space
        self.encrypted = encrypted

    # Force str and unicode types in case any of the properties are unicode
    def _to_string(self):
        s = ("%(type)s instance (%(id)s) -- \n"
             "  mountpoint = %(mountpoint)s  lv = %(lv)s"
             "  thin = %(thin)s  btrfs = %(btrfs)s\n"
             "  weight = %(weight)s  fstype = %(fstype)s  encrypted = %(enc)s\n"
             "  size = %(size)s  max_size = %(max_size)s  grow = %(grow)s\n" %
             {"type": self.__class__.__name__, "id": "%#x" % id(self),
              "mountpoint": self.mountpoint, "lv": self.lv, "btrfs": self.btr,
              "weight": self.weight, "fstype": self.fstype, "size": self.size,
              "enc": self.encrypted, "max_size": self.max_size, "grow": self.grow,
              "thin": self.thin})

        return s

    def __str__(self):
        return stringize(self._to_string())

    def __unicode__(self):
        return unicodeize(self._to_string())

    def __eq__(self, other):
        return isinstance(other, PartSpec) and vars(self) == vars(other)


class KickstartSpecification:
    """Specification of kickstart data.

    This specification can be used to get the corresponding
    handler and parser to parse and handle kickstart data
    described by this specification.

    You should call get_kickstart_handler to get the kickstart
    handler for this specification.

    You should call get_kickstart_parser to get the kickstart
    parser for this specification.

    A specification is defined by these attributes:

    version       - version of this specification
    commands      - mapping of kickstart command names to
                    classes that represent them
    commands_data - mapping of kickstart data names to
                    classes that represent them
    sections      - mapping of kickstart sections names to
                    classes that represent them
    sections_data - mapping of kickstart sections data names to
                    classes that represent them
    addons        - mapping of kickstart addons names to
                    classes that represent them

    """

    version = VERSION
    commands = {}
    commands_data = {}
    sections = {}
    sections_data = {}
    addons = {}


class NoKickstartSpecification(KickstartSpecification):
    """Specification for no kickstart data."""
    pass


class KickstartSpecificationHandler(KickstartHandler):
    """Handler defined by a kickstart specification."""

    def __init__(self, specification):
        super().__init__()
        self.version = specification.version

        for name, command in specification.commands.items():
            self.registerCommand(name, command)

        for name, data in specification.commands_data.items():
            self.registerData(name, data)

        for name, data in specification.sections_data.items():
            self.registerSectionData(name, data)

        if specification.addons:
            self.addons = AddonRegistry()

        for name, data in specification.addons.items():
            self.registerAddonData(name, data)

    def registerSectionData(self, name, data):
        """Register data used by a section."""
        obj = data()
        setattr(self, name, obj)
        self._registerWriteOrder(obj)

    def registerAddonData(self, name, data):
        """Register data used by %addon."""
        obj = data()
        setattr(self.addons, name, obj)
        self._registerWriteOrder(obj)

    def _registerWriteOrder(self, obj):
        """Write the object at the end of the output."""
        write_priority = 0

        if self._writeOrder:
            write_priority = max(self._writeOrder.keys()) + 100

        self._writeOrder[write_priority] = [obj]


class KickstartSpecificationParser(KickstartParser):
    """Parser defined by a kickstart specification."""

    def __init__(self, handler, specification):
        super().__init__(handler)

        for section in specification.sections.values():
            self.registerSection(section(handler))

        if specification.addons:
            self.registerSection(AddonSection(handler))

    def setupSections(self):
        """Do not setup any default sections."""
        pass


class StorageKickstartSpecification(KickstartSpecification):
    """Kickstart specification of the storage module."""

    commands = {
        "autopart": AutoPart,
        "bootloader": COMMANDS.Bootloader,
        "btrfs": BTRFS,
        "clearpart": ClearPart,
        "fcoe": Fcoe,
        "ignoredisk": IgnoreDisk,
        "iscsi": Iscsi,
        "iscsiname": IscsiName,
        "logvol": COMMANDS.LogVol,
        "mount": COMMANDS.Mount,
        "nvdimm": Nvdimm,
        "part": COMMANDS.Partition,
        "partition": COMMANDS.Partition,
        "raid": COMMANDS.Raid,
        "reqpart": COMMANDS.ReqPart,
        "snapshot": Snapshot,
        "volgroup": COMMANDS.VolGroup,
        "zerombr": COMMANDS.ZeroMbr,
        "zfcp": ZFCP,
        "zipl": COMMANDS.Zipl
    }

    commands_data = {
        "BTRFSData": COMMANDS.BTRFSData,
        "FcoeData": COMMANDS.FcoeData,
        "IscsiData": COMMANDS.IscsiData,
        "LogVolData": COMMANDS.LogVolData,
        "MountData": COMMANDS.MountData,
        "NvdimmData": COMMANDS.NvdimmData,
        "PartData": COMMANDS.PartData,
        "RaidData": COMMANDS.RaidData,
        "SnapshotData": COMMANDS.SnapshotData,
        "VolGroupData": COMMANDS.VolGroupData,
        "ZFCPData": COMMANDS.ZFCPData,
    }
