from collections import Iterable
from itertools import cycle

from npyscreen import BoxTitle, MultiLine


class ItemLine(MultiLine):
    def __init__(self, *args, **kwargs):
        self._my_widgets = []
        self.__release_timeout_callback = None
        super().__init__(slow_scroll=True, *args, **kwargs)

    def display_value(self, item) -> None:
        value = item
        if isinstance(item, Iterable):
            value = next(iter(item), None)
        return super().display_value(value)

    def get_selected_objects(self) -> None:
        value = self.value
        if value is not None:
            value = self.values[self.value]
        return value

    def when_value_edited(self) -> None:
        if self.value is not None:
            self.values[self.value][1]()

    def calculate_area_needed(self):
        y, x = super().calculate_area_needed()

        for item in self._my_widgets:
            text_len = len(item.value or [])
            if text_len > x:
                x = text_len

        return y, x

    def resize(self):
        super().resize()
        y, x = self.calculate_area_needed()
        for widget in self._my_widgets:
            widget.width = x + 2
            widget.resize()

    def retranslate(self, items=None):
        if not isinstance(items, Iterable):
            return
        item = cycle(items)
        new_values = []
        for first, second in self.values:
            new_values.append((next(item), second))
        self.values = new_values

    def set_value(self, value):
        self.value = value
        self._old_value = value

    def set_release_timeout_callback(self, release_timeout_callback = None):
        self.__release_timeout_callback = release_timeout_callback

    def handle_input(self, _input):
        super().handle_input(_input)
        if self.__release_timeout_callback:
            self.__release_timeout_callback()

class MainBoxTitle(BoxTitle):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    _contained_widget = ItemLine

    def calculate_area_needed(self):
        max_height, max_width = super().calculate_area_needed()

        name_len = len(self.name or []) + 10
        if name_len > max_width:
            max_width = name_len

        if hasattr(self, 'entry_widget'):
            y, x = self.entry_widget.calculate_area_needed()
            if x > max_width:
                max_width = x

            max_height = len(self.entry_widget._my_widgets) + 2

        return max_height, max_width

    def set_relyx(self, y, x):
        super().set_relyx(y, x)
        if hasattr(self, 'entry_widget'):
            self.entry_widget.set_relyx(y + 1, x + 2)

    def retranslate(self, title=None, items=None):
        self.name = title
        self.entry_widget.retranslate(items)

    def set_value(self, value):
        if hasattr(self, 'entry_widget'):
            self.entry_widget.set_value(value)
        else:
            # probably trying to set the value before the textarea is initialised
            self.__tmp_value = value

    def set_release_timeout_callback(self, release_timeout_callback = None):
        self.entry_widget.set_release_timeout_callback(release_timeout_callback)
