import curses
from typing import Optional, Callable, TYPE_CHECKING

from npyscreen import FormBaseNew, notify_confirm

from astra_installer.settings.config import ConfigLoader
from astra_installer.settings.manager import SettingsManager
from astra_installer.translator import Translator
from astra_installer.utils import Utils
from astra_installer_launcher.ui.hotkeys_widget import HotkeysWidget
from astra_installer_launcher.ui.main_box_title import MainBoxTitle
from astra_installer_launcher.ui.select_form import SelectForm
from astra_installer_launcher.ui.select_language_form import SelectLanguage
from astra_installer_launcher.ui.translatable_text import TranslatableText
from astra_installer_launcher.utils import Config, clear_console

if TYPE_CHECKING:
    from gettext import gettext as _


class MainForm(FormBaseNew):
    def __init__(self, *args, **kwargs):
        self.welcome = None  # type: Optional[TranslatableText]
        self.main_actions_box = None  # type: Optional[MainBoxTitle]
        self.hotkeys = None  # type: Optional[HotkeysWidget]
        self.autoinstaller_item = None  # type: Optional[dict]
        self.keypress_waiting_time = 0
        self.translator = Translator
        self.setup()
        super().__init__(*args, **kwargs)

    @staticmethod
    def setup() -> None:
        config = ConfigLoader().load()
        SettingsManager.create_settings(config)

    def create(self) -> None:
        height, width = self.useable_space()

        items_for_main_actions = []
        for item_data in Config.parse_main_menu_data():
            item = (Config.menu_data_text(item_data), lambda data=item_data: self.h_line_toggled(data))
            items_for_main_actions.append(item)
            if item_data.get('execute_if_preseed', False):
                self.autoinstaller_item = item_data

        self.nextrely += (height - len(items_for_main_actions) - 10) // 2

        self.welcome = self.add(TranslatableText,
                                editable=False)

        self.nextrely += 1

        self.main_actions_box = self.add(MainBoxTitle,
                                         max_height=len(items_for_main_actions) + 2,  # TODO: del
                                         values=items_for_main_actions)

        self.hotkeys = self.add(HotkeysWidget)

        self.add_handlers({curses.KEY_F2: self.h_select_language})
        self.add_handlers({curses.KEY_F10: self.h_shutdown})

        self.set_timeout(items_for_main_actions[0][0])
        self.retranslate()

    def h_line_toggled(self, data) -> None:
        self.release_timeout()
        self.main_actions_box.set_value(None)
        self.perform_function(data)

    def perform_function(self, data) -> None:
        execute = data.get('execute', None)
        exit_and_execute = data.get('exit_and_execute', False)
        action_text = Config.menu_data_text(data)

        if not callable(execute):
            notify_confirm(title=_('Error'),
                           message=_('Menu item is not callable'))
        elif exit_and_execute is True:
            self.exit_program(last_command=lambda: MainForm.execute_function(action_text, execute))
        else:
            clear_console()
            self.execute_function(action_text, execute)

    def h_select_language(self, input) -> None:
        if self.keypress_timeout != 0:
            self.release_timeout()

        locales = SettingsManager.installer_locale.locales

        form_y = 2
        form_lines = len(locales) + 5
        if form_lines + form_y > self.hotkeys.rely:
            form_lines = self.hotkeys.rely - form_y

        form = SelectLanguage(locales=locales,
                              current_locale=SettingsManager.installer_locale.selected,
                              parent=self,
                              lines=form_lines,
                              show_atx=2,
                              show_aty=form_y)

        form.edit()

        locale = form.locale()
        if locale is not None:
            self.set_locale(locale)

    def set_locale(self, locale: str) -> None:
        SettingsManager.installer_locale.change_locale(locale)
        self.retranslate()
        self.resize()

    def h_shutdown(self, input) -> None:
        if self.keypress_timeout != 0:
            self.release_timeout()

        items = [(_('Shutdown computer'), Utils.poweroff),
                 (_('Reboot computer'), Utils.reboot),
                 (_('Exit to console'), None)]
        title = _('Shutdown')
        footer = 'F10, ESC - {cancel_text} | ENTER, SPACE - {apply_text}'.format(cancel_text=_('Cancel'),
                                                                                 apply_text=_('Apply'))

        max_len = max(len(title), len(footer))
        for item in items:
            max_len = max(len(item[0]), max_len)

        form_y = 2
        form_lines = len(items) + 5
        if form_lines + form_y > self.hotkeys.rely:
            form_lines = self.hotkeys.rely - form_y

        form = SelectForm(name=title,
                          items=items,
                          footer=footer,
                          parent=self,
                          lines=form_lines,
                          columns=max_len + 6,
                          show_atx=2,
                          show_aty=form_y)
        form.add_handlers({curses.KEY_F10: form.h_cancel})

        form.edit()

        if form.value is not None:
            self.exit_program(form.value[1])

    def exit_program(self, last_command=None) -> None:
        self.editing = False
        self.parentApp.exit(last_command)

    def retranslate(self) -> None:
        self.welcome.retranslate(_('Astra Linux OS installer'))
        self.main_actions_box.retranslate(title=_('Select an action'),
                                          items=Config.get_main_menu_texts())
        self.hotkeys.retranslate()

    def resize(self) -> None:
        height, width = self.useable_space()

        y, x = self.welcome.calculate_area_needed()
        self.welcome.set_relyx(self.welcome.rely, (width - x) // 2)
        self.welcome.width = x
        self.welcome.height = y
        self.welcome.resize()

        y, x = self.main_actions_box.calculate_area_needed()
        self.main_actions_box.set_relyx(self.main_actions_box.rely, (width - x) // 2)
        self.main_actions_box.width = x
        self.main_actions_box.height = y
        self.main_actions_box.resize()

        y, x = self.hotkeys.calculate_area_needed()
        self.hotkeys.set_relyx(height - y - 2, self.hotkeys.relx)
        self.hotkeys.width = x + 1
        self.hotkeys.height = y
        self.hotkeys.resize()

        self.display()
        self.welcome.display()
        self.main_actions_box.display()
        self.hotkeys.display()

    def get_timeout_text(self, current_item):
        return "{item} {text} {timeout} {units}" \
                .format(item=current_item,
                        text=_("will be started in"),
                        timeout=self.keypress_waiting_time,
                        units=_("sec"))

    def set_timeout(self, current_item):
        self.keypress_waiting_time = Config.get_keypress_waiting_time()
        if self.keypress_waiting_time > 0:
            self.keypress_timeout = 10
            self.main_actions_box.set_release_timeout_callback(self.release_timeout)
            self.hotkeys.set_timeout(self.get_timeout_text(current_item))

    def release_timeout(self):
        self.keypress_timeout = 0
        self.hotkeys.set_timeout("")
        self.retranslate()
        self.resize()
    
    def timetick(self):
        self.keypress_waiting_time -= 1
        self.hotkeys.set_timeout(self.get_timeout_text(self.main_actions_box.get_values()[0][0]))
        self.retranslate()
        self.resize()

    def while_waiting(self) -> None:
        self.timetick()
        if self.keypress_waiting_time <= 0:
            self.release_timeout()
            try:
                toogle_line = self.main_actions_box.get_values()[0][1]
                toogle_line()
            except Exception as e:
                print(e)

    @staticmethod
    def execute_function(action_text: str, execute_function: Callable) -> None:
        print(_('"{action}" is running...').format(action=action_text))
        execute_function()
